#include "BIN_IO.h"


BIN_IO::BIN_IO (string fileName, Region* destination) {
	_data = destination;
	_fileName = fileName;
	MPI_Comm_rank(BASIN_COMM, &_myRank);
}

BIN_IO::~BIN_IO () { }

bool BIN_IO::OpenFile () {
	_file.open (_fileName.c_str (), std::ifstream::in);

	if (not _file.is_open ()) {
		return false;
	}

	_file.seekg (0, ios::end);
	_fileSize = _file.tellg ();
	_file.seekg (0, ios::beg);
	return true;
}

bool BIN_IO::CloseFile () {
	_file.close ();
	return true;
}

bool BIN_IO::GetDatasets () {

	/* Get Header */
	int dummy, numDMParticles, numGasParticles;
	_file.read (reinterpret_cast<char*> (&dummy), sizeof (dummy));
	_file.read (reinterpret_cast<char*> (&numDMParticles), sizeof (numDMParticles));
	_file.read (reinterpret_cast<char*> (&numGasParticles), sizeof (numGasParticles));
	float omegaMatterZ0, omegaBaryon, lambda, hubble, scaleFactorAtEpoch0, scaleFactor, timeInUnit;
	_file.read (reinterpret_cast<char*> (&omegaMatterZ0), sizeof (omegaMatterZ0));
	_file.read (reinterpret_cast<char*> (&omegaBaryon), sizeof (omegaBaryon));
	_file.read (reinterpret_cast<char*> (&lambda), sizeof (lambda));
	_file.read (reinterpret_cast<char*> (&hubble), sizeof (hubble));
	_file.read (reinterpret_cast<char*> (&scaleFactorAtEpoch0), sizeof (scaleFactorAtEpoch0));
	_file.read (reinterpret_cast<char*> (&scaleFactor), sizeof (scaleFactor));
	_file.read (reinterpret_cast<char*> (&timeInUnit), sizeof (timeInUnit));
	double scaleUnit, massUnit, timeUnit;
	_file.read (reinterpret_cast<char*> (&scaleUnit), sizeof (scaleUnit));
	_file.read (reinterpret_cast<char*> (&massUnit), sizeof (massUnit));
	_file.read (reinterpret_cast<char*> (&timeUnit), sizeof (timeUnit));


	/*
	std::cout << "                          Dummy: " << dummy << endl;
	std::cout << "Number of Dark Matter Particles: " << numDMParticles << endl;
	std::cout << "        Number of Gas Particles: " << numGasParticles << endl;
	std::cout << "          Omega Matter at z = 0: " << omegaMatterZ0 << endl;
	std::cout << "                   Omega Baryon: " << omegaBaryon << endl;
	std::cout << "                         Lambda: " << lambda << endl;
	std::cout << "                         Hubble: " << hubble << endl;
	std::cout << "  Scale Factor at Initial Epoch: " << scaleFactorAtEpoch0 << endl;
	std::cout << "                   Scale Factor: " << scaleFactor << endl;
	std::cout << "              Time in this Unit: " << timeInUnit << endl;
	std::cout << "                     Scale Unit: " << scaleUnit << endl;
	std::cout << "                      Mass Unit: " << massUnit << endl;
	std::cout << "                      Time Unit: " << timeUnit << endl;
	*/

	List* list = new List (_data);

	/* Get Positions */
	{
		Attribute xPos = Attribute (numDMParticles, 0.0);
		vector <int> procIndices = xPos.get_gen_distribution_index (_myRank);

		unsigned int procParticles = procIndices[1] - procIndices[0] + 1;
		double* xArray = new double[procParticles];
		double* yArray = new double[procParticles];
		double* zArray = new double[procParticles];

		float x, y, z;
		_file.seekg (3 * (sizeof (x)) * procIndices[0], ios::cur);
		for (int i = 0; i < procParticles; ++i) {
			_file.read (reinterpret_cast<char*> (&x), sizeof (x));
			_file.read (reinterpret_cast<char*> (&y), sizeof (y));
			_file.read (reinterpret_cast<char*> (&z), sizeof (z));
			xArray[i] = static_cast<double> (x);
			yArray[i] = static_cast<double> (y);
			zArray[i] = static_cast<double> (z);
		}

		
		xPos.put (procIndices[0], procIndices[1], xArray);
		delete [] xArray;
		list->add_attribute ("xPos", xPos);

		Attribute yPos = Attribute (numDMParticles, 0.0);
		yPos.put (procIndices[0], procIndices[1], yArray);
		delete [] yArray;
		list->add_attribute ("yPos", yPos);

		Attribute zPos = Attribute (numDMParticles, 0.0);
		zPos.put (procIndices[0], procIndices[1], zArray);
		delete [] zArray;
		list->add_attribute ("zPos", zPos);
	}

	/* Get Velocities */
	{
		Attribute xVel = Attribute (numDMParticles, 0.0);
		vector <int> procIndices = xVel.get_gen_distribution_index (_myRank);

		unsigned int procParticles = procIndices[1] - procIndices[0] + 1;
		double* xArray = new double[procParticles];
		double* yArray = new double[procParticles];
		double* zArray = new double[procParticles];

		float x, y, z;
		_file.seekg (3 * (sizeof (x)) * (numDMParticles - procIndices[1] - 1), ios::cur);
		for (int i = 0; i < procParticles; ++i) {
			_file.read (reinterpret_cast<char*> (&x), sizeof (x));
			_file.read (reinterpret_cast<char*> (&y), sizeof (y));
			_file.read (reinterpret_cast<char*> (&z), sizeof (z));
			xArray[i] = static_cast<double> (x);
			yArray[i] = static_cast<double> (y);
			zArray[i] = static_cast<double> (z);
		}

		xVel.put (procIndices[0], procIndices[1], xArray);
		delete [] xArray;
		list->add_attribute ("xVel", xVel);

		Attribute yVel = Attribute (numDMParticles, 0.0);
		yVel.put (procIndices[0], procIndices[1], yArray);
		delete [] yArray;
		list->add_attribute ("yVel", yVel);

		Attribute zVel = Attribute (numDMParticles, 0.0);
		zVel.put (procIndices[0], procIndices[1], zArray);
		delete [] zArray;
		list->add_attribute ("zVel", zVel);
	}

	/* Get ID's */
	{
		Attribute ids = Attribute (numDMParticles, 0);
		vector <int> procIndices = ids.get_gen_distribution_index (_myRank);

		unsigned int procParticles = procIndices[1] - procIndices[0] + 1;
		int* idsArray = new int[procParticles];

		int id;
		_file.seekg (3 * (sizeof (id)) * (numDMParticles - procIndices[1] - 1), ios::cur);
		for (int i = 0; i < procParticles; ++i) {
			_file.read (reinterpret_cast<char*> (&id), sizeof (id));
			idsArray[i] = id;
		}

		ids.put (procIndices[0], procIndices[1], idsArray);
		delete [] idsArray;
		list->add_attribute ("id", ids);
	}

	_data->add_list (_fileName, list);

	return true;
}

bool BIN_IO::NextDataset () {
	/* This format has no notion of a next dataset */
	return false;
}

bool BIN_IO::GetGroups () {
	/* This format has no notion of a group */
	return true;
}

bool BIN_IO::NextGroup () {
	/* This format has no notion of a next group */
	return false;
}
