#include "basin.h"

int BASIN_ID, BASIN_NPROCS;

MPI_Comm BASIN_COMM;

vector<int> basin_initialize(int &ID, int &NP,MPI_Comm c)
{
  /*

  - This function assigns the ID and the number of processes to the 
    global variables, BASIN_ID, BASIN_GLOBAL.
    
  -It returns a vector<int> with two elements 
    (V[0]=number of processes,V[1]=id).
    The user who wants to use it this way will type

    vector<int> a(2);
    a=basin_initialize();
    and then uses a[0] and a[1] if the id or the nprocs are needed

  - The last option is to use it by passing two integers by reference 
    as arguments and get the values of id and number of processes in those 
    two integers; the function will still return a vector<int> but that will 
    not be used.The global variables will still be set as they are needed for 
    the definition of bout
    int myid,nprocs;
    basin_initialize(myid,nprocs);

    So you have several different options; maybe too many for such a 
    simple function.  

  */



  vector<int> nproc_and_id(2);
  MPI::Init_thread(MPI_THREAD_MULTIPLE);
  if (MPI_Comm_dup(MPI_COMM_WORLD, &BASIN_COMM) != MPI_SUCCESS)
  BASIN_COMM = MPI_COMM_WORLD;

  MPI_Comm_size(BASIN_COMM, &nproc_and_id[0]);
  MPI_Comm_rank(BASIN_COMM, &nproc_and_id[1]);

  BASIN_NPROCS=nproc_and_id[0];
  BASIN_ID=nproc_and_id[1];

  //if the default values for ID and NP are used these would not be necessary
  NP=nproc_and_id[0];
  ID=nproc_and_id[1];

  return nproc_and_id;

}


/// Calls MPI_Finalize
void basin_finalize()
{
  MPI_Finalize();
}





