#include "Context.h"
#include "Data.h"

Context::Context()
{
  this->parent=NULL;
  this->lower_left_corner.push_back(0.0);
  this->upper_right_corner.push_back(1.0);
  this->n_grid_points.push_back(2);
  this->set_order_type(NORMAL);
}


Context::Context(const Context& c) 
{
  vector<vector<double> > temp;
  temp=c.get_boundary();
  vector<int> tempnx;
  tempnx=c.get_n_grid_points();
  this->set_boundary(temp[0],temp[1]);
  this->set_parent(c.get_parentData_pointer());
  this->set_order_type(c.get_order_type());
}


  
Context::Context(vector<double> ll, vector<double> ur, vector<int> nx, Data* parent)
{
  //TODO: ADD A CHECK THAT THE SIZE OF ALL THIS VECTORS IS THE SAME
  this->set_boundary(ll,ur);
  this->set_n_grid_points(nx);
  this->set_parent(parent);
  this->set_order_type(NORMAL);
}


void Context::set_boundary(vector<double> ll, vector<double> ur) 
{
  this->lower_left_corner=ll;
  this->upper_right_corner=ur;
}

void Context::set_n_grid_points(vector<int> nxyz)
{
  this->n_grid_points=nxyz;
  int ndims=nxyz.size();
}

vector<int> Context::get_n_grid_points() const
{
  return this->n_grid_points;
}

vector<int> Context::get_index(int i0) {
  vector<int> nx;
  nx=this->get_n_grid_points();
  int ndims=nx.size();
  vector<int>index(ndims,0);
  vector<int>  nskip(ndims,1);
  for (int i=0; i < ndims-1; i++) {
    for (int j=i+1; j < ndims; j++) {
      nskip[i]=nskip[i]*nx[j];
    }
  }
  for (int i=0; i < ndims; i++) {
    index[i]=(int(i0/nskip[i])) % nx[i];
    if ((this->get_order_type() == FFT_HALF_HERMITIAN)&&(i != ndims-1)&&
	(index[i] > nx[i]/2)) {index[i]=index[i]-nx[i];}
  }
  return index;
}

vector<vector<double> > Context::get_boundary() const
{
  vector<vector<double> > temp;
  temp.push_back(lower_left_corner);
  temp.push_back(upper_right_corner);
  return temp;
}
  
void Context::set_LowerCorner(const vector<double> ll)
{
  this->lower_left_corner=ll;
}

void Context::set_UpperCorner(const vector<double> ur)
{
  this->upper_right_corner=ur;
}
vector<double> Context::get_LowerCorner() const
{
  return this->lower_left_corner;
}
  
vector<double> Context::get_UpperCorner() const
{
  return this->upper_right_corner;
}

void Context::set_parent(Data* pare)
{
  this->parent=pare;
}
Data* Context::get_parentData_pointer() const
{
  return this->parent;
}
List* Context::get_parentList_pointer() const
{
  return dynamic_cast<List*>(this->parent);
}

Grid* Context::get_parentGrid_pointer() const
{
  return dynamic_cast<Grid*>(this->parent);
}

void Context::set_order_type(int t) {
  order_type=t;
}

int Context::get_order_type() const {
  return order_type;
}

		 
