#include "Data.h"
#include "Region.h"
//************************************************************
//
// Constructor functions

Grid::Grid() {parentRegion=NULL;}
Grid::Grid(Region* parent,const vector<int>& nxNew) : Data(parent){
  myGeometry=Geometry(*(parent -> get_geometry()));
  nx=nxNew;
  myGeometry.set_ndims(nxNew.size());
  offset=nx;
  offset[myGeometry.get_ndims()-1]=1;  
  if (myGeometry.get_ndims() > 1) {
    for (int i=myGeometry.get_ndims()-2; i>=0; i--) {
      offset[i]=offset[i+1]*nx[i+1];
    }
  }
}

Grid::Grid(const vector<int>&nxNew) {
  parentRegion=NULL;
  myGeometry=Geometry(nxNew.size(),0,0,0,vector<double>(nxNew.size(),0),vector<double>(nxNew.size(),1));
  nx=nxNew;
  offset=nx;
  offset[myGeometry.get_ndims()-1]=1;  
  if (myGeometry.get_ndims() > 1) {
    for (int i=myGeometry.get_ndims()-2; i>=0; i--) {
      offset[i]=offset[i+1]*nx[i+1];
    }
  }
  

} 

// Grids may also be constructed such that they have a different
// geometry from their parent region or particle.  For example , they
// may be a cartesian image of a galaxy which is, itself, within a
// spherical geometry...  In this case, we'll want to construct from
// geometry...

Grid::Grid(Geometry* geo,const vector<int>& nxNew) {
  myGeometry=Geometry(*geo);
  nx=nxNew;
  
  offset=nx;
  offset[myGeometry.get_ndims()-1]=1;  
  if (myGeometry.get_ndims() > 1) {
    for (int i=myGeometry.get_ndims()-2; i>=0; i--) {
      offset[i]=offset[i+1]*nx[i+1];
    }
  }
}

Grid::Grid(string filename, Region* parent){
  myGeometry=Geometry(*(parent -> get_geometry()));
  nx=vector<int>(3,20);
  offset=nx;
  offset[myGeometry.get_ndims()-1]=1;  
  if (myGeometry.get_ndims() > 1) {
    for (int i=myGeometry.get_ndims()-2; i>=0; i--) {
      offset[i]=offset[i+1]*nx[i+1];
    }

  }

  ifstream fin(filename.c_str());
  if(fin.is_open())
    {
      
      float tempStr;
      vector<float> data;
      while(fin>>tempStr)
	{
	  data.push_back(tempStr);
	}

      if (add_attribute("dens",float(0))==-1)
	cerr << "failed to add attribute" << endl;
      if (set_attribute("dens",&data[0])==-1)
	cerr << "failed to set attribute" << endl;
      fin.close();
    }
}
//************************************************************
//
// Functions to add/delete Data from a Grid.  
//

// template<class T> int addData(string name,T* newdata) is defined in the header

//******************************************************************************
// Probing data from a grid...


// Returns the index for a particular ordered n-tuplet.
int Grid::calculate_index(vector<int> x0) {
  int idx=0;
  for (int i=0; i < myGeometry.get_ndims(); i++) {
    idx+=x0[i]*offset[i];
  }
  return idx;
}

vector<int> Grid::calculate_coords(int idx) {
  int st=myGeometry.get_spacetype();
  vector<int> x0(get_ndims(),0);

  if (st == REAL) {
    for (int i=0; i < get_ndims(); i++) {
      x0[i]=int(idx/offset[i]);
      idx-=x0[i]*offset[i];
    }
  }
  else if (st == FOURIER_HERMITE) {
    for (int i=0; i < get_ndims(); i++) {
      x0[i]=int(idx/offset[i]);
      idx-=x0[i]*offset[i];
      if ((i < get_ndims()-1)&&(x0[i] > nx[i]/2)) {
	x0[i]=x0[i]-nx[i];
      }
    }
  }
  else if (st == FOURIER_GENERAL) {
    for (int i=0; i < get_ndims(); i++) {
      x0[i]=int(idx/offset[i]);
      idx-=x0[i]*offset[i];
    }
  }
  

  return x0;
}


// Plot and return an appropriate graphics_grid
