#include "basin_base.h"
#include "Data.h"
#include "Region.h"
#include <limits.h>

//************************************************************
//
// Constructor functions

List:: List() {nElements = 0; posKey = "";}

List::List(List* listptr){
  nElements=listptr->get_nelements();
  int n=listptr->get_num_attributes();
  for (int i=0; i < n; i++) {
    string mykey=listptr->get_attribute_key(i);
    cerr << i << ":" << mykey <<"\n";
    Attribute* attrptr=listptr->get_attribute_pointer(mykey);
    Attribute newattr(*attrptr);
    add_attribute(mykey,newattr);
  }
  int ntags=listptr->get_ntags();
  for (int i=1; i <= ntags; i++) {
    string tagname=listptr->get_tagname(i);
    cerr << i << ":" << tagname << "\n";
    string tagvalue=listptr->get_tag(tagname);
    set_tag(tagname,tagvalue);
  }
}

List::List(Region* parent, int nElementsNew) : Data(parent)
{
  nElements = nElementsNew;
  for (int i = 0; i <= nElements; i++) selection.push_back(i);
}

//************************************************************
//
// Functions to add/delete Data from a List.
// User should always call reset_selection() after list modification.
//
// template<class T> int addData(string name, T* newdata) is defined in
// the header

//! Set (and check) nElements from the List attributes.
//! Note that there is no check that all attributes are of the same length.
 int List::set_nelements()
{
  bool nSet = false;
  int nMin = INT_MAX, nMax = -1;
  for (map<string,Attribute>::iterator i = attributes.begin();
       i != attributes.end(); i++) {
    int n = i->second.size();
    if (n < nMin) nMin = n;
    if (n > nMax) nMax = n;
      nSet = true;
  }
 
    // If nSet = false, the map contains no entries.

    if (nSet) {
	if (nMin != nMax) {
	    cerr << "List::set_nelements: attributes have different lengths"
		 << endl;
	    nMax = max(nMin, nMax);
	    
	}
    } else{
      nMax = 0;
      nElements = nMax;	// nMax is consistent with append_attribute_element()  
      return -1;}
    
    nElements = nMax;	// nMax is consistent with append_attribute_element()
    return 0;
}

//******************************************************************************
//
// Filtering and selection of List elements.

// Reset the selection list.
int List::reset_selection()
{
  //cerr<<"IN SELECTION"<<endl;
    selection.clear();
    //cerr<<"IN SELECTION 2"<<endl;
    //    selectionHistory.clear();
    set_nelements();
    //cerr<<"IN SELECTION 3"<<endl;
    int myRank;
    MPI_Comm_rank(BASIN_COMM, &myRank );
    int lo,hi;
    //cerr<<"IN SELECTION 4"<<endl;
    string key=attributes.begin()->first;
    get_distribution_index(key,myRank,&lo,&hi);
    //cerr<<"IN SELECTION 5"<<endl;
    for (int i = 0; i <= nElements; i++) selection.push_back(i);
    return 0;
    //cerr<<"OUT OF SELECTION"<<endl;
}


int List::print_selection(ostream &s)		// default = cout
{
    // Convention: tools never add leading/trailing newlines.
    // s << endl;

    s << "selection: ";
    for (list<int>::iterator i = selection.begin(); i != selection.end(); i++)
	s << *i << " ";
    //s << endl;
    return 0;
}

int List::print_selection_history(ostream &s)	// default = cout
{
    // Convention: tools never add leading/trailing newlines.
    //s << endl;

    s << "selection history:" << endl;
    for (vector<string>::iterator i = selectionHistory.begin();
	 i != selectionHistory.end(); i++) s << *i;	// newline included

    // s << endl;
    return 0;
}

//
// Probing data from a list

// Return the index of the particle nearest a given point.
int List::calculate_idx_nearest(const vector<double>& x0) {
  // DG: I REALLY should use a general metric of some sort.  For now,
  // just assuming cartesian.  I know, it's pretty weenie
  int iBest=0;
  float d2Min=myGeometry.calculate_distance_squared(x0,get_pos(0)); 
  float d2;
  for (int i=1; i < nElements; i++) {
    d2=myGeometry.calculate_distance_squared(x0,get_pos(i));
    if (d2 < d2Min) {
      d2Min=d2;
      iBest=i;
    }
  }
  return iBest;
}


//Will replace much of this code with a call to an external sort
int List::calculate_nth_nearest(int particleDepth,
				const vector<double>& testPos)
{
  int currentDepth=1;
  int nearestParticleIndex = 0;
  float lastDistance=0;
  while(currentDepth<=particleDepth)
    {
      float minDist = FLT_MAX;	  
      float currentDistance=FLT_MAX;

      for(int i=1; i<nElements; i++)
	{
	  //find distance between the two elements (make sure not same particle)
	  currentDistance=
	    myGeometry.calculate_distance_squared(get_pos(i),testPos);
	  //see if it is the closest object, then assign it
	  //also check if it was farther then the one we found on the last pass
	  if ( currentDistance < minDist && currentDistance > lastDistance)
	    {
	      minDist=currentDistance; 
	      nearestParticleIndex = i;
	    }
	}
      lastDistance=minDist;
      currentDepth++;
    }
  return nearestParticleIndex;
}


vector<double> List::get_pos(int idx) {
  vector<double> posOut;
  /*  Vector3 *v;
  if((attributes[posKey].get_genvector_element(idx,v))!=0) {
    cerr << "Could not get element from " << posKey << endl;
    return posOut;
  }
  for (int i=0; i < myGeometry.get_ndims(); i++) {
    posOut.push_back((*v)[i]);
    }*/

  // HERE 
  return posOut;
}

//Ugly temp. workaround for STL sort function... there is a better
//way...?  Again i apologize....
/*
namespace comparison_space
{
  List* compListPtr=NULL;
  string myAttribute = "";
   int fncompare(const int elem1, const int elem2)
  {
      if (compListPtr!=NULL && myAttribute!="") {
      double *test1;
      double *test2;
      if((compListPtr->get_attribute_element(myAttribute, *(int*)elem1,test1 ))!=0){
	cerr << "Warning, could not retrieve element1\n";
	return 0;
      }
      if((compListPtr->get_attribute_element(myAttribute, *(int*)elem2,test2 ))!=0){
	cerr << "Warning, could not retrive element2\n";
	return 0;
      }
      if( *test1 > *test2 )  return 1;
      else if( *test2 > *test1 ) return -1;
      else return 0;
    }
    else
      {
	cerr<<"Warning, compListPtr = NULL or the myAttribute=NULL,"
	    <<"ie there is no list to sort!"<<endl;
	return 0;
      }
}
}
*/




