#include "TEXT_IO.h"

TEXT_IO::TEXT_IO (string fileName, Region* destination) {
  _data = destination;
  _fileName = fileName;
  _isFirstData = true;
  MPI_Comm_rank(BASIN_COMM, &_myRank);
  MPI_Comm_size(BASIN_COMM, &_nProcs);
  
  _equalKey = "=";
  _gridKey = "isGrid";
  _nDimsKey = "nDims";
  _llKey = "lower_left";
  _urKey = "upper_right";
  _nTotKey = "N";
  // _posNameKey = "posName";
  _colNameKey = "colName";
  _colTypeKey = "colType";
  //_colRankKey = "colRank";
  _nColKey = "nColumns";
  _nxKey = "dim_vector";
  // _metricKey = "metric";
  // _boundaryKey = "boundary";
  // _spaceKey = "space";
  _extraVarName = "var";	
  _doubleTypeKey = "double";
  _intTypeKey = "int";
  _stringTypeKey = "str";
  _floatTypeKey = "float";
  _vec2TypeKey = "vector2";
  _vec3TypeKey = "vector3";
  _delimiters = " \t";
  
  _iList = _iGrid = 0;
  

}

TEXT_IO::~TEXT_IO () { }

bool TEXT_IO::OpenFile () {
  _fileStr.open(_fileName.c_str());
  if(_fileStr.is_open()) return true;
  else return false;
}

bool TEXT_IO::CloseFile () {
  _fileStr.close();
  _fileStr.clear();
  if (_fileStr.is_open()) return false;
  return true;
}

void TEXT_IO::SkipLines()
{
  string tempStr;

  while(_fileStr.peek() == '\n' || _fileStr.peek() == ' ' || _fileStr.peek() == '\t') {
    if(_fileStr.peek() == '\n')
      getline(_fileStr, tempStr);
    else if(_fileStr.peek() == ' ' || _fileStr.peek() == '\t')
      _fileStr.ignore();
  }
}

void TEXT_IO::SkipInLine(string &data, int &pos)
{
  int i = pos;
  while(i >= 0 && i < data.size() && (data[i] == ' ' || data[i] == '\t'))
    i++;
  pos = i;
}

void TEXT_IO::RevSkipInLine(string &data, int &pos)
{
  int i = pos;
  while(i >= 0 && i < data.size() && (data[i] == ' ' || data[i] == '\t'))
    i--;
  pos = i;
}

bool TEXT_IO::AssignVariable(string &data, map<const string, string> &header)
{
  int begin_pos = 0;
  int end_pos = 0;
  int equal_pos = 0;
  int key_length = 0;
  int value_length = 0;
  string key;
  string value;
  if(data.size() > 1) {

                           // Find the key in the data string
    
    begin_pos = data.find(';', 0);
    if(begin_pos == -1)
      return false;
    begin_pos++;
    
    equal_pos = data.find(_equalKey, begin_pos);
    if(equal_pos == -1)
      return false;

    key = strtok(&data[begin_pos], " \t");

                           // Find the corresponding data value
    begin_pos = equal_pos + 1;
    end_pos = data[data.size() - 1];
    
    SkipInLine(data, begin_pos);
    RevSkipInLine(data, end_pos);

    //value = strtok(&data[begin_pos], " \t");
    value = data.substr(begin_pos, end_pos-begin_pos+1);
    
                           // Put data into map
    header[key.c_str()] = value;

  }
   
  return true;
    
}

bool TEXT_IO::ReadHeader(map<const string, string> &header)
{
  string tempStr;

  SkipLines();

  while(_fileStr.peek() == ';') {
    getline(_fileStr, tempStr);
    AssignVariable(tempStr, header);
    SkipLines();
  }

  if(header.size() == 0) return false;
  return true;
}

bool TEXT_IO::ReadDataLine(vector<string> &data, int capacity = 0)
{
  int beginPos = 0;
  int endPos = 0;
  string tempStr;
  string dataStr;
  string varData;
  bool endline = false;
  data.resize(0);
  data.reserve(capacity);

  SkipLines();
 
  if(_fileStr.peek() != ';' && _fileStr.peek() != -1)
    getline(_fileStr, tempStr);
  else
    return false;

  char tempStr1[tempStr.size()];

  strcpy(tempStr1, tempStr.c_str());

  char *token = strtok(tempStr1, _delimiters.c_str());

  while(token != NULL) {
    data.push_back(string(token));
    token = strtok(NULL, _delimiters.c_str());
  }

  return true;
}

bool TEXT_IO::IsGrid() {

  int retVal;
  if(_myRank == 0) {
    if(_header.count(_gridKey) == 1) {
      string tempStr = _header[_gridKey.c_str()];
      retVal = strtol(&tempStr[0],NULL,10);
    }
    else
      retVal = 0;
  }
  
  MPI_Bcast(&retVal, 1, MPI_INT, 0, BASIN_COMM);
  
  if(retVal) return true;
  return false;
}

int TEXT_IO::NumDimensions() {
  
  int retVal;
  if(_myRank == 0) {
    if(_header.count(_nDimsKey) == 1) {
      string tempStr = _header[_nDimsKey.c_str()];
      retVal = strtol(&tempStr[0],NULL,10);
    }
    else
      retVal = 0;
  }
  
  MPI_Bcast(&retVal, 1, MPI_INT, 0, BASIN_COMM);
  
  return retVal;
}

unsigned long int TEXT_IO::NumElements(int nCols) {
  int retVal;
  int rowCount = 0;
  if(_myRank == 0) {
    if(_header.count(_nTotKey) == 1) {
      //Number of Elements is Given by Header 
      string tempStr = _header[_nTotKey.c_str()];
      retVal = strtol(&tempStr[0],NULL,10);
      if(retVal < 0) retVal = 0;
    }
    else {
      //Number of Elements is not Given by Header
      streampos beginPos = _fileStr.tellg();
      bool isLine = true;
      vector<string> line;
      while(isLine){
	isLine = ReadDataLine(line, nCols);	
	if(isLine) 
	  ++rowCount;
      }
	  
      retVal = rowCount;
      _fileStr.clear();
      _fileStr.seekg(beginPos);

    }
  }

  MPI_Bcast(&retVal, 1, MPI_INT, 0, BASIN_COMM);

  return retVal;
}

vector<int> TEXT_IO::NumCell() {

  vector<int> nx;
  char *endptr;
  int nDims;
  
  if(_myRank == 0) {
    if(_header.count(_nxKey) == 1) {
      string tempStr = _header[_nxKey.c_str()];
      endptr = &tempStr[0];
      bool empty = false;
      while(int(*endptr) != 0)
	nx.push_back(strtol(endptr, &endptr, 10));
    }
    nDims = nx.size();
  }
   
  MPI_Bcast(&nDims, 1, MPI_INT, 0, BASIN_COMM);

  if(_myRank != 0)
    nx.resize(nDims);
  MPI_Bcast(&nx[0], nDims, MPI_INT, 0, BASIN_COMM);

  MPI_Barrier(BASIN_COMM);

  return nx;  
}

vector<double> TEXT_IO::LowerLeft() {
  vector<double> ll;
  char *endptr;
  int nDims;
  if(_myRank == 0) {
    if(_header.count(_llKey) == 1) {
      string tempStr = _header[_llKey.c_str()];
      endptr = &tempStr[0];
      while(int(*endptr) != 0)
	ll.push_back(strtol(endptr, &endptr, 10));
    }
    else
      ll.resize(0);
  
    nDims = ll.size();
  }
  
  MPI_Bcast(&nDims, 1, MPI_INT, 0, BASIN_COMM);
  if(_myRank != 0)
    ll.resize(nDims);
  MPI_Bcast(&ll[0], nDims, MPI_INT, 0, BASIN_COMM);

  return ll;  
}

vector<double> TEXT_IO::UpperRight() {
  vector<double> ur;
  char *endptr;
  int nDims;
  if(_myRank == 0) {
    if(_header.count(_urKey)) {
      string tempStr = _header[_urKey.c_str()];
      endptr = &tempStr[0];
      while(int(*endptr) != 0)
	ur.push_back(strtol(endptr, &endptr, 10));
    }
    else
      ur.resize(0);
    
    nDims = ur.size();
  }
  

  MPI_Bcast(&nDims, 1, MPI_INT, 0, BASIN_COMM);
  if(_myRank != 0)
    ur.resize(nDims);
  MPI_Bcast(&ur[0], nDims, MPI_INT, 0, BASIN_COMM);

  return ur;  
}

void TEXT_IO::AttInfo (unsigned int nCols, vector<string> &colName, vector<int> &colType, vector<int> &colRank1, vector<int> &colRank2) {

  colName.resize(0);
  colType.resize(0);
  colRank1.resize(0);
  colRank2.resize(0);
  
  bool found = true;

  string tempStr;
  int extraNameCount = 0;
  int replaceNameCount = 0;
  _nInts = _nDoubles = _nStrings = _nVec2ds = _nVec3ds = _nFloats = 0;
  

  if(_myRank == 0) {
    for(int i = 0; i < nCols && colName.size() < nCols; i++) {
      stringstream sstr;
      sstr << i;
      string key = _colNameKey + sstr.str();

      if(_header.count(key)) {
	tempStr = _header[key.c_str()];
	if(tempStr == "")
	  tempStr = _extraVarName + sstr.str();
      }
      else
	tempStr = _extraVarName + sstr.str();

      colName.push_back(tempStr);

      key = _colTypeKey + sstr.str();
      tempStr = "";
      if(_header.count(key))
	 tempStr = _header[key.c_str()];
      else
	tempStr = _doubleTypeKey;
      if(tempStr == "")
	tempStr = _doubleTypeKey;


      if (tempStr == _intTypeKey) {
	// integer
	colType.push_back(0);
	colRank1.push_back(_nInts);
	colRank2.push_back(-1);
	//colName.push_back(tempStr.substr(0,loc-2));
	_nInts++;
      }
      else if (tempStr == _stringTypeKey) {
	// string
	colType.push_back(2);
	colRank1.push_back(_nStrings);
	colRank2.push_back(-1);
	//colName.push_back(tempStr.substr(0,loc-2));
	_nStrings++;
      }
#ifdef VECTOR23
      else if (tempStr == _vec2TypeKey) {
	// vec2d
	colType.push_back(3);
	colType.push_back(3);
	colRank1.push_back(_nVec2ds);
	colRank1.push_back(_nVec2ds);
	//key = colRankKey + sstr.str() + equalKey;
	//keyLoc = header.find(key, keyLoc);
	//if(keyLoc == -1) return -1;
	//int myRank2 = strtol(&header[keyLoc + key.size()], &endptr, 10);
	//int myRank2=atoi(tempStr.substr(loc-3,1).c_str());
	
	//colRank2.push_back(myRank2);
	colRank2.push_back(0);
	colRank2.push_back(1);
	colName.push_back(colName[colName.size()-1]);
	
	//colName.push_back(tempStr.substr(0,loc-4));
	//if (myRank2==0)
	_nVec2ds++;
      }
      else if (tempStr == _vec3TypeKey) {
	// vec3d
	colType.push_back(4);
	colType.push_back(4);
	colType.push_back(4);
	colRank1.push_back(_nVec3ds);
	colRank1.push_back(_nVec3ds);
	colRank1.push_back(_nVec3ds);
	//key = colRankKey + sstr.str() + equalKey;
	//keyLoc = header.find(key, keyLoc);
	//if(keyLoc == -1) return -1;
	//int myRank2 = strtol(&header[keyLoc + key.size()], &endptr, 10);
	//int myRank2=atoi(tempStr.substr(loc-3,1).c_str());
	colRank2.push_back(0);
	colRank2.push_back(1);
	colRank2.push_back(2);
	colName.push_back(colName[colName.size()-1]);
	colName.push_back(colName[colName.size()-1]);
	//colName.push_back(tempStr.substr(0,loc-4));
	//if (myRank2 ==0)
	_nVec3ds++;
      }
#endif
      else if (tempStr == _floatTypeKey){
	colType.push_back(5);
	colRank1.push_back(_nFloats);
	colRank2.push_back(-1);
	//colName.push_back(tempStr.substr(0,loc-2));
	_nFloats++;
      }
      else {
	// double
	colType.push_back(1);
	colRank1.push_back(_nDoubles);
	colRank2.push_back(-1);
	//colName.push_back(tempStr.substr(0,loc-2));
	_nDoubles++;
      }
    }

    for(int count = 0; count < _nInts; count++) {
      _iArray.resize(_nInts);
      _iArray[count] = new vector<int>;
    }
    for(int count = 0; count < _nDoubles; count++) {
      _dArray.resize(_nDoubles);
      _dArray[count] = new vector<double>;
    }
    for(int count = 0; count < _nStrings; count++) {
      _sArray.resize(_nStrings);
      _sArray[count] = new vector<char>;
    }
#ifdef VECTOR23
    for(int count = 0; count < _nVec2ds; count++) {
      _vec2DArray.resize(_nVec2ds);
      _vec2DArray[count] = new vector<Vector2>;
    }    
    for(int count = 0; count < _nVec3ds; count++) {
      _vec3DArray.resize(_nVec3ds);
      _vec3DArray[count] = new vector<Vector3>;
    }
#endif
    for(int count = 0; count < _nFloats; count++) {
      _fArray.resize(_nFloats);
      _fArray[count] = new vector<float>;
    }
  }

  //Bcast Column Names
  vector<int> nameSizes(nCols);
  if(_myRank != 0) {
    colName.resize(nCols);
    colType.resize(nCols);
    colRank1.resize(nCols);
    colRank2.resize(nCols);
  }


  if(_myRank == 0)
    for(int i = 0; i < nCols; i++)
      nameSizes[i] = colName[i].size();

  MPI_Bcast(&nameSizes[0], nCols, MPI_INT, 0, BASIN_COMM);
  if(_myRank != 0) {
    for(int i = 0; i < nCols; i++)
      colName[i].resize(nameSizes[i]);
  }

  for(int i = 0; i < nCols; i++) {
    MPI_Barrier(BASIN_COMM);
    MPI_Bcast(&(colName[i])[0], nameSizes[i], MPI_CHAR, 0, BASIN_COMM);
  }

  //Bcast Column Type Info
  MPI_Bcast(&colType[0], nCols, MPI_INT, 0, BASIN_COMM);
  MPI_Bcast(&colRank1[0], nCols, MPI_INT, 0, BASIN_COMM);
  MPI_Bcast(&colRank2[0], nCols, MPI_INT, 0, BASIN_COMM);
}

void TEXT_IO::FillLine(vector<string>&line, vector<int> &colType, vector<int> &colRank1, 
		       vector<int> &colRank2) {

  int nCols = colType.size();
  
  int itemp;
  double dtemp;
  float ftemp;
  char ctemp;

#ifdef VECTOR23
  Vector3 tmpvec;
  Vector2 tmpvec2;
#endif

  for (int col=0; col < int(colType.size()); col++) {
    if (colType[col] == 0) {
      itemp = strtol(&(line[col])[0], NULL, 10);
      _iArray[colRank1[col]] -> push_back(itemp);
    }
    else if (colType[col] == 1) {
      dtemp = strtod(&(line[col])[0], NULL);
      _dArray[colRank1[col]] -> push_back(dtemp);
    }
    else if (colType[col] == 2) {
      ctemp = line[col][0];
      _sArray[colRank1[col]] -> push_back(ctemp);
    }    
#ifdef VECTOR23
    else if (colType[col] == 3) {
      dtemp = strtod(&(line[col])[0], NULL);
      if (colRank2[col] == 0) 
	tmpvec2[0]=dtemp;
      else if (colRank2[col] == 1){ 
	tmpvec2[1]=dtemp;
	_vec2DArray[colRank1[col-1]] -> push_back(tmpvec2);	
      }
    }
    else if (colType[col] == 4) {
      dtemp = strtod(&(line[col])[0], NULL);
      if (colRank2[col] == 0) 
	tmpvec[0]=dtemp;
      else if (colRank2[col] == 1) 
	tmpvec[1]=dtemp;
      else if (colRank2[col] == 2) {
	tmpvec[2]=dtemp;
	_vec3DArray[colRank1[col-2]] -> push_back(tmpvec);	
      }
    }
#endif
    else if (colType[col] == 5){
      dtemp = static_cast<float>(strtod(&(line[col])[0], NULL));
      _fArray[colRank1[col]] -> push_back(ftemp);
    }
  }
}

int TEXT_IO::ReadData(int isGrid, vector<string> &colName, vector<int> colType, vector<int> &colRank1, vector<int> &colRank2, string index) {

  int retVal = 1;
  vector<string> line;
  int nCols = colName.size();

  if(_hasHeader) {
    //Read the Data
    for(int proc = 0; proc < _nProcs && _myRank == 0; proc++) {
      int rowCount;
      int iLo, iHi;
      if(isGrid)
	_newGrid -> get_distribution_index(colName[0], proc, &iLo, &iHi);
      else
	_newList -> get_distribution_index(colName[0], proc, &iLo, &iHi);
      
      //cerr << "iLo = " << iLo << ", iHi = " << iHi << endl;
      
      bool isLine = true;
      
      
      //Reserve Space in Arrays
      for(int count = 0; count < _nInts; count++) {
	int itemp;
	_iArray[count] -> resize(0);
	_iArray[count] -> reserve(iHi-iLo+1);
      }
      for(int count = 0; count < _nDoubles; count++) {
	double dtemp;
	_dArray[count] -> resize(0);
	_dArray[count] -> reserve(iHi-iLo+1);
      }
      for(int count = 0; count < _nStrings; count++) {
	string ctemp;
	_sArray[count] -> resize(0);
	_sArray[count] -> reserve(iHi-iLo+1);
      }
#ifdef VECTOR23
      for(int count = 0; count < _nVec2ds; count++) {
	Vector2 v2temp;
	_vec2DArray[count] -> resize(0);
	_vec2DArray[count] -> reserve(iHi-iLo+1);
      }
      for(int count = 0; count < _nVec3ds; count++) {
	Vector3 v3temp;
	_vec3DArray[count] -> resize(0);
	_vec3DArray[count] -> reserve(iHi-iLo+1);
      }
#endif
      for(int count = 0; count < _nFloats; count++) {
	float ftemp;
	_fArray[count] -> resize(0);
	_fArray[count] -> reserve(iHi-iLo+1);
      }
      
      
      rowCount = iLo;
      while(rowCount <= iHi && isLine){
	
	isLine = ReadDataLine(line, nCols);

	if(isLine) {
	  FillLine(line, colType, colRank1, colRank2);
	  ++rowCount;
	}
	
      }
      
      if(rowCount-1 != iHi) {
	if(isGrid) {
	  retVal = 0;
	  cerr << "Error: RegionGrid" << index << " contains too few elements." << endl;	
	}      
	
	else
	  cerr << "Error: RegionList" << index << " contained fewer elements than indicated."
	       << endl << "Attributes will be correctly sized but some elements will be inaccurate."
	       << endl;
      }

      
      //Fill in Grid/List
      
      int errorCheck=-1;
      for(int col=0; col< int(colName.size()) && retVal;col++) {	    
	if(isGrid) {
	  if (colType[col] == 0) 
	    errorCheck=_newGrid->set_attribute(colName[col],iLo,iHi,&(*_iArray[colRank1[col]])[0]);
	  else if (colType[col] == 1) 
	    errorCheck=_newGrid->set_attribute(colName[col],iLo,iHi,&(*_dArray[colRank1[col]])[0]);
	  else if (colType[col] == 2) cerr <<"hello, fix needed here"<<endl;
	  //errorCheck=_newGrid->set_attribute(colName[col],iLo,iHi,&(*sArray[colRank1[col]])[0]);
#ifdef VECTOR23
	  else if ((colType[col] == 3) && (colRank2[col] == 0)) 
	    errorCheck=_newGrid->set_attribute(colName[col],iLo,iHi,&(*_vec2DArray[colRank1[col]])[0]);
	  else if ((colType[col] == 4) && (colRank2[col] == 0)) 
	    errorCheck=_newGrid->set_attribute(colName[col],iLo,iHi,&(*_vec3DArray[colRank1[col]])[0]);
#endif
	  else if (colType[col] == 5)
	    errorCheck=_newGrid->set_attribute(colName[col],iLo,iHi, &(*_fArray[colRank1[col]])[0]);
	}
	else {
	  if (colType[col] == 0)
	    errorCheck=_newList->set_attribute(colName[col],iLo,iHi,&(*_iArray[colRank1[col]])[0]);
	  else if (colType[col] == 1) 
	    errorCheck=_newList->set_attribute(colName[col],iLo,iHi,&(*_dArray[colRank1[col]])[0]);
	  else if (colType[col] == 2) cerr <<"hello, fix needed here"<<endl;
	  //	      errorCheck=_newList->set_attribute(colName[col],iLo,iHi,&(*_sArray[colRank1[col]])[0]);
#ifdef VECTOR23
	  else if ((colType[col] == 3) && (colRank2[col] == 0)) 
	    errorCheck=_newList->set_attribute(colName[col],iLo,iHi,&(*_vec2DArray[colRank1[col]])[0]);
	  else if ((colType[col] == 4) && (colRank2[col] == 0)) 
	    errorCheck=_newList->set_attribute(colName[col],iLo,iHi,&(*_vec3DArray[colRank1[col]])[0]);
#endif
	  else if (colType[col] == 5)
	    errorCheck=_newList->set_attribute(colName[col],iLo,iHi, &(*_fArray[colRank1[col]])[0]);
	  //cerr << "errorCheck = " << errorCheck << endl;
	}
	
	
	
	if (errorCheck!=0){
	  for(int count = 0; count < _nInts; count++) delete _iArray[count];
	  for(int count = 0; count < _nDoubles; count++)  delete _dArray[count];
	  for(int count = 0; count < _nStrings; count++) delete _sArray[count];
#ifdef VECTOR23
	  for(int count = 0; count < _nVec2ds; count++) delete _vec2DArray[count];
	  for(int count = 0; count < _nVec3ds; count++) delete _vec3DArray[count];
#endif 
	  for(int count = 0; count < _nFloats; count++) delete _fArray[count];
	  retVal = 0;
	}
	
      }
      
      
    } // eo for(int proc = 0...
  }

  else {
    unsigned long rowCount = 0;
   
    if(_myRank == 0) {
   
      
      bool isLine = true;
      while(isLine){
	isLine = ReadDataLine(line, nCols);	
	if(isLine) {
	  FillLine(line, colType, colRank1, colRank2);
	  ++rowCount;
	}
      }

      //cerr << "size = " << _iArray[0] -> size() << endl;
    }


    MPI_Bcast(&rowCount, 1, MPI_UNSIGNED_LONG, 0, BASIN_COMM);
    
    
    //Make the list
    stringstream sstr;
    sstr << _iList;
    index = sstr.str();
    _newList = _data -> add_list("RegionList" + index, rowCount);
    vector<double> ll;
    ll.push_back(0);
    vector<double> ur;
    ur.push_back(1.0);
    vector<int> temp_nE;
    temp_nE.push_back(rowCount);
    _newList->set_context(ll,ur,temp_nE);
    _iList++;
      
    MPI_Barrier(BASIN_COMM);
    // Now, add all of the attributes.
    int errorCheck = 0;
    for (int i=0; i < int(colName.size()); i++) {
      if (colType[i] == 0) 
	errorCheck=_newList->add_attribute(colName[i],int(0));
      else if (colType[i] == 1) 
	errorCheck=_newList->add_attribute(colName[i],double(0));
      else if (colType[i] == 2)
	errorCheck=_newList->add_attribute(colName[i],char());
#ifdef VECTOR23
      else if ((colType[i] == 3) &&(colRank2[i] == 0)) 
	errorCheck=_newList->add_attribute(colName[i],Vector2());
      else if ((colType[i] == 4) &&(colRank2[i] == 0)) 
	errorCheck=_newList->add_attribute(colName[i],Vector3());
#endif
      else if (colType[i] == 5)
	errorCheck=_newList->add_attribute(colName[i], float(0));
    }
    if (errorCheck!=0) return 0;


    if(_myRank == 0) {

      //Fill in List
      errorCheck=-1;
      for(int col=0; col< int(colName.size()) && retVal;col++) {
	if (colType[col] == 0)
	  errorCheck=_newList->set_attribute(colName[col],0,rowCount-1,&(*_iArray[colRank1[col]])[0]);
	else if (colType[col] == 1) 
	  errorCheck=_newList->set_attribute(colName[col],0,rowCount-1,&(*_dArray[colRank1[col]])[0]);
	else if (colType[col] == 2) cerr <<"hello, fix needed here"<<endl;
	//	      errorCheck=_newList->set_attribute(colName[col],0,rowCount,&(*_sArray[colRank1[col]])[0]);
#ifdef VECTOR23
	else if ((colType[col] == 3) && (colRank2[col] == 0)) 
	  errorCheck=_newList->set_attribute(colName[col],0,rowCount-1,&(*_vec2DArray[colRank1[col]])[0]);
	else if ((colType[col] == 4) && (colRank2[col] == 0)) 
	  errorCheck=_newList->set_attribute(colName[col],0,rowCount-1,&(*_vec3DArray[colRank1[col]])[0]);
#endif
	else if (colType[col] == 5)
	  errorCheck=_newList->set_attribute(colName[col],0,rowCount-1, &(*_fArray[colRank1[col]])[0]);
	//cerr << "errorCheck = " << errorCheck << endl;
	
	if (errorCheck!=0){
	  for(int count = 0; count < _nInts; count++) delete _iArray[count];
	  for(int count = 0; count < _nDoubles; count++)  delete _dArray[count];
	  for(int count = 0; count < _nStrings; count++) delete _sArray[count];
#ifdef VECTOR23
	  for(int count = 0; count < _nVec2ds; count++) delete _vec2DArray[count];
	  for(int count = 0; count < _nVec3ds; count++) delete _vec3DArray[count];
#endif
	  for(int count = 0; count < _nFloats; count++) delete _fArray[count];
	  retVal = 0;
	}
	
      }
    }

  } // eo else after if(_hasHeader)
  
  MPI_Bcast(&retVal, 1, MPI_INT, 0, BASIN_COMM);

  return retVal;

}

bool TEXT_IO::NextDataset () {
  
  int retVal;
  int headerInt;

  _header.clear();

  if(_myRank == 0) {
    ReadHeader(_header);
    if(_isFirstData) {
      if(_header.size() == 0)
	headerInt = 0;
      else
	headerInt = 1;
      _isFirstData = false;
    }
    else {
      if(_header.size() == 0)
	retVal = 0;
      else
	retVal = 1;
    }
  }    
  MPI_Bcast(&retVal, 1, MPI_INT, 0, BASIN_COMM);
  MPI_Bcast(&headerInt, 1, MPI_INT, 0, BASIN_COMM);

  _hasHeader = static_cast<bool>(headerInt);

  if(retVal) return true;
  return false;    
}

bool TEXT_IO::MakeDataObjects () {

  int retVal = 1;
  _newList = NULL;
  _newGrid = NULL;

  vector<double> ll, ur;
  vector<string> line, colName;
  vector<int> numCell, colType, colRank1, colRank2;
  int nCols, nDims;
  string index;
  int isGrid = 0;

  if(_myRank == 0) {
    streampos beginPos = _fileStr.tellg();
    ReadDataLine(line, 0);
    _fileStr.clear();
    _fileStr.seekg(beginPos);
    nCols = line.size();
  }


  unsigned long int nTot;

  MPI_Bcast(&nCols, 1, MPI_INT, 0, BASIN_COMM);

  AttInfo(nCols, colName, colType, colRank1, colRank2);

  if(_hasHeader) {
    //Look for necessary information
    isGrid = IsGrid();
    if(isGrid) {

      numCell = NumCell();
      ll = LowerLeft();
      ur = UpperRight();


      if(ll.size() != numCell.size())
	ll.resize(numCell.size(), 0.0);
      if(ur.size() != numCell.size())
	ur.resize(numCell.size(), 1.0);

      if(numCell.size() == 0)
	isGrid = false;
      else {
	nTot = 1;
	for(int i = 0; i < numCell.size(); i++)
	  nTot *= numCell[i];
      }

      if(nDims > 0 && nDims < numCell.size()) {
	numCell.resize(nDims);
	ll.resize(nDims);
	ur.resize(nDims);
      }
      
    }

    if(!isGrid) {
      nTot = NumElements(nCols);
      ll = LowerLeft();
      ur = UpperRight();
    }

    if(isGrid) {
      // Make the grid
      stringstream sstr;
      sstr << _iGrid;
      index = sstr.str();
      _newGrid = _data -> add_grid("RegionGrid" + index, numCell);
      _newGrid->set_context(ll,ur,numCell);
      _iGrid++;
      

      
      MPI_Barrier(BASIN_COMM);
      // Now, add all of the attributes.
      int errorCheck = 0;
      for (int i=0; i < int(colName.size()); i++) {
	if (colType[i] == 0) 
	  errorCheck=_newGrid->add_attribute(colName[i],int(0));
	else if (colType[i] == 1) 
	  errorCheck=_newGrid->add_attribute(colName[i],double(0));
	else if (colType[i] == 2)
	  errorCheck=_newGrid->add_attribute(colName[i],char());
#ifdef VECTOR23
	else if ((colType[i] == 3) &&(colRank2[i] == 0)) 
	  errorCheck=_newGrid->add_attribute(colName[i],Vector2());
	else if ((colType[i] == 4) &&(colRank2[i] == 0)) 
	  errorCheck=_newGrid->add_attribute(colName[i],Vector3());
#endif
	else if (colType[i] == 5)
	  errorCheck=_newGrid->add_attribute(colName[i], float(0));
      }
      if (errorCheck!=0) return false;

    } // eo if(isGrid)
 
    
    else {
      
      //Make the list
      stringstream sstr;
      sstr << _iList;
      index = sstr.str();
      _newList = _data -> add_list("RegionList" + index, nTot);
      vector<int> temp_nE;
      temp_nE.push_back(nTot);
      _newList->set_context(ll,ur,temp_nE);
      _iList++;
            
      
      MPI_Barrier(BASIN_COMM);
      // Now, add all of the attributes.
      int errorCheck = 0;
      for (int i=0; i < int(colName.size()); i++) {
	if (colType[i] == 0) 
	  errorCheck=_newList->add_attribute(colName[i],int(0));
	else if (colType[i] == 1) 
	  errorCheck=_newList->add_attribute(colName[i],double(0));
	else if (colType[i] == 2)
	  errorCheck=_newList->add_attribute(colName[i],char());
#ifdef VECTOR23
	else if ((colType[i] == 3) &&(colRank2[i] == 0)) 
	  errorCheck=_newList->add_attribute(colName[i],Vector2());
	else if ((colType[i] == 4) &&(colRank2[i] == 0)) 
	  errorCheck=_newList->add_attribute(colName[i],Vector3());
#endif
	else if (colType[i] == 5)
	  errorCheck=_newList->add_attribute(colName[i], float(0));
      }
      if (errorCheck!=0) return false;

    } // eo else after if(isGrid)
   

  } //eo if(_hasHeader)
  
  retVal = ReadData(isGrid, colName, colType, colRank1, colRank2, index);

  MPI_Barrier(BASIN_COMM);

  return retVal;
}

bool TEXT_IO::NextGroup () {
  return false;
}
  
  
bool TEXT_IO::CreateFile() {
  int retVal;
  if(_myRank == 0) {
    _fileStr.open(_fileName.c_str(), fstream::out);
    retVal = !_fileStr.fail();
  }
  MPI_Bcast(&retVal, 1, MPI_INT, 0, BASIN_COMM);

  return retVal;
}

bool TEXT_IO::CloseWrittenFile() {
  if(_myRank == 0) {
    _fileStr.close();
    _fileStr.clear();
  }
  return true;
}

bool TEXT_IO::WriteDatasets() {
  map<string, Grid*>* grids = _data -> get_gridArray ();
  map<string, Grid*>::iterator gridIter;
  for (gridIter = (*grids).begin (); gridIter != (*grids).end (); ++gridIter) {
    WriteData(gridIter->second, true);
  }
  
  map<string, List*>* lists = _data -> get_listArray ();
  map<string, List*>::iterator listIter;
  for (listIter = (*lists).begin (); listIter != (*lists).end (); ++listIter) {
    WriteData(listIter->second, false);
  }
  return true;
}

bool TEXT_IO::WriteData(Data *data, bool isGrid) {
  
  int nAttributes = data -> get_num_attributes();

  vector <string> attName;
  vector <string> attType;
  vector<int> colRank1;

  _nInts = _nDoubles = _nStrings = _nVec2ds = _nVec3ds = _nFloats = 0;

  attName.reserve(nAttributes);
  attType.reserve(nAttributes);
  colRank1.reserve(nAttributes);

  for (unsigned int iAttr = 0; iAttr < nAttributes; ++iAttr) {
    Attribute* tempAttr = data -> get_attribute_pointer (iAttr);
    Attribute_Data_Type type = (*tempAttr).get_data_type ();
    attName.push_back(tempAttr -> get_name());
    
    attName.size();

    if (type == BASIN_DATA_NONE) {
      return false;
    }
    else  if (type == BASIN_DATA_INT) {
      attType.push_back(_intTypeKey);
      colRank1.push_back(_nInts);
      _nInts++;
    } 
    else if (type == BASIN_DATA_FLOAT) {
      attType.push_back(_floatTypeKey);
      colRank1.push_back(_nFloats);
      _nFloats++;
    } 
    else if (type == BASIN_DATA_DOUBLE) {
      attType.push_back(_doubleTypeKey);
      colRank1.push_back(_nDoubles);
      _nDoubles++;
    } 
    else if (type == BASIN_DATA_CHAR) {
      attType.push_back(_stringTypeKey);
      colRank1.push_back(_nStrings);
      _nStrings++;
    } 
    else if (type == BASIN_DATA_VECTOR2) {
      attType.push_back(_vec2TypeKey);
      colRank1.push_back(_nVec2ds);
      _nVec2ds++;
    } 
    else if (type == BASIN_DATA_VECTOR3) {
      attType.push_back(_vec3TypeKey);
      colRank1.push_back(_nVec3ds);
      _nVec3ds++;
    }
    else
      return false;  
  }

  for(int count = 0; count < _nInts; count++) {
    _iArray.resize(_nInts);
    _iArray[count] = new vector<int>;
  }
  for(int count = 0; count < _nDoubles; count++) {
    _dArray.resize(_nDoubles);
    _dArray[count] = new vector<double>;
  }
  for(int count = 0; count < _nStrings; count++) {
    _sArray.resize(_nStrings);
    _sArray[count] = new vector<char>;
  }
#ifdef VECTOR23
  for(int count = 0; count < _nVec2ds; count++) {
    _vec2DArray.resize(_nVec2ds);
    _vec2DArray[count] = new vector<Vector2>;
  }    
  for(int count = 0; count < _nVec3ds; count++) {
    _vec3DArray.resize(_nVec3ds);
    _vec3DArray[count] = new vector<Vector3>;
  }
#endif
  for(int count = 0; count < _nFloats; count++) {
    _fArray.resize(_nFloats);
    _fArray[count] = new vector<float>;
  }

  PrintHeader(data, isGrid, attName, attType);
  PrintAttributes(data, attName, attType, colRank1);
  
  return true;  
  
}

bool TEXT_IO::PrintAttributes(Data* data, vector<string> attName, vector<string> attType, vector<int> colRank1) {

  int nCols = attName.size();

  int iLo, iHi;
 
  for(int proc = 0; proc < _nProcs; proc++) {
    data -> get_distribution_index(attName[0], proc, &iLo, &iHi);
    if(_myRank == 0) {
      for(int col = 0; col < nCols; col++) {
	string type = attType[col];
	Attribute *tempAtt = data -> get_attribute_pointer(attName[col]);
	
	for(int count = 0; count < _nInts; count++)
	  _iArray[count] -> resize(iHi-iLo+1);
	for(int count = 0; count < _nDoubles; count++)
	  _dArray[count] -> resize(iHi-iLo+1);
	for(int count = 0; count < _nStrings; count++)
	  _sArray[count] -> resize(iHi-iLo+1);
#ifdef VECTOR23
	for(int count = 0; count < _nVec2ds; count++)
	  _vec2DArray[count] -> resize(iHi-iLo+1);
	for(int count = 0; count < _nVec3ds; count++)
	  _vec3DArray[count] -> resize(iHi-iLo+1);
#endif
	for(int count = 0; count < _nFloats; count++)
	  _fArray[count] -> resize(iHi-iLo+1);
	
	
	if (type == _intTypeKey)
	  tempAtt -> get(iLo, iHi, &(*_iArray[colRank1[col]])[0]);
	else if (type == _floatTypeKey)
	  tempAtt -> get(iLo, iHi, &(*_fArray[colRank1[col]])[0]);
	else if (type == _doubleTypeKey)
	  tempAtt -> get(iLo, iHi, &(*_dArray[colRank1[col]])[0]);
	else if (type == _stringTypeKey)
	  tempAtt -> get(iLo, iHi, &(*_sArray[colRank1[col]])[0]);
#ifdef VECTOR23
	else if (type == _vec2TypeKey)
	  tempAtt -> get(iLo, iHi, &(*_vec2DArray[colRank1[col]])[0]);
	else if (type == _vec3TypeKey)
	  tempAtt -> get(iLo, iHi, &(*_vec3DArray[colRank1[col]])[0]);
#endif	
      }
      
      
      for(int i = 0; i < iHi-iLo+1; i++) {
	for(int col = 0; col < nCols; col++) {
	  string type = attType[col];
	  if (type == _intTypeKey)
	    _fileStr << _iArray[colRank1[col]] -> at(i);
	  else if (type == _floatTypeKey)
	    _fileStr << _fArray[colRank1[col]] -> at(i);
	  else if (type == _doubleTypeKey)
	    _fileStr << _dArray[colRank1[col]] -> at(i);
	  else if (type == _stringTypeKey)
	    _fileStr << _sArray[colRank1[col]] -> at(i);
#ifdef VECTOR23
	  else if (type == _vec2TypeKey)
	    _fileStr << _vec2DArray[colRank1[col]] -> at(i);
	  else if (type == _vec3TypeKey)
	    _fileStr << _vec3DArray[colRank1[col]] -> at(i);
#endif	  
	  if(col == nCols-1)
	    _fileStr << endl;
	  else
	    _fileStr << " ";
	}
      }

    }
  }

  if(_myRank == 0)
    _fileStr << endl;
  return true;
}


bool TEXT_IO::PrintHeader(Data *data, bool isGrid, vector<string> attName, vector<string> attType) {

  Context *context = data -> get_context_pointer();
  
  vector<int> nx = context -> get_n_grid_points();
  unsigned long nTot;
  int nAttributes = data -> get_num_attributes();
  
  if(nx.size() > 0)
    nTot = nx[0];
  else if(!isGrid)
    return false;

  const vector<double> ll = context -> get_LowerCorner();
  const vector<double> ur = context -> get_UpperCorner();

  if(_myRank == 0) {
    if(isGrid) {
      _fileStr << "; " << _gridKey << " " << _equalKey << " 1" << endl;
      _fileStr << "; " << _nxKey << " " << _equalKey << " ";
      for(int i = 0; i < ur.size(); i++) {
	_fileStr << nx[i] << " ";
      }
      _fileStr << endl;
    }
    else {
      _fileStr << "; " << _gridKey << " " << _equalKey << " 0" << endl;
      _fileStr << "; " << _nTotKey << " " << _equalKey << " " << nTot  << endl;
    }
    if(isGrid || ll.size() > 0 && ur.size() > 0) {
      _fileStr << "; " << _llKey << " " << _equalKey << " ";
      for(int i = 0; i < ll.size(); i++) {
	_fileStr << ll[i] << " ";
      }
      _fileStr << endl;
    
    _fileStr << "; " << _urKey << " " << _equalKey << " ";
    for(int i = 0; i < ur.size(); i++) {
      _fileStr << ur[i] << " ";
    }
    _fileStr << endl;
    }
    for(int i = 0; i < nAttributes; i++) {
      stringstream sstr;
      sstr << i;
      _fileStr << "; " << _colNameKey << sstr.str() << " " << _equalKey << " " << attName[i] << endl;
      _fileStr << "; " << _colTypeKey << sstr.str() << " " << _equalKey << " " << attType[i] << endl;
    }

    _fileStr << endl;
  }
}
