#include "basin.h"
#include "simple_plot.h"
#include "statistics.h"
#include "transforms.h"
#include "cosmology.h"
#include "linearalgebra.h" 
#include "visit_module.h"

using namespace std;


Attribute& fix(const Attribute& attr) {
  int iLo,iHi;
  Attribute dum=attr;
  dum.get_gen_distribution_index(BASIN_ID, &iLo, &iHi);
  double* mydat=(double*)malloc(sizeof(double)*(iHi-iLo+1));
  int* dataout=(int*)malloc(sizeof(int)*(iHi-iLo+1));
  dum.get(iLo,iHi,mydat);
  for (int i=0; i < iHi-iLo+1; i++) {
    dataout[i]=int(mydat[i]);
  }
  int idum=0;
  Attribute* out_ptr=new Attribute(attr.size(),idum);
  out_ptr->put(iLo,iHi,dataout);
  MPI_Barrier(BASIN_COMM);

  free(mydat);
  free(dataout);
  return(*out_ptr);
}

int main( int argc, char *argv[] )
{
  basin_initialize(); 
  // Open the SDSS data 
  Region reg("/home/goldberg/data/DR6_gals_spectra.fit");
  cosmology lcdm;
  List* listptr=reg.get_list_pointer();

  // This is fine.  It's very specific
  Attribute* rlumptr= lcdm.redshift_to_distance(listptr,"z","rlum",3);
  Attribute* rcomptr= lcdm.redshift_to_distance(listptr,"z","rcom");

  // Can we figure out a way to combine these into a single command?
  Attribute* raptr=listptr->get_attribute_pointer("RA");
  Attribute* magptr=listptr->get_attribute_pointer("petroMag_r");

  Attribute absmag=*(magptr)+5.-5.*(log10(abs(*(rlumptr)))+6.);
  int n=absmag.size();
  berr << "original size=" << n << "\n";

  // Does this work?
  double rmax=600.;
  Attribute sel=((absmag > -100)&&(absmag < -20)&&
		 (*raptr > 72)&&(*raptr < 288)&&(*rcomptr<rmax));

  // We _should_ be able to do a selection criterion, and then simply do a 
  // batch job, don't you think?
  Attribute ra1=where(sel,*(raptr));
  Attribute dec1=where(sel,*(listptr->get_attribute_pointer("Dec")));
  Attribute rlum1=where(sel,*(rlumptr));
  Attribute absmag1=where(sel,absmag);
  Attribute rcom1=where(sel,*(rcomptr));

  berr << "Selected size=" << absmag1.size() << "\n";

  // Also, we should be able to do these adds in batch, don't you think?
  List* selptr=reg.add_list("selected",absmag1.size());
  selptr->add_attribute("ra",ra1);
  selptr->add_attribute("dec",dec1);
  selptr->add_attribute("rlum",rlum1);
  selptr->add_attribute("absmag",absmag1);
  selptr->add_attribute("rcom",rcom1);

  // This is fine.
  int nx_box=40;
  vector<int> nx1(3,nx_box);
  equatorial_to_xyz(selptr,"ra","dec","rcom","x","y","z");
  Grid* densities=compute_density(selptr,"x","y","z","rho",nx1);
  Context *dens_context=densities->get_context_pointer();
  vector<double>ll,ur;
  ll=dens_context->get_LowerCorner();
  ur=dens_context->get_UpperCorner();
  double L=ur[0]-ll[0];
  double dL=L/nx_box;

  // Why don't we have a regular "get_attribute", which returns a reference?
  // or perhaps, get_attribute_reference, just to make it clear?
  Attribute* rho= densities->get_attribute_pointer("rho");


  // Compute the window function
  // This is legit, I think.  It's an actual algorithm using a bunch of 
  // native BASIN functions.
  n=rcom1.size();
  int seed,niter=3;
  Attribute dum,id,rcom_mock;
  Attribute xtmp,ytmp,ztmp; 
  Attribute denstmp;
  Attribute window;
  for (int iter=0; iter < niter; iter++) {
    berr << iter << "\n";
    seed=iter;
    id=fix(n*attribute_random(n,seed));
    rcom_mock=resort(rcom1,id);
    equatorial_to_xyz(&ra1,&dec1,&rcom_mock,&xtmp,&ytmp,&ztmp);
    denstmp=compute_density(xtmp,ytmp,ztmp,nx1);    
    if (iter == 0) {
      window=denstmp;
    } else {
      window=window+denstmp;
    }
  }
  window=window/double(niter);

  densities->add_attribute("window",window);
  Attribute diff=(*rho)-window; 
  densities->add_attribute("diff",diff);

  Attribute diff_k=fft(diff); 
  Attribute window_k=fft(window);

  Attribute Pdiff=real(diff_k*conj(diff_k));
  Attribute Pwindow=real(window_k*conj(window_k));
  Pwindow.set_context(window_k.get_context_pointer());
  berr << Pwindow.get_context_pointer()->get_order_type()  << "\n";

  // eq. 21 from Hamilton (2005), which describes the Baumgart & Fry 
  // approach yields:

  // W^2 P(k)_est=[<diff_k* diff_k^\ast>-N(0)]=\int W(k-k')^2 P(k')d^3/(2*pi)^3
  
  Attribute k1=attribute_index(window_k,1);
  Attribute k2=attribute_index(window_k,2);
  Attribute k3=attribute_index(window_k,3);
  Attribute k=sqrt(k1*k1+k2*k2+k3*k3);
  
  berr <<"pt1\n";

  /*
  fourier_ptr->add_attribute("k",k);
  */

  vector<double> kbin,Pbin,Pest;
  for (int i=0; i < nx1[0]/2; i++) {
    kbin.push_back(double(i));
  }
  Pbin=bin_data(k,Pdiff,kbin);
  // Create the convolution kernel

  berr << "pt2\n";

  ConvKernel W(Pwindow,2);
  berr << "About to do sum\n";
  double wtot=sum(Pwindow);
  berr << "total convolution kernel = " << wtot << "\n";
  berr << "function complete!\n";
  for (int i=0; i < nx1[0]/2; i++) {
    Pest.push_back(Pbin[i]/wtot);
  }
  

  // Now, convolve the sample power spectrum with the window...
  vector<double>Pbin_conv;
  for (int iter=0 ; iter < 3; iter++) {
    Attribute Pconv=convolve(Pest, W, Pwindow.get_dims(), FFT_HALF_HERMITIAN);
    berr << "iteration " << iter<< "\n";
    Pbin_conv=bin_data(k,Pconv,kbin);
    //if (BASIN_ID == 0) {plot(kbin,Pbin_conv,Pbin);}
    // Convolve again...
    for (int i=1; i < nx1[0]/2; i++) {
      Pest[i]=Pest[i]*Pbin[i]/Pbin_conv[i];
    }
  }
  // Now, put everything into real units...
  vector<double> kfin,Pfin;
  for (int i=1; i < nx1[0]/2; i++) {
    kfin.push_back(kbin[i]*2.*3.14159/L);
    Pfin.push_back(L*L*L*Pest[i]);		   
  }
  //if (BASIN_ID == 0) {plot(kfin,Pfin,1,3,"Power spectrum");}


  berr << "About to do visit stuff\n";

  visit_module_start();
  visit_expose_discrete(*selptr,"x","y","z","rcom","galaxies");
  visit_module_join();


  basin_finalize();  

  return( 0 ); 
}
  
 
 
 
