#include "basin_package.h"
#include "cosmosims.h"

// Takes a simple gadget code, reads it, computes the power spectrum.
// The i/o will be a template for the later calculations

Region& read_gadget(string infile) {
  Region* reg=new Region();
  int i,j,idum,nbytes,npartall[6],dumarr[10];
  unsigned int npart[6];
  double massarr[6],a,time,fdum,length,omega0,omegal,h;
  float pos,posarr[3];
  double xval=0,z_init;
  int nmax=10000;
  double x[nmax],y[nmax],z[nmax];
  int id[nmax];
  string typearr[6];
  typearr[0]="gas";
  typearr[1]="halo";
  typearr[2]="disk";
  typearr[3]="bulge";
  typearr[4]="stars";
  typearr[5]="bndry";
  int ntypes=6;

  List* listptr;
  FILE* fd;  
  int myid;
  MPI_Comm_rank(BASIN_COMM,&myid);
  cerr << infile << "\n";
  if (myid == 0) {
    fd=fopen(infile.c_str(),"r");
    fread(&nbytes,sizeof(int),1,fd);
    fread(npart,sizeof(unsigned int),6,fd);
    fread(massarr,sizeof(double),6,fd);
    fread(&time,sizeof(double),1,fd);
    fread(&z_init,sizeof(double),1,fd);
    fread(dumarr,sizeof(int),10,fd); // just a silly array
    fread(&length,sizeof(double),1,fd);
    fread(&omega0,sizeof(double),1,fd);
    fread(&omegal,sizeof(double),1,fd);
    fread(&h,sizeof(double),1,fd);
    fseek(fd,260,SEEK_SET);
    cerr << npart[0] << " " << npart[1]<< "\n";
  }
  MPI_Barrier(BASIN_COMM);
  MPI_Bcast(npart,6,MPI_INT,0,BASIN_COMM); // Now everybody knows how many particles we have.  
  MPI_Bcast(&a,1,MPI_DOUBLE,0,BASIN_COMM);
  MPI_Bcast(&length,1,MPI_DOUBLE,0,BASIN_COMM);
  MPI_Bcast(&omega0,1,MPI_DOUBLE,0,BASIN_COMM);
  MPI_Bcast(&omegal,1,MPI_DOUBLE,0,BASIN_COMM);
  MPI_Bcast(&h,1,MPI_DOUBLE,0,BASIN_COMM);
  MPI_Bcast(massarr,6,MPI_DOUBLE,0,BASIN_COMM);
  
  

  for (int itype =0 ; itype < ntypes; itype++) {
    if (npart[itype] > 0) {
      listptr=reg->add_list(typearr[itype],int(npart[itype]));
      listptr->add_attribute("x",xval);
      listptr->add_attribute("y",xval);
      listptr->add_attribute("z",xval);
      listptr->add_attribute("vx",xval);
      listptr->add_attribute("vy",xval);
      listptr->add_attribute("vz",xval);
      listptr->add_attribute("ids",i);

      stringstream ss1,ss2,ss3,ss4,ss5,ss6;
      ss1 << time;
      ss2 << massarr[itype];
      ss3 << length;
      ss4 << omega0;
      ss5 << omegal; 
      ss6 << h;

      listptr->set_tag("a",ss1.str());
      listptr->set_tag("mass",ss2.str());
      listptr->set_tag("BoxSize",ss3.str());
      listptr->set_tag("Omega0",ss4.str());
      listptr->set_tag("OmegaLambda",ss5.str());
      listptr->set_tag("HubbleParam",ss6.str());
    }
  }
  MPI_Barrier(BASIN_COMM);

  // Now, read in the data
  if (myid == 0) {
    // this is a totally ghetto way to read
    
    // First positions
    int ndum;
    fread(&ndum,sizeof(int),1,fd);
    fread(&ndum,sizeof(int),1,fd);
    
    for (int itype=0; itype < ntypes; itype++) {
      if (npart[itype] > 0) {
	listptr=reg->get_list_pointer(typearr[itype]);
	Attribute *xptr=listptr->get_attribute_pointer("x");
	Attribute *yptr=listptr->get_attribute_pointer("y");
	Attribute *zptr=listptr->get_attribute_pointer("z");
	int n=0;
	int istart=0;
	for (int i=0; i < npart[itype]; i++) {
	  fread(posarr,sizeof(float),3,fd);
	  x[n]=posarr[0];
	  y[n]=posarr[1];
	  z[n]=posarr[2];
	  n++;
	  if ((n == nmax)||(i == npart[itype]-1)) {
	    xptr->put(istart,istart+n-1,x);
	    yptr->put(istart,istart+n-1,y);
	    zptr->put(istart,istart+n-1,z);
	    istart=i;		      
	    n=0;
	  }
	}  
      }
    }

    // padding
    fread(&ndum,sizeof(int),1,fd);
    fread(&ndum,sizeof(int),1,fd);

    // Now do the same with velocities...
    for (int itype=0; itype < ntypes; itype++) {
      if (npart[itype] > 0) {
	listptr=reg->get_list_pointer(typearr[itype]);
	Attribute *vxptr=listptr->get_attribute_pointer("vx");
	Attribute *vyptr=listptr->get_attribute_pointer("vy");
	Attribute *vzptr=listptr->get_attribute_pointer("vz");
	int n=0;
	int istart=0;
	for (int i=0; i < npart[itype]; i++) {
	  fread(posarr,sizeof(float),3,fd);
	  x[n]=posarr[0];
	  y[n]=posarr[1];
	  z[n]=posarr[2];
	  n++;
	  if ((n == nmax)||(i == npart[itype]-1)) {
	    vxptr->put(istart,istart+n-1,x);
	    vyptr->put(istart,istart+n-1,y);
	    vzptr->put(istart,istart+n-1,z);
	    istart=i;		      
	    n=0;
	  }
	}  
      }
      
      //padding
      fread(&ndum,sizeof(int),1,fd);
      fread(&ndum,sizeof(int),1,fd);

      // And finally, the particle IDs
      if (npart[itype] > 0) {
	listptr=reg->get_list_pointer(typearr[itype]);
	Attribute *idptr=listptr->get_attribute_pointer("ids");
	int n=0;
	int istart=0;
	for (int i=0; i < npart[itype]; i++) {
	  fread(&j,sizeof(int),1,fd);
	  id[n]=j;
	  n++;
	  if ((n == nmax)||(i == npart[itype]-1)) {
	    idptr->put(istart,istart+n-1,id);
	    istart=i;		      
	    n=0;
	  }
	}  
      }
    }	    
    fclose(fd);
  }
  
  MPI_Barrier(BASIN_COMM);
  
  return *(reg);
}


//**************************************************

int write_gadget(Region* reg,string outfile) {
  string typearr[6];
  typearr[0]="gas";
  typearr[1]="halo";
  typearr[2]="disk";
  typearr[3]="bulge";
  typearr[4]="stars";
  typearr[5]="bndry";
  int ntypes=6;
  int nmax=10000;
  float posarr[nmax*3];
  double x[nmax],y[nmax],z[nmax];
  int id[nmax];
  int npart[6],nparttot=0;
  double massarr[6],a;
  double Omega0,OmegaLambda,HubbleParam,BoxSize;
  int nbytes=256,idum;
  FILE* fd;
  List* listptr;

  // For now, we assume that we are only writing if we are processor 0.
  int myid;
  MPI_Comm_rank(BASIN_COMM,&myid);
  
  struct gadget_header head;
  head.num_files=1;
  for (int i=0; i < 24; i++) head.dumarr[i]=0;
  head.flag_sfr=0;
  head.flag_feedback=0;
  head.flag_cooling=0;
  // Get some info for the header
  for (int i=0; i < ntypes; i++) {
    if (reg->query_list_key(typearr[i]) == 1) {
      listptr=reg->get_list_pointer(typearr[i]);
      npart[i]=head.npart[i]=head.npartTotal[i]=listptr->get_nelements();
      stringstream(listptr->get_tag("mass")) >> head.massarr[i];
      stringstream(listptr->get_tag("a")) >> head.a;
      head.z=1./head.a-1.;
      stringstream(listptr->get_tag("BoxSize")) >> head.BoxSize;
      stringstream(listptr->get_tag("Omega0")) >> head.Omega0;
      stringstream(listptr->get_tag("OmegaLambda")) >> head.OmegaLambda;
      stringstream(listptr->get_tag("HubbleParam")) >> head.HubbleParam;
            
    } else {
      npart[i]=head.npart[i]=head.npartTotal[i]=0;
      head.massarr[i]=0.;
    }
    nparttot+=head.npart[i];
  }

  if (myid==0) {
    fd=fopen(outfile.c_str(),"w");
    fwrite(&nbytes,sizeof(int),1,fd);
    fwrite((const char*)&head,nbytes,1,fd);
    fwrite(&nbytes,sizeof(int),1,fd);


    
    // Now for the data.
    int nbytebin=nparttot*sizeof(float)*3;
    fwrite(&nbytebin,sizeof(int),1,fd);

    for (int itype=0; itype < ntypes; itype++) {
      if (npart[itype] > 0) {
	listptr=reg->get_list_pointer(typearr[itype]);
	Attribute *xptr=listptr->get_attribute_pointer("x");
	Attribute *yptr=listptr->get_attribute_pointer("y");
	Attribute *zptr=listptr->get_attribute_pointer("z");
	int n=0;
	for (int i=0; i < npart[itype]; i++) {
	  if ((n==nmax)||(i==npart[itype]-1)) {
	    int i_curr=0;
	    int lo,hi;
	    while (i_curr < npart[itype]) {
	      lo=i_curr;
	      hi=min(i_curr+nmax-1,npart[itype]-1);
	      //cerr << lo <<  " " << hi << "\n";
	      xptr->get(lo,hi,x);
	      yptr->get(lo,hi,y);
	      zptr->get(lo,hi,z);
	      for (int j=0; j < hi-lo+1; j++) {
		posarr[0+3*j]=x[j];
		posarr[1+3*j]=y[j];
		posarr[2+3*j]=z[j];
	      }
	      fwrite(posarr,sizeof(float),3*(hi-lo+1),fd);
	      i_curr=hi+1;
	    } 
	  }
	}
      }
    }

    // Now read/write velocities
    // padding
    fwrite(&nbytebin,sizeof(int),1,fd);
    fwrite(&nbytebin,sizeof(int),1,fd);

    for (int itype=0; itype < ntypes; itype++) {
      if (npart[itype] > 0) {
	listptr=reg->get_list_pointer(typearr[itype]);
	Attribute *xptr=listptr->get_attribute_pointer("vx");
	Attribute *yptr=listptr->get_attribute_pointer("vy");
	Attribute *zptr=listptr->get_attribute_pointer("vz");
	int n=0;
	for (int i=0; i < npart[itype]; i++) {
	  if ((n==nmax)||(i==npart[itype]-1)) {
	    int i_curr=0;
	    int lo,hi;
	    while (i_curr < npart[itype]) {
	      lo=i_curr;
	      hi=min(i_curr+nmax-1,npart[itype]-1);
	      xptr->get(lo,hi,x);
	      yptr->get(lo,hi,y);
	      zptr->get(lo,hi,z);
	      for (int j=0; j < hi-lo+1; j++) {
		posarr[0+3*j]=x[j];
		posarr[1+3*j]=y[j];
		posarr[2+3*j]=z[j];
	      }
	      fwrite(posarr,sizeof(float),3*(hi-lo+1),fd);
	      i_curr=hi+1;
	    } 
	  }
	}
      }
    }

    // Now read/write indices.
    
    fwrite(&nbytebin,sizeof(int),1,fd);
    nbytebin=sizeof(int)*nparttot;
    fwrite(&nbytebin,sizeof(int),1,fd);

    for (int itype=0; itype < ntypes; itype++) {
      if (npart[itype] > 0) {
	listptr=reg->get_list_pointer(typearr[itype]);
	Attribute *idptr=listptr->get_attribute_pointer("ids");
	int n=0;
	for (int i=0; i < npart[itype]; i++) {
	  if ((n==nmax)||(i==npart[itype]-1)) {
	    int i_curr=0;
	    int lo,hi;
	    while (i_curr < npart[itype]) {
	      lo=i_curr;
	      hi=min(i_curr+nmax-1,npart[itype]-1);
	      idptr->get(lo,hi,id);
	      fwrite(id,sizeof(int),(hi-lo+1),fd);
	      i_curr=hi+1;
	    } 
	  }
	}
      }
    } 
    fwrite(&nbytebin,sizeof(int),1,fd);

    fclose(fd);
  }
  

  return 0;
}
