#include "basin_package.h"
#include "cosmology.h"
#include "cosmosims.h"
#include "transforms.h"

// For now, we're assuming CDM only
Region& initial_conditions(cosmology cosmo,int nx,double L,double z_init,
			   unsigned int seed) {
  // L is in h^1 Mpc
  Region* outreg=new Region();
  int npts=nx*nx*nx;
  int lo,hi;
  List* cdm_ptr=outreg->add_list("halo",npts);
  vector<int>nxgrid(3,nx);
  Attribute* px=new Attribute(nxgrid,double(0.));
  fftw_complex comp_dum;
  vector<int> nx_k=px->get_dims();
  nx_k[2]=nx_k[2]/2+1;
  Attribute* px_k=new Attribute(nx_k,comp_dum);
  *px_k=fft(*px);

  Context* r=px->get_context_pointer();
  Context* c=px_k->get_context_pointer();

  cerr << "Context order " << c->get_order_type() << "\n";

  double k0=2.*3.14159/L;
  
  Attribute* k=new Attribute(nx_k,0.);
  *k=sqrt(pow(k0*attribute_index(*px_k,1),2)+
		   pow(k0*attribute_index(*px_k,2),2)+
		   pow(k0*attribute_index(*px_k,3),2)+1.e-6);
  // Do I need any factors of L,nx, etc.?
  fftw_complex comp_i;
  comp_i.re=0.;
  comp_i.im=1.;
  
  // Generate Gaussian-random conditions.
  double a=1./(1.+z_init);
  double norm=pow(L,-1.5);
  cerr << z_init << " " << cosmo.hubble(a) << " " << cosmo.hubble(1.) << "\n";
  double vfact=100.*cosmo.hubble(a)/cosmo.hubble(1.)*sqrt(a);

  Attribute *x1=new Attribute(nx_k,0.);
  Attribute *x2=new Attribute(nx_k,0.);
  Attribute *delta_k=new Attribute(nx_k,comp_i);

  *x1=attribute_random(nx_k,seed);
  *x2=attribute_random(nx_k,seed+1);
  *delta_k=sqrt(-2*log((*x1)))*cos(2.*3.14159*(*x2))
      +comp_i*(sqrt(-2.*log((*x1)))*sin(2.*3.14159*(*x2)));
  delete x1;
  delete x2;
  cerr << "Phase computed\n";
  *delta_k=norm*sqrt(cosmo.power_spectrum(*k,z_init)/2)*(*delta_k);

  cerr << "Delta_k computed\n";

  Attribute* tmp=new Attribute(nxgrid,0.);
  cerr << "tmp allocated\n";
  if (tmp == NULL) {
    cerr << "But it is null!\n";
  } 

  // X-direction
  *px_k=comp_i*(*delta_k)*(k0*attribute_index(*px_k,1)/((*k)*(*k)));
  *px=vfact*fft(*px_k);
  *tmp=(attribute_index((*px),1)+0.5)*L/nx+(*px)/vfact;
  tmp->get_gen_distribution_index(BASIN_ID,&lo,&hi);  
  double* xdat=(double*)malloc(sizeof(double)*(hi-lo+1));
  double* vdat=(double*)malloc(sizeof(double)*(hi-lo+1));
  tmp->get(lo,hi,xdat);
  px->get(lo,hi,vdat);
  Attribute* x=new Attribute(npts,0.);
  Attribute* vx=new Attribute(npts,0.);
  x->put(lo,hi,xdat);
  vx->put(lo,hi,vdat);
  cdm_ptr->add_attribute("x",*x);
  cdm_ptr->add_attribute("vx",*vx);

  // Y-direction
  *px_k=comp_i*(*delta_k)*(k0*attribute_index(*px_k,2)/((*k)*(*k)));
  *px=vfact*fft(*px_k);
  *tmp=(attribute_index(*px,2)+0.5)*L/nx+(*px)/vfact;
  tmp->get(lo,hi,xdat);
  px->get(lo,hi,vdat);
  Attribute* y=new Attribute(npts,0.);
  Attribute* vy=new Attribute(npts,0.);
  y->put(lo,hi,xdat);
  vy->put(lo,hi,vdat);
  cdm_ptr->add_attribute("y",*y);
  cdm_ptr->add_attribute("vy",*vy);

  // Z-direction
  *px_k=comp_i*(*delta_k)*(k0*attribute_index(*px_k,3)/((*k)*(*k)));
  delete delta_k;
  *px=vfact*fft(*px_k);
  *tmp=(attribute_index(*px,3)+0.5)*L/nx+(*px)/vfact;
  tmp->get(lo,hi,xdat);
  // tmp is no longer needed
  delete tmp;
  px->get(lo,hi,vdat);
  Attribute* z=new Attribute(npts,0.);
  Attribute* vz=new Attribute(npts,0.);
  z->put(lo,hi,xdat);
  vz->put(lo,hi,vdat);
  cdm_ptr->add_attribute("z",*z);
  cdm_ptr->add_attribute("vz",*vz);

  // Clean up the temporary memory;

  delete k;
  delete px;
  delete px_k;

  Attribute idx=attribute_index(npts);
  cdm_ptr->add_attribute("ids",idx);

  // We also need to add a few things to the main list:
  // For now, we only have one list, which means that we need only add to
  // that list.
  cerr << "About to write header\n";

  stringstream ss1,ss2,ss3,ss4,ss5,ss6;
  ss1 << a;
  cdm_ptr->set_tag("a",ss1.str());
  
  // In h^-1 M_odot
  double mass=cosmo.get_omega_m()*pow(L,3)/npts*2.76e11;
  ss2 << mass;
  cdm_ptr->set_tag("mass",ss2.str());

  ss3 << L;
  cdm_ptr->set_tag("BoxSize",ss3.str());

  ss4 << cosmo.get_omega_m();
  cdm_ptr->set_tag("Omega0",ss4.str());

  ss5 << cosmo.get_omega_l();
  cdm_ptr->set_tag("OmegaLambda",ss5.str());

  ss6 << cosmo.get_h();
  cdm_ptr->set_tag("HubbleParam",ss6.str());
  free(xdat);
  free(vdat);

  cerr << "all done!\n";

  return *outreg;
 
}
