#include "basin.h"

using namespace std;

int main( int argc, char *argv[] )
{

    int  myid, numprocs;
    /*    

    MPI_Init( &argc, &argv );
    MPI_Comm_size( BASIN_COMM, &numprocs );
    MPI_Comm_rank( BASIN_COMM, &myid );

    */

    // The three lines above can be replaced by

    basin_initialize(myid,numprocs);

    /* 
       see ba_kernel for different ways to use ba_initialize
       e.g. you could use it this way
       vector<int> np_id(2);
       np_id=basin_initialize();
       myid=np_id[1];
       numprocs=np_id[0];
       
       np_id[0] would contain the number of processes
       np_id[1] would contain the process id
    */
   
    MPI_Barrier( BASIN_COMM);



    /* READ FILES, GET THE DATA AND DISTRIBUTE THEM AMONG ALL THE PROCESSES*/

    Region Reg1("../data/cluster1.dat");
    Region Reg2("../data/cluster2.dat");
   

    /* Get the list out of the two regions above */

    List * l1 = Reg1.get_list_pointer("RegionList0");
    List * l2 = Reg2.get_list_pointer("RegionList0");



    /* A few attributes to play with soon.
       At this point the attributes do not know what they are going to be.
       they might become doubles, float, Vector3....who knows?
    */


    Attribute a1,a2,a3;
    Attribute b1,b2,b3;
    Attribute x1,x2,x3;
    Attribute y1,y2,y3,y4,y5,y6;





    /* A few variables to test some operations */
    double d1=4.0;
    float f1=3.0;
    int i1=2;


    /* Finally let's get some attribute out of Lists and play with them*/

    /* in l1 mass is a double*/
    a1=l1->copy_attribute("mass");
    a2=l1->copy_attribute("pos");

    /* 
       a1 is a distribute double attribute now
       a2 is a distribute Vector3 attribute now
    */

    /* in l2 mass is a float */

    b1=l2->copy_attribute("mass");
    b2=l2->copy_attribute("pos");
    /* 
       b1 is a distribute double attribute now
       b2 is a distribute Vector3 attribute now
    */



    /* ...and now let's do some some operations on these Attributes */

    x1=1+a1-(i1-pow(a1,2.0))+exp(a1/10)-sqrt(a1+f1);
    x2=b1+a1;
    x3=a2*(a2+1.0); /* this is the scalar product of two Vector3 */
    cout<<"ALL1="<<sum(a1<2 || (a1>4 && x3<1))<<endl;
    cout<<"ALL1B="<<sum(a1<2)<<endl;
    cout<<"ALL1C="<<sum(a1>4 && x3<1)<<endl;
    Attribute tw=where(a1<2 || a1 >90,a1+2.0,a1+3.0);
    cout<<a1[0]<<"====="<<tw[0]<<endl;
    int tt=1;
    Attribute tw2=where(a1<67,tt,a1);
    cout<<a1[0]<<"HHH="<<tw2[0]<<endl;
    cout<<a1[10]<<"HHH="<<tw2[10]<<endl;
    /* 
       All operations involving mixed datatypes and many other things
       were not allowed (or they were and they were to lead to a run-time 
       error, e.g. sqrt(string))  in the previous version of the BASIN_KERNEL 
       and implementing them would have required a lot of non-clean coding.
      
    */

    /* 

    /*    Now let's check whether the values are correct and in doing so 
          use the [] operator */
    if (myid==0)
      {
	cout<<"E:If you are using the original demo files,the first element of x1 should be" <<endl; 
	cout<<"5523.13...and hopefully it is..." <<x1[0]<<endl;
	cout<<"============================="<<endl;
	cout<<"E:If you are using the original demo files,the second element of x2 should be "<<endl;
	cout<<"153.68...and hopefully it is..." <<x2[1]<<endl;
	cout<<"============================="<<endl;
	cout<<"E:If you are using the original demo files,the third element of x3 should be "<<endl;
	cout<<"0.233812...and hopefully it is..." <<x3[2]<<endl;
	cout<<"============================="<<endl;
      }


    if (myid==0) cout<<"If you try to call a function not defined for a given type (in this case I will try to call sin(Vector3))"<<endl;
    if (myid==0) cout<<"you get..."<<endl;

    y1=sin(a2);
	cout<<"============================="<<endl;


    /* 
       This MPI_BARRIER HERE SHOULD JUST TO MAKE message printing less messy
    */
    MPI_Barrier(BASIN_COMM);

    /* 
       Now I want to know the range of values of an Attribute owned 
       by each process. Note that the name of the function used to do this 
       is to be changed. For n-Dimensional attributes the function 
       to call is a different one. 
    */
    
    int jl,jh;
    x1.get_gen_distribution_index(myid,&jl,&jh);
    cout<<"C:I am process "<<myid<<" and I own elements of attribute x1 from "<<jl<<" to "<<jh<<endl; 
    /* 
       This MPI_BARRIER HERE SHOULD JUST TO MAKE message printing less messy
    */
    MPI_Barrier(BASIN_COMM);
    if (myid==0)
      {
	cout<<"SU: And why is this useful?"<<endl;
	cout<<"E: Are you kidding me?!! Because access to local elements is faster"<<endl; 
	cout<<"..and if you can you should work on the elements you own!"<<endl;
	cout<<"***Note that each operation or math operation defined for an Attribute works this way"<<endl;
	cout<<"every time this is possible***"<<endl;
	cout<<"The calculation is done in parallel and each process is working" <<endl;
	cout<<"on its own elements"<<endl;
	cout<<"SU: Now if I want to change the third and the fouth element of an Attribute."<<endl;
	cout<<"How do I do that?"<<endl;
	cout<<"E: Why would you want to do that?"<<endl;
	cout<<"SU: OK I understand...it's not possible, is it?"<<endl;
	cout<<"E: YES IT IS POSSIBLE!!!!"<<endl;
	cout<<"LOOK:"<<endl;
	  }
    /* 
       This MPI_BARRIER IS HERE JUST TO MAKE message printing less messy
    */
    MPI_Barrier(BASIN_COMM);



    vector<double> a;
    a.push_back(99.99999);
    a.push_back(101.001);
    if (myid==0) cout<<"E:a is an stl vector a[0]="<<a[0]<<" a[1]"<<a[1]<<endl;
    if (myid==0) cout<<"E:BEFORE CALLING a1.put(3,4,a) "<<a1[3]<<" "<<a1[4]<<endl;
    a1.put(3,4,a);
    if (myid==0) cout<<"E:AFTER CALLING a1.put(3,4,a) "<<a1[3]<<" "<<a1[4]<<endl;
    l1->add_attribute("newAtt",x1);


    /* 
       This MPI_BARRIER IS HERE JUST TO MAKE message printing less messy
    */
    MPI_Barrier(BASIN_COMM);

    
    if (myid==0) cout<<"E: And now let's Sort Attribute a1 and put "<<endl;
    if (myid==0) cout<<"the sorted Attribute in y2: y2=sort(a1)"<<endl;
    y2=sort(a1);
    if (myid==0) cout<<"the min is of course in y2[0]="<<y2[0]<<endl;
    if (myid==0) cout<<"You can get the length of the Attribute"<<endl;
    if (myid==0) cout<<"and easily calculate the percentiles"<<endl;
    vector<int> ll;
    ll=y2.get_dims();
    if (myid==0) cout<<"min="<<y2[0]<<endl;
    int ll1=int(ll[0]*0.1);
    if (myid==0) cout<<"10percent="<<y2[ll1]<<endl;
    ll1=int(ll[0]*0.5);
    if (myid==0) cout<<"50percent="<<y2[ll1]<<endl;
    ll1=int(ll[0]*0.9);
    if (myid==0) cout<<"90percent="<<y2[ll1]<<endl;
    if (myid==0) cout<<"max="<<y2[ll[0]]<<endl;
    if (myid==0) cout<<"==================="<<endl;
    if (myid==0) cout<<"E: This means that the user can do very easily"<<endl;
    if (myid==0) cout<<"what List_statistics does and, even better, get the statistics he needs not only the one predefined in List_statistics!"<<endl;
    if (myid==0) cout<<"===================================="<<endl;
    if (myid==0) cout<<"===================================="<<endl;
    if (myid==0) cout<<"SU: OK, OK"<<endl;
    if (myid==0) cout<<"SU: Now another question for you:"<<endl;
    if (myid==0) cout<<"SU: You sorted one attribute, say mass, and...."<<endl;
    if(myid==0)  cout<<"I want to know the value of another attribute corresponding to a given mass percentile"<<endl;
    if (myid==0) cout<<"SU: For example, what is the position of the object with 90percentile mass?"<<endl; 
    if (myid==0) cout<<"E: Do you mean that you want to know to have the original index of the sorted attribute?"<<endl; 
    if (myid==0) cout<<"SU: Yes, sir"<<endl; 
    if (myid==0) cout<<"E: You can do that and as for your specific question"<<endl; 

  
  
    y3=index_sort(a1);
    int index=y3[ll1];
    if (myid==0) cout<<"The position is "<<a2[y3[ll1]]<<endl;

    if (myid==0) cout<<"===================================="<<endl;
    if (myid==0) cout<<"SU: I am sure you don't have any collective operation. Do you?"<<endl;
    if (myid==0) cout<<"E: Indeed at this point I have only one collective operation: SUM"<<endl;  
    if (myid==0) cout<<"E: sum(Attribute) will calculate the sum of all elements of a 1D Attribute (you can use this to calculate the mean for example)"<<endl;

    float xsum=sum(b1);
    if (myid==0) cout<<"E:For example the sum of masses in our file is "<<xsum<<endl; 
    bout<<"====================="<<endl;
    bout<<"Now it is easier to print a statement just from proc 0"<<endl;
    bout<<"Just use bout<<"<<endl;
    berr<<"or berr<<"<<endl;


    if (myid==0) cout<<"====================="<<endl;
    if (myid==0) cout<<"Complex numbers and operations on complex numbers are also available"<<endl;
    fftw_complex zc;
    zc.re=1.0;
    zc.im=2.0;
    if (myid==0) cout<<"For example I have defined zc="<<zc<<endl;
    if (myid==0) cout<<"I can have Attribute holding complex numbers"<<endl;
    if (myid==0) cout<<"Here I define a new Attribute  equal to zc+a1*zc/10"<<endl;
    y4=zc+a1*zc/10.;
    y5=imag(y4);
    if (myid==0) cout<<" and I print one element of a1 and the corresponding element of y4"<<endl;
    
    if (myid==0) cout<<"      "<<endl;
    if (myid==0) cout<<"a1[2]="<<a1[2]<<" y4[2]="<<y4[2]<<" imag(y4)[2]="<<y5[2]<<endl;
    y5=real(a1);
    if (myid==0) cout<<"a1[2]="<<a1[2]<<" and of course real(a1)[2]="<<y5[2]<<endl;
    y5=imag(a1);
    if (myid==0) cout<<"a1[2]="<<a1[2]<<" and of course imag(a1)[2]="<<y5[2]<<endl;
    y6=attribute_random(20);
    if (myid==0) cout<<"EV:We even have a random number function"<<endl;
    if (myid==0) cout<<"for example if you type y6=attribute_random(20) you get 1d Attribute with 20 numbers uniformly distributed between 0 and 1"<<endl;
    if (myid==0)
      {
	for (int kr=0;kr<20;kr++) cout<<"y6["<<kr<<"]="<<y6[kr]<<endl;
      }
    if (myid==0) cout<<"or, for a multi-dimensional Attribute y6=attribute_random(nd) where nd is a vector<int> with the dimensions of your matrix"<<endl;
    vector<int> hj(2);
    hj[0]=5;
    hj[1]=5;
    y6=attribute_random(hj);
    if (myid==0)
      {
	for (int kr1=0;kr1<5;kr1++)
	  { 
	    	for (int kr2=0;kr2<5;kr2++)
		  {
		    cout<<"y6["<<kr1<<","<<kr2<<"]="<<y6(kr1,kr2)<<endl;
		  }
	  }
      }
    unsigned int seed=124;
    unsigned int seed1=125;
    y6=attribute_random(hj,seed);
    if (myid==0)
      {
	for (int kr1=0;kr1<5;kr1++)
	  { 
	    	for (int kr2=0;kr2<5;kr2++)
		  {
		    //		    cout<<"y6["<<kr1<<","<<kr2<<"]="<<y6(kr1,kr2)<<endl;
		  }
	  }
      }
    //    MPI_Finalize();
    /* For consistency with initialize MPI_finalize is replaced by */
    cerr<<a1.get_data_type()<<endl;
    cerr<<y6.get_data_type()<<endl;
    Attribute hjl=Attribute(10,double(31.0));
    cerr<<(Attribute(10,double(0))).get_data_type()<<endl;
    vector<bv1> tempv;
    tempv=hjl.get_data_elements(0,2);
    cout<<tempv[0]<<" "<<tempv[1]<<endl;
    Attribute ax;
    ax=attribute_random(1000,123);
    Attribute ay(1000,1);
    Attribute az;
    Attribute aw;
    cout<<"HHHHH1-"<<ay.get_context_pointer()->get_UpperCorner()[0]<<endl;
    ay=ay+ax;
    cout<<"HHHHH2-"<<ay.get_context_pointer()->get_UpperCorner()[0]<<endl;
    cout<<"HHHHH3-"<<az.get_context_pointer()->get_UpperCorner()[0]<<endl;
    az=ax+ay;
    cout<<"HHHHH4-"<<az.get_context_pointer()->get_UpperCorner()[0]<<endl;
    cout<<"HHHHH5-"<<aw.get_context_pointer()->get_UpperCorner()[0]<<endl;
    aw=ax*2;
    cout<<"HHHHH6-"<<aw.get_context_pointer()->get_UpperCorner()[0]<<endl;
    basin_finalize();
    return( 0 );
}
