
/*
 *  C++ core and functions for BASIN.
 *  Copyright (C) 2006  BASIN team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* 
   File va.cc

   Global Arrays for the BASIN project

      Author:Enrico Vesperini 
      Department of Physics
      Drexel University
      Date:February 19, 2005
      Revised:****

For an N-dim matrix A[i_0][i_1][i_2]....[i_N-1]
loi[k] and hi[k] give the range of values for the index i_k
vec_dim[k] is the number of elements for the index i_k

A N-dimensional matrix is distributed by i_0 (each process gets
i_0/number_processes "rows")  

For 2d 
0 index for rows
1 index for columns

*/
#ifndef BA
#define BA

#include "ba_trait.h"
//#include "promote-old.h"
template <class T, class U>
class Conversion
{
    typedef char Small;
    class Big { char dummy[2]; };
    static Small Test(U);
    static Big Test(...);
    static T MakeT();
public:
    enum { exists = 
        (sizeof(Test(MakeT())) == sizeof(Small)) };
};

template <typename T>
ba<T>::ba(){
  vector<int> tmp_d(1);
  tmp_d[0]=0;
   ba_internal_create(tmp_d,BASIN_COMM,0);
   //  cerr<<"default constructor"<<win<<endl;
};

template <typename T>
  MPI_Win ba<T>::ba_win(){return win;}


template <typename T>
  int ba<T>::ba_internal_create(vector<int> vec_dim,MPI_Comm c,int reserv=0)
  {
    comm = c;
    MPI_Comm_rank(comm,&rank);
    int kd;
    vector<int>::iterator pos;
    mpi_type_trait<T> mpt;
    last_process_reserve=reserv;
    data_T=mpt.get();

    MPI_Comm_size(comm,&n_proc);
    MPI_Type_size(data_T,&size_T);
    dim_v=0;
    dimensions=vec_dim;
    number_of_dimensions=vec_dim.size();
    kd=vec_dim[0]%n_proc;
    last_index_put=0;
    map_u=new int[n_proc];
	loc_dim_v=vec_dim[0]/n_proc;
	map_u[rank]=0;
    /* this is to take into account that a reserve will be allocated for the
       last process; so map_u[n_proc-1] is actually larger than needed
       based on the initial number of elements in the BA */
    //    if (rank==n_proc-1) map_u[rank]+=last_process_reserve;
    /* information about data mapping is broadcast to all processes*/
    for (int kr=0;kr<n_proc;kr++)
      {
	//       MPI_Bcast(&map_u[kr],1,MPI_INT,kr,comm);
	map_u[kr]=0;
      }
    /* Allocate memory*/
      MPI_Alloc_mem(1*size_T,((MPI_Info) 0),&g);
      MPI_Win_create(g,1,size_T,((MPI_Info) 0),comm,&win);
      //      cerr<<"in internal_create"<<win<<endl;
    return -1;
  }



template <typename T>
  int ba<T>::ba_create(vector<int> vec_dim,MPI_Comm c,int reserv=0)
  {
    int td;
    if(dim_v==0) 
      {
	ba_free();
      }
    comm = c;
    MPI_Comm_rank(comm,&rank);
    int kd;
    vector<int>::iterator pos;
    mpi_type_trait<T> mpt;
    last_process_reserve=reserv;
    data_T=mpt.get();

    MPI_Comm_size(comm,&n_proc);
    MPI_Type_size(data_T,&size_T);
    dim_v=1;
    for (pos=vec_dim.begin();pos<vec_dim.end();++pos)
    dim_v=dim_v*(*pos);
    dimensions=vec_dim;
    number_of_dimensions=vec_dim.size();
    kd=vec_dim[0]%n_proc;
    last_index_put=0;
    map_u=new int[n_proc];
    if(rank<=(kd-1))
      {
	loc_dim_v=vec_dim[0]/n_proc+1;
      }
    else
      {
	loc_dim_v=vec_dim[0]/n_proc;
      }
    for (int kp=0;kp<n_proc;kp++)
      {
	if(kp<=(kd-1))
	  {
	    td=vec_dim[0]/n_proc+1;
	    map_u[kp]=(kp+1)*td-1;
	  }
	else
	  {
	    td=vec_dim[0]/n_proc;
	    map_u[kp]=kd*td+(kd-1)+(kp-kd+1)*td;
	  }
      }
    /* this is to take into account that a reserve will be allocated for the
       last process; so map_u[n_proc-1] is actually larger than needed
       based on the initial number of elements in the BA */
    //    if (rank==n_proc-1) map_u[rank]+=last_process_reserve;
    map_u[n_proc-1]+=last_process_reserve;
    /* information about data mapping is broadcast to all processes*/
    /*    for (int kr=0;kr<n_proc;kr++)
	  {
	  MPI_Bcast(&map_u[kr],1,MPI_INT,kr,comm);
	  }*/
    /* Allocate memory*/
    if (rank==n_proc-1)
      {/* allocate reserve memory for the last process
	  so to be able to append data with push_back function
       */
	MPI_Alloc_mem(((loc_dim_v)*(dim_v/dimensions[0])+(last_process_reserve)*(dim_v)/dimensions[0])*size_T,((MPI_Info) 0),&g);
	MPI_Win_create(g,(loc_dim_v)*(dim_v/dimensions[0])+(last_process_reserve)*(dim_v)/dimensions[0],size_T,((MPI_Info) 0),comm,&win);
	//  cerr<<"in ba_createA"<<win<<endl;
      }
    else
      {
	/* give all the other processes only the memory they need*/
	
	MPI_Alloc_mem(((loc_dim_v)*(dim_v/dimensions[0]))*size_T,((MPI_Info) 0),&g);
	MPI_Win_create(g,(loc_dim_v)*(dim_v/dimensions[0]),size_T,((MPI_Info) 0),comm,&win);
      }
    //  cerr<<"in ba_createF"<<win<<endl;
    
    return -1;
  }

template <typename T>
  ba<T>::ba(const ba<T> & a)
{
  if (a.dimensions[0]!=0)
    {
      //      cerr<<"calling ba_create from the copy contsructor "<<win<<" "<<dim_v<<" "<<a.dimensions[0]<<endl;
      this->ba_internal_create(a.dimensions,a.comm,0);
      this->ba_create(a.dimensions,a.comm,0);
      for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
	{
	  T temp=a.g[i];
	  this->g[i]=temp;
	}
    }
  else
    {
      this->ba_internal_create(a.dimensions,a.comm,0);
    }
  //  cerr<<"copy constructor"<<win<<endl;

};



 template <typename T> 
  ba<T>::ba(const T a)
{

  //  this->ba_create(a.dimensions,a.comm,0);
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      T temp=a;
      this->g[i]=temp;
    }
};

  
template <typename T>
void ba<T>::ba_free()
{
  if (&win !=NULL) MPI_Win_free(&win);
  if (g!=NULL) MPI_Free_mem(g);
  g=NULL;
  delete []map_u;
}
template <typename T>
ba<T>::~ba(void){
  this->ba_free();
}

template <typename T>
void ba<T>::ba_distribution(int proc_rank,int *ilo, int *ihi) const
{
  if (proc_rank==0) 
    {
      *ilo=0;
    }
  else
    {

      *ilo=map_u[proc_rank-1]+1;


    }
  *ihi=map_u[proc_rank];

}



template <typename T>
vector<int> ba<T>::ba_distribution(int proc_rank) const
{
  vector<int> limits(2);
  this->ba_distribution(proc_rank,&limits[0],&limits[1]);
  return limits;
}


template <typename T>
void ba<T>::ba_distribution_index(int proc_rank,int *ilo, int *ihi) const
{

  if (proc_rank==0) 
    {
      *ilo=0;
    }
  else
    {

      *ilo=(map_u[proc_rank-1]+1)*(dim_v/dimensions[0]);


    }
  *ihi=(map_u[proc_rank]+1)*(dim_v/dimensions[0])-1;

}

template <typename T>
vector<int> ba<T>::ba_distribution_index(int proc_rank) const
{
  vector<int> limits(2);
  this->ba_distribution_index(proc_rank,&limits[0],&limits[1]);
  return limits;
}


/* This function is to append data in the last processor beyond the initial
   size of the BA but within the allocated reserve.
   This means that 
   target rank=n_proc-1
*/
template <typename T>
int ba<T>::ba_push_back(int number_elems, T *ptr)
{
  int target_rank;
  int ih=0;
  int temp_lo,temp_hi;
  int number_of_elems=number_elems;
  int low[1],upp[1];
  target_rank=n_proc-1;
  this->ba_distribution(target_rank,&low[0],&upp[0]);
  temp_lo=dim_v+last_index_put;
  temp_hi=dim_v+last_index_put+number_of_elems-1;
  target_rank=n_proc-1;
  if ((map_u[n_proc-1]*dim_v)/dimensions[0]<temp_hi) 
    {
      cerr<<"not allowed to use ba_push_back; not enough reserve memory"<<endl;
      return -1;
    }

  MPI_Win_lock( MPI_LOCK_EXCLUSIVE, target_rank, MPI_MODE_NOCHECK, 
		win );
      
  //==================================================
  //       ||     ||  ||        ||     ||  ||  ||   ||
  //   values put                     displ   
  
  
  MPI_Put(&ptr[ih],number_of_elems,data_T,target_rank,temp_lo-(low[0]*dim_v)/dimensions[0],number_of_elems,data_T,win);
  MPI_Win_unlock( target_rank, win ); 
  // update last_index_put
  last_index_put+=number_of_elems;

      return 0;
}
/*
template <typename T>
  int ba<T>::ba_put(const vector<int> loi,const vector<int> hi,T *ptr)
  {
    int target_rank,id_rank;
    int init_target_rank,final_target_rank;
    int ih=0;
    double time1,time2;
    int low[n_proc],upp[n_proc];
    int temp_lo,temp_hi;

    for (int h=0;h<n_proc;h++)
      {
	this->ba_distribution(h,&low[h],&upp[h]);
	if (loi[0]>=low[h]) 
	  { 
	    init_target_rank=h;
	  }
	if (hi[0]==loi[0]) 
	  {
	    final_target_rank=init_target_rank;
	  }
	else if (hi[0]>=low[h]) 
	  {
	    final_target_rank=h;
	  }
      }
    for (int h=init_target_rank;h<final_target_rank+1;h++)
      {
	    target_rank=h;
	    if (h==init_target_rank) 
	      {
		temp_lo=loi[0];
		if (hi[0]>upp[h])
		  {
		    temp_hi=upp[h];
		  }
		else
		  {
		    temp_hi=hi[0];
		  }
	      }
	    else if (h==final_target_rank) 
	      {
		temp_lo=low[h];
		temp_hi=hi[0];
	      }
	    else 
	      {
		temp_lo=low[h];
		temp_hi=upp[h];
	      }

	    MPI_Win_lock(MPI_LOCK_EXCLUSIVE, target_rank, MPI_MODE_NOCHECK, 
	           win);
	//==================================================
	//       ||     ||  ||        ||     ||  ||  ||   ||
	//   values get                     displ   
    MPI_Put(&ptr[ih],(temp_hi-temp_lo+1)*(dim_v/dimensions[0]),data_T,target_rank,(temp_lo-low[h])*(dim_v/dimensions[0]),(temp_hi-temp_lo+1)*(dim_v/dimensions[0]),data_T,win);
    ih=ih+(temp_hi-temp_lo+1)*(dim_v/dimensions[0]);
    
    MPI_Win_unlock( target_rank, win ); 
      }
    return -1;
    }*/
template <typename T> 
int ba<T>::ba_get(const vector<int> loi,const vector<int> hi,T* ptr) const
{
  int target_rank;
  int init_target_rank=0,final_target_rank=0;
  MPI_Aint extentd;
  int local_dim_v;
  int ih=0;
  int low[n_proc],upp[n_proc];
  int temp_lo,temp_hi;
  int nd_minus_one;

  int displ,R_dim;
  //Error check
  //===========
  if(loi.size() != hi.size())
    {
      cerr<<"ERROR in ba_get:the size of lower and upper limit vectors differ"<<endl;
      return -1;
    }
  if(int(loi.size()) != number_of_dimensions || int(hi.size())!=number_of_dimensions)
    {
      cerr<<"ERROR in ba_get:size of lower and/or of upper limit vector differ from the number of dimensions of the array"<<endl; 
      return -1;
    }
  if(loi.size() == hi.size() && loi>hi)
    {
      cerr<<"ERROR in ba_get:lower larger than upper limit "<<endl;
      return -1;
    }
  //===========


  if (number_of_dimensions>1) 
    {
      nd_minus_one=number_of_dimensions-1;
    }
  else
    {
      nd_minus_one=1;
    }
  vector<int> number_data(nd_minus_one);
  vector<int> stride_data(nd_minus_one);
  int ka;
  
  MPI_Datatype nonc_da_ty3;
  mpi_type_trait< non_c_d_md<T> > mpt3;
  
  local_dim_v=1;
  for (int jl=1;jl<number_of_dimensions;jl++)
    local_dim_v*=(hi[jl]-loi[jl]+1);
  
  for (int h=0;h<n_proc;h++)
    {
      this->ba_distribution(h,&low[h],&upp[h]);
      if (loi[0]>=low[h]) 
	{ 
	  init_target_rank=h;
	}
      if (hi[0]==loi[0]) 
	{
	  final_target_rank=init_target_rank;
	}
      else if (hi[0]>=low[h]) 
	{
	  final_target_rank=h;
	}
    }
  /* Here we assume a N-dim matrix A[i0,i1,i2....,i(N-1)]
   NOTE THAT 
   number_data[0], stride_data[0] are those associated to index i(N-1)
   and so on.*/
  if (number_of_dimensions>1)
    {
      ka=number_of_dimensions-1;
      number_data[0]=hi[ka]-loi[ka]+1;
      stride_data[0]=dimensions[ka];
      for (int kx=1;kx<number_of_dimensions-1;kx++)
	{
	  ka-=1;
	  number_data[kx]=hi[ka]-loi[ka]+1;
	  stride_data[kx]=dimensions[ka];
	}
    }
  else
    {
      number_data[0]=1;
      stride_data[0]=1;
    }
  nonc_da_ty3=mpt3.get(number_data,stride_data);
  MPI_Type_extent(nonc_da_ty3,&extentd);
  for (int h=init_target_rank;h<final_target_rank+1;h++)
    {
      target_rank=h;
      if (h==init_target_rank) 
	{
	  temp_lo=loi[0];
	  if (hi[0]>upp[h])
	    {
	      temp_hi=upp[h];
	    }
	  else
	    {
	      temp_hi=hi[0];
	    }
	}
      else if (h==final_target_rank) 
	{
	  temp_lo=low[h];
	  temp_hi=hi[0];
	}
      else 
	{
	  temp_lo=low[h];
	  temp_hi=upp[h];
	}
      displ=(temp_lo-low[h])*dim_v/dimensions[0];
      R_dim=dim_v/dimensions[0];
      for (int ki=1;ki<number_of_dimensions;ki++)
	{
	  displ+=loi[ki]*R_dim/dimensions[ki];
	  R_dim=R_dim/dimensions[ki];
	}

      MPI_Win_lock(MPI_LOCK_EXCLUSIVE, target_rank, MPI_MODE_NOCHECK, 
		   win);
     
      MPI_Get(&(ptr[ih]),(temp_hi-temp_lo+1)*(local_dim_v),data_T,target_rank,displ,(temp_hi-temp_lo+1),nonc_da_ty3,win);
            
      MPI_Win_unlock( target_rank, win ); 
      ih=ih+(temp_hi-temp_lo+1)*(local_dim_v);
    }
  MPI_Type_free(&nonc_da_ty3);
  return 0;
}

//******************************************************************************************
//! Experimental -- ba_get using the indices of the array rather than the pairs describing the 
//! corners.

//*********************
template <typename T>
int ba<T>::ba_get(const int loi,const int hii,T* ptr) const
{
  int target_rank;
  int init_target_rank=0,final_target_rank=0;
  MPI_Aint extentd;
  int local_dim_v;
  int ih=0;
  int temp_lo,temp_hi;
  
  int lo_target;


  //Error check
  //===========
  if(loi > hii)
    {
      cerr<<"ERROR in ba_put:lower limit larger than upper limit "<<endl;
      return -1;
    }
  //===========
  
  vector<int> number_data(1);
  vector<int> stride_data(1);
  
  MPI_Datatype nonc_da_ty2;
  mpi_type_trait< non_c_d_md<T> > mpt2;
  
  local_dim_v=1;
  
  init_target_rank=0;
  final_target_rank=0;
  for (int h=1;h<n_proc;h++)
    {
      if (loi >= (map_u[h-1]+1)*(dim_v/dimensions[0])) 
	{ 
	  init_target_rank=h;
	}
      if (hii >=(map_u[h-1]+1)*(dim_v/dimensions[0])) 
	{
	  final_target_rank=h;
	}
    }
  
  number_data[0]=1;
  stride_data[0]=1;
  
  nonc_da_ty2=mpt2.get(number_data,stride_data);
  
  for (int h=init_target_rank;h<final_target_rank+1;h++)
    {
      target_rank=h;
      if (h==init_target_rank) 
	{
	  temp_lo=loi;
	}
      else {
	temp_lo=(map_u[h-1]+1)*(dim_v/dimensions[0]);
      }
      if (h==final_target_rank) 
	{
	  temp_hi=hii;
	}
      else {
	temp_hi=(map_u[h]+1)*(dim_v/dimensions[0])-1;
      }
      
      MPI_Win_lock(MPI_LOCK_EXCLUSIVE, target_rank, MPI_MODE_NOCHECK, 
		   win);
      //==================================================
      //       ||     ||  ||        ||     ||  ||  ||   ||
      //   values get                     displ   
      MPI_Type_extent(nonc_da_ty2,&extentd);
      if (h == 0) {
	lo_target=0;
      }
      else {
	lo_target=(map_u[h-1]+1)*(dim_v/dimensions[0]);
      }

      //      MPI_Get(&ptr[ih],(temp_hi-temp_lo+1),data_T,target_rank,(temp_lo-lo_target),
      //	      (temp_hi-temp_lo+1),nonc_da_ty2,win);
            MPI_Get(&ptr[ih],(temp_hi-temp_lo+1),data_T,target_rank,(temp_lo-lo_target),(temp_hi-temp_lo+1),data_T,win);
      MPI_Win_unlock( target_rank, win ); 
      ih=ih+temp_hi-temp_lo+1;
    }

    MPI_Type_free(&nonc_da_ty2);
    return 0;
}



//======================
template <typename T>
int ba<T>::ba_get_opt(const int index,T* ptr) const
{
  int target_rank=0;
  int ih=0;
  int temp_lo;
  int lo_target;
  if (index >(dim_v-1)) 
    {
      cerr<<"ERROR:index larger than the size of the BA.No value returned"<<endl; 
      return -1;
    }

  for (int h=1;h<n_proc;h++)
    {
      if (index >= (map_u[h-1]+1)*(dim_v/dimensions[0])) 
	{ 
	  target_rank=h;
	}
    }
  
  
  temp_lo=index;

  MPI_Win_lock(MPI_LOCK_EXCLUSIVE, target_rank, MPI_MODE_NOCHECK, 
	       win);
  lo_target=0;
  if (target_rank !=0) 
    {
      lo_target=(map_u[target_rank-1]+1)*(dim_v/dimensions[0]);
    }
  
  MPI_Get(&ptr[ih],1,data_T,target_rank,(temp_lo-lo_target),1,data_T,win);
  MPI_Win_unlock( target_rank, win ); 
  return 0;
}




template <typename T> 
int ba<T>::ba_get_opt(const vector<int> index,T* ptr) const
{
  int target_rank=0;
  int local_dim_v;
  int ih=0;
  int low[n_proc],upp[n_proc];
  int temp_lo,temp_hi;
  int nd_minus_one;

  int displ,R_dim;
  //Error check
  //===========
  if(int(index.size()) != number_of_dimensions)
    {
      cerr<<"ERROR in ba_get:size of lower and/or of upper limit vector differ from the number of dimensions of the array"<<endl; 
      return -1;
    }
  //===========


  if (number_of_dimensions>1) 
    {
      nd_minus_one=number_of_dimensions-1;
    }
  else
    {
      nd_minus_one=1;
    }
  vector<int> number_data(nd_minus_one,1);
  //  vector<int> stride_data(nd_minus_one);
  
  
  local_dim_v=1;
  //  for (int jl=1;jl<number_of_dimensions;jl++)
  //local_dim_v*=(hi[jl]-loi[jl]+1);

  for (int h=0;h<n_proc;h++)
    {
      this->ba_distribution(h,&low[h],&upp[h]);
      if (index[0]>=low[h]) 
	{ 
	  target_rank=h;
	}
    }

  /* Here we assume a N-dim matrix A[i0,i1,i2....,i(N-1)]
   NOTE THAT 
   number_data[0], stride_data[0] are those associated to index i(N-1)
   and so on.*/


  /*  
      int ka;
      if (number_of_dimensions>1)
    {
      ka=number_of_dimensions-1;
      //      number_data[0]=1;
      stride_data[0]=dimensions[ka];
      for (int kx=1;kx<number_of_dimensions-1;kx++)
	{
	  ka-=1;
	  //  number_data[kx]=1;
	  stride_data[kx]=dimensions[ka];
	}
    }
  else
    {
      //      number_data[0]=1;
      stride_data[0]=1;
      }*/
  //  temp_lo=index[0];
  //temp_hi=index[0];
  
  displ=(index[0]-low[target_rank])*dim_v/dimensions[0];
  R_dim=dim_v/dimensions[0];
  for (int ki=1;ki<number_of_dimensions;ki++)
    {
      displ+=index[ki]*R_dim/dimensions[ki];
      R_dim=R_dim/dimensions[ki];
    }

  MPI_Win_lock(MPI_LOCK_EXCLUSIVE, target_rank, MPI_MODE_NOCHECK, 
	       win);
     
  //  MPI_Get(&(ptr[ih]),(temp_hi-temp_lo+1)*(local_dim_v),data_T,target_rank,displ,(temp_hi-temp_lo+1),data_T,win);
  MPI_Get(&(ptr[ih]),1,data_T,target_rank,displ,1,data_T,win);
  MPI_Win_unlock( target_rank, win ); 
  return 0;
}









//====================

template <typename T>
int ba<T>::ba_put(const vector<int> loi,const vector<int> hi,T *ptr){
  int target_rank;
  int init_target_rank=0,final_target_rank=0;
  MPI_Aint	extentd;
  int local_dim_v;
  int ih=0;
  int low[n_proc],upp[n_proc];
  int temp_lo,temp_hi;
  int nd_minus_one;

  int displ,R_dim;
  //Error check
  //===========
  if(loi.size() != hi.size())
    {
      cerr<<"ERROR in ba_put:the size of lower and upper limit vectors differ"<<endl;
      return -1;
    }
  if(int(loi.size()) != number_of_dimensions || int(hi.size())!=number_of_dimensions)
    {
      cerr<<"ERROR in ba_put:size of lower and/or of upper limit vector differ from the number of dimensions of the array"<<endl; 
      return -1;
    }
  if(loi.size() == hi.size() && loi>hi)
    {
      cerr<<"ERROR in ba_put:lower limit larger than upper limit "<<endl;
      return -1;
    }
  //===========

  if (number_of_dimensions>1) 
    {
      nd_minus_one=number_of_dimensions-1;
    }
  else
    {
      nd_minus_one=1;
    }
 vector<int> number_data(nd_minus_one);
 vector<int> stride_data(nd_minus_one);
 int ka;
 
 MPI_Datatype nonc_da_ty2;
 mpi_type_trait< non_c_d_md<T> > mpt2;
  
 local_dim_v=1;
 for (int jl=1;jl<number_of_dimensions;jl++)
   local_dim_v*=(hi[jl]-loi[jl]+1);
 
 for (int h=0;h<n_proc;h++)
   {
     this->ba_distribution(h,&low[h],&upp[h]);
     if (loi[0]>=low[h]) 
       { 
	  init_target_rank=h;
	}
      if (hi[0]==loi[0]) 
	{
	  final_target_rank=init_target_rank;
	}
      else if (hi[0]>=low[h]) 
	{
	  final_target_rank=h;
	}
      
    }
 
 if (number_of_dimensions>1)
    {
      ka=number_of_dimensions-1;
      number_data[0]=hi[ka]-loi[ka]+1;
      stride_data[0]=dimensions[ka];
      for (int kx=1;kx<number_of_dimensions-1;kx++)
	{
	  ka-=1;
	  number_data[kx]=hi[ka]-loi[ka]+1;
	  stride_data[kx]=dimensions[ka];
	}
    }
  else
    {
      number_data[0]=1;
      stride_data[0]=1;
}
  nonc_da_ty2=mpt2.get(number_data,stride_data);
  for (int h=init_target_rank;h<final_target_rank+1;h++)
    {
      target_rank=h;
      if (h==init_target_rank) 
	{
	  temp_lo=loi[0];
	  if (hi[0]>upp[h])
	    {
	      temp_hi=upp[h];
	    }
	  else
	    {
	      temp_hi=hi[0];
	    }
	}
      else if (h==final_target_rank) 
	{
	  temp_lo=low[h];
	  temp_hi=hi[0];
	}
      else 
	{
	  temp_lo=low[h];
	  temp_hi=upp[h];
	}
      displ=(temp_lo-low[h])*dim_v/dimensions[0];
      R_dim=dim_v/dimensions[0];
      for (int ki=1;ki<number_of_dimensions;ki++)
	{
	  displ+=loi[ki]*R_dim/dimensions[ki];
	  R_dim=R_dim/dimensions[ki];
	}

      MPI_Win_lock(MPI_LOCK_EXCLUSIVE, target_rank, MPI_MODE_NOCHECK, 
		   win);
      //==================================================
      //       ||     ||  ||        ||     ||  ||  ||   ||
      //   values get                     displ   
      MPI_Type_extent(nonc_da_ty2,&extentd);

      MPI_Put(&ptr[ih],(temp_hi-temp_lo+1)*(local_dim_v),data_T,target_rank,displ,(temp_hi-temp_lo+1),nonc_da_ty2,win);
      
      MPI_Win_unlock( target_rank, win ); 
      ih=ih+(temp_hi-temp_lo+1)*(local_dim_v);
    }
  
  MPI_Type_free(&nonc_da_ty2);
  return 0;
}
template <typename T> 
int ba<T>::ba_putF(const vector<int> loi,const vector<int> hi,T *ptr)
{

  //  ba_put(loi,hi,ptr,(bool)Conversion<T,U>::exists);
  ba_put(loi,hi,ptr);
  return 0;
}


//******************************************************************************************
//! Experimental -- ba_put using the indices of the array rather than the pairs describing the 
//! corners.

//*********************
template <typename T>
int ba<T>::ba_put(const int loi,const int hii,T *ptr)
{
  int target_rank;
  int init_target_rank=0,final_target_rank=0;
  MPI_Aint	extentd;
  int local_dim_v;
  int ih=0;
  int temp_lo,temp_hi;

  int lo_target;
  
  //Error check
  //===========
  if(loi > hii)
    {
      cerr<<"ERROR in ba_put:lower limit larger than upper limit "<<endl;
      return -1;
    }
  //===========
  vector<int> number_data(1);
  vector<int> stride_data(1);
  
  MPI_Datatype nonc_da_ty2;
  mpi_type_trait< non_c_d_md<T> > mpt2;
  
  local_dim_v=1;
  
  init_target_rank=0;
  final_target_rank=0;
  for (int h=1;h<n_proc;h++)
    {
      if (loi >= (map_u[h-1]+1)*(dim_v/dimensions[0]))
	{ 
	  init_target_rank=h;
	}
      if (hii >=(map_u[h-1]+1)*(dim_v/dimensions[0]))
	{
	  final_target_rank=h;
	}
    }
  number_data[0]=1;
  stride_data[0]=1;
  nonc_da_ty2=mpt2.get(number_data,stride_data);
  for (int h=init_target_rank;h<final_target_rank+1;h++)
    {
      target_rank=h;
      if (h==init_target_rank) 
	{
	  temp_lo=loi;
	}
      else {
	temp_lo=(map_u[h-1]+1)*(dim_v/dimensions[0]);
      }
      if (h==final_target_rank) 
	{
	  temp_hi=hii;
	}
      else {
	temp_hi=(map_u[h]+1)*(dim_v/dimensions[0])-1;
      }
      MPI_Win_lock(MPI_LOCK_EXCLUSIVE, target_rank, MPI_MODE_NOCHECK, 
		   win);
      //==================================================
      //       ||     ||  ||        ||     ||  ||  ||   ||
      //   values get                     displ   
      MPI_Type_extent(nonc_da_ty2,&extentd);
      if (h == 0) {
	lo_target=0;
      }
      else {
	lo_target=(map_u[h-1]+1)*(dim_v/dimensions[0]);
      }
      //      MPI_Put(&ptr[ih],(temp_hi-temp_lo+1),data_T,target_rank,(temp_lo-lo_target),
      //      (temp_hi-temp_lo+1),nonc_da_ty2,win);
            MPI_Put(&ptr[ih],(temp_hi-temp_lo+1),data_T,target_rank,(temp_lo-lo_target),
		    (temp_hi-temp_lo+1),data_T,win);

      MPI_Win_unlock( target_rank, win ); 
      ih=ih+temp_hi-temp_lo+1;
    }

  MPI_Type_free(&nonc_da_ty2);
  return 0;
}
//========================NEW
template <typename T>
int ba<T>::ba_add(const vector<int> loi,const vector<int> hi,T *ptr){
  int target_rank;
  int init_target_rank=0,final_target_rank=0;
  MPI_Aint	extentd;
  int local_dim_v;
  int ih=0;
  int low[n_proc],upp[n_proc];
  int temp_lo,temp_hi;
  int nd_minus_one;

  int displ,R_dim;
  //Error check
  //===========
  if(loi.size() != hi.size())
    {
      cerr<<"ERROR in ba_put:the size of lower and upper limit vectors differ"<<endl;
      return -1;
    }
  if(int(loi.size()) != number_of_dimensions || int(hi.size())!=number_of_dimensions)
    {
      cerr<<"ERROR in ba_put:size of lower and/or of upper limit vector differ from the number of dimensions of the array"<<endl; 
      return -1;
    }
  if(loi.size() == hi.size() && loi>hi)
    {
      cerr<<"ERROR in ba_put:lower limit larger than upper limit "<<endl;
      return -1;
    }
  //===========

  if (number_of_dimensions>1) 
    {
      nd_minus_one=number_of_dimensions-1;
    }
  else
    {
      nd_minus_one=1;
    }
 vector<int> number_data(nd_minus_one);
 vector<int> stride_data(nd_minus_one);
 int ka;
 
 MPI_Datatype nonc_da_ty2;
 mpi_type_trait< non_c_d_md<T> > mpt2;
  
 local_dim_v=1;
 for (int jl=1;jl<number_of_dimensions;jl++)
   local_dim_v*=(hi[jl]-loi[jl]+1);
 
 for (int h=0;h<n_proc;h++)
   {
     this->ba_distribution(h,&low[h],&upp[h]);
     if (loi[0]>=low[h]) 
       { 
	  init_target_rank=h;
	}
      if (hi[0]==loi[0]) 
	{
	  final_target_rank=init_target_rank;
	}
      else if (hi[0]>=low[h]) 
	{
	  final_target_rank=h;
	}
      
    }
 
 if (number_of_dimensions>1)
    {
      ka=number_of_dimensions-1;
      number_data[0]=hi[ka]-loi[ka]+1;
      stride_data[0]=dimensions[ka];
      for (int kx=1;kx<number_of_dimensions-1;kx++)
	{
	  ka-=1;
	  number_data[kx]=hi[ka]-loi[ka]+1;
	  stride_data[kx]=dimensions[ka];
	}
    }
  else
    {
      number_data[0]=1;
      stride_data[0]=1;
}
  nonc_da_ty2=mpt2.get(number_data,stride_data);
  for (int h=init_target_rank;h<final_target_rank+1;h++)
    {
      target_rank=h;
      if (h==init_target_rank) 
	{
	  temp_lo=loi[0];
	  if (hi[0]>upp[h])
	    {
	      temp_hi=upp[h];
	    }
	  else
	    {
	      temp_hi=hi[0];
	    }
	}
      else if (h==final_target_rank) 
	{
	  temp_lo=low[h];
	  temp_hi=hi[0];
	}
      else 
	{
	  temp_lo=low[h];
	  temp_hi=upp[h];
	}
      displ=(temp_lo-low[h])*dim_v/dimensions[0];
      R_dim=dim_v/dimensions[0];
      for (int ki=1;ki<number_of_dimensions;ki++)
	{
	  displ+=loi[ki]*R_dim/dimensions[ki];
	  R_dim=R_dim/dimensions[ki];
	}

      MPI_Win_lock(MPI_LOCK_EXCLUSIVE, target_rank, MPI_MODE_NOCHECK, 
		   win);
      //==================================================
      //       ||     ||  ||        ||     ||  ||  ||   ||
      //   values get                     displ   
      MPI_Type_extent(nonc_da_ty2,&extentd);

      MPI_Accumulate(&ptr[ih],(temp_hi-temp_lo+1)*(local_dim_v),data_T,target_rank,displ,(temp_hi-temp_lo+1),nonc_da_ty2,MPI_SUM,win);
      
      MPI_Win_unlock( target_rank, win ); 
      ih=ih+(temp_hi-temp_lo+1)*(local_dim_v);
    }
  
  MPI_Type_free(&nonc_da_ty2);
  return 0;
}

//******************************************************************************************
//! Experimental -- ba_put using the indices of the array rather than the pairs describing the 
//! corners.

//*********************
template <typename T>
int ba<T>::ba_add(const int loi,const int hii,T *ptr)
{
  int target_rank;
  int init_target_rank=0,final_target_rank=0;
  MPI_Aint	extentd;
  int local_dim_v;
  int ih=0;
  int temp_lo,temp_hi;

  int lo_target;
  
  //Error check
  //===========
  if(loi > hii)
    {
      cerr<<"ERROR in ba_put:lower limit larger than upper limit "<<endl;
      return -1;
    }
  //===========
  vector<int> number_data(1);
  vector<int> stride_data(1);
  
  MPI_Datatype nonc_da_ty2;
  mpi_type_trait< non_c_d_md<T> > mpt2;
  
  local_dim_v=1;
  
  init_target_rank=0;
  final_target_rank=0;
  for (int h=1;h<n_proc;h++)
    {
      if (loi >= (map_u[h-1]+1)*(dim_v/dimensions[0]))
	{ 
	  init_target_rank=h;
	}
      if (hii >=(map_u[h-1]+1)*(dim_v/dimensions[0]))
	{
	  final_target_rank=h;
	}
    }
  number_data[0]=1;
  stride_data[0]=1;
  nonc_da_ty2=mpt2.get(number_data,stride_data);
  for (int h=init_target_rank;h<final_target_rank+1;h++)
    {
      target_rank=h;
      if (h==init_target_rank) 
	{
	  temp_lo=loi;
	}
      else {
	temp_lo=(map_u[h-1]+1)*(dim_v/dimensions[0]);
      }
      if (h==final_target_rank) 
	{
	  temp_hi=hii;
	}
      else {
	temp_hi=(map_u[h]+1)*(dim_v/dimensions[0])-1;
      }
      MPI_Win_lock(MPI_LOCK_EXCLUSIVE, target_rank, MPI_MODE_NOCHECK, 
		   win);
      //==================================================
      //       ||     ||  ||        ||     ||  ||  ||   ||
      //   values get                     displ   
      MPI_Type_extent(nonc_da_ty2,&extentd);
      if (h == 0) {
	lo_target=0;
      }
      else {
	lo_target=(map_u[h-1]+1)*(dim_v/dimensions[0]);
      }
      //      MPI_Put(&ptr[ih],(temp_hi-temp_lo+1),data_T,target_rank,(temp_lo-lo_target),
      //      (temp_hi-temp_lo+1),nonc_da_ty2,win);
            MPI_Accumulate(&ptr[ih],(temp_hi-temp_lo+1),data_T,target_rank,(temp_lo-lo_target),
		    (temp_hi-temp_lo+1),data_T,MPI_SUM,win);

      MPI_Win_unlock( target_rank, win ); 
      ih=ih+temp_hi-temp_lo+1;
    }

  MPI_Type_free(&nonc_da_ty2);
  return 0;
}




//=======================NEW
template <typename T>
int ba<T>::ba_multiply(const vector<int> loi,const vector<int> hi,T *ptr){
  int target_rank;
  int init_target_rank=0,final_target_rank=0;
  MPI_Aint	extentd;
  int local_dim_v;
  int ih=0;
  int low[n_proc],upp[n_proc];
  int temp_lo,temp_hi;
  int nd_minus_one;

  int displ,R_dim;
  //Error check
  //===========
  if(loi.size() != hi.size())
    {
      cerr<<"ERROR in ba_put:the size of lower and upper limit vectors differ"<<endl;
      return -1;
    }
  if(int(loi.size()) != number_of_dimensions || int(hi.size())!=number_of_dimensions)
    {
      cerr<<"ERROR in ba_put:size of lower and/or of upper limit vector differ from the number of dimensions of the array"<<endl; 
      return -1;
    }
  if(loi.size() == hi.size() && loi>hi)
    {
      cerr<<"ERROR in ba_put:lower limit larger than upper limit "<<endl;
      return -1;
    }
  //===========

  if (number_of_dimensions>1) 
    {
      nd_minus_one=number_of_dimensions-1;
    }
  else
    {
      nd_minus_one=1;
    }
 vector<int> number_data(nd_minus_one);
 vector<int> stride_data(nd_minus_one);
 int ka;
 
 MPI_Datatype nonc_da_ty2;
 mpi_type_trait< non_c_d_md<T> > mpt2;
  
 local_dim_v=1;
 for (int jl=1;jl<number_of_dimensions;jl++)
   local_dim_v*=(hi[jl]-loi[jl]+1);
 
 for (int h=0;h<n_proc;h++)
   {
     this->ba_distribution(h,&low[h],&upp[h]);
     if (loi[0]>=low[h]) 
       { 
	  init_target_rank=h;
	}
      if (hi[0]==loi[0]) 
	{
	  final_target_rank=init_target_rank;
	}
      else if (hi[0]>=low[h]) 
	{
	  final_target_rank=h;
	}
      
    }
 
 if (number_of_dimensions>1)
    {
      ka=number_of_dimensions-1;
      number_data[0]=hi[ka]-loi[ka]+1;
      stride_data[0]=dimensions[ka];
      for (int kx=1;kx<number_of_dimensions-1;kx++)
	{
	  ka-=1;
	  number_data[kx]=hi[ka]-loi[ka]+1;
	  stride_data[kx]=dimensions[ka];
	}
    }
  else
    {
      number_data[0]=1;
      stride_data[0]=1;
}
  nonc_da_ty2=mpt2.get(number_data,stride_data);
  for (int h=init_target_rank;h<final_target_rank+1;h++)
    {
      target_rank=h;
      if (h==init_target_rank) 
	{
	  temp_lo=loi[0];
	  if (hi[0]>upp[h])
	    {
	      temp_hi=upp[h];
	    }
	  else
	    {
	      temp_hi=hi[0];
	    }
	}
      else if (h==final_target_rank) 
	{
	  temp_lo=low[h];
	  temp_hi=hi[0];
	}
      else 
	{
	  temp_lo=low[h];
	  temp_hi=upp[h];
	}
      displ=(temp_lo-low[h])*dim_v/dimensions[0];
      R_dim=dim_v/dimensions[0];
      for (int ki=1;ki<number_of_dimensions;ki++)
	{
	  displ+=loi[ki]*R_dim/dimensions[ki];
	  R_dim=R_dim/dimensions[ki];
	}

      MPI_Win_lock(MPI_LOCK_EXCLUSIVE, target_rank, MPI_MODE_NOCHECK, 
		   win);
      //==================================================
      //       ||     ||  ||        ||     ||  ||  ||   ||
      //   values get                     displ   
      MPI_Type_extent(nonc_da_ty2,&extentd);

      MPI_Accumulate(&ptr[ih],(temp_hi-temp_lo+1)*(local_dim_v),data_T,target_rank,displ,(temp_hi-temp_lo+1),nonc_da_ty2,MPI_PROD,win);
      
      MPI_Win_unlock( target_rank, win ); 
      ih=ih+(temp_hi-temp_lo+1)*(local_dim_v);
    }
  
  MPI_Type_free(&nonc_da_ty2);
  return 0;
}

//******************************************************************************************
//! Experimental -- ba_put using the indices of the array rather than the pairs describing the 
//! corners.

//*********************
template <typename T>
int ba<T>::ba_multiply(const int loi,const int hii,T *ptr)
{
  int target_rank;
  int init_target_rank=0,final_target_rank=0;
  MPI_Aint	extentd;
  int local_dim_v;
  int ih=0;
  int temp_lo,temp_hi;

  int lo_target;
  
  //Error check
  //===========
  if(loi > hii)
    {
      cerr<<"ERROR in ba_put:lower limit larger than upper limit "<<endl;
      return -1;
    }
  //===========
  vector<int> number_data(1);
  vector<int> stride_data(1);
  
  MPI_Datatype nonc_da_ty2;
  mpi_type_trait< non_c_d_md<T> > mpt2;
  
  local_dim_v=1;
  
  init_target_rank=0;
  final_target_rank=0;
  for (int h=1;h<n_proc;h++)
    {
      if (loi >= (map_u[h-1]+1)*(dim_v/dimensions[0]))
	{ 
	  init_target_rank=h;
	}
      if (hii >=(map_u[h-1]+1)*(dim_v/dimensions[0]))
	{
	  final_target_rank=h;
	}
    }
  number_data[0]=1;
  stride_data[0]=1;
  nonc_da_ty2=mpt2.get(number_data,stride_data);
  for (int h=init_target_rank;h<final_target_rank+1;h++)
    {
      target_rank=h;
      if (h==init_target_rank) 
	{
	  temp_lo=loi;
	}
      else {
	temp_lo=(map_u[h-1]+1)*(dim_v/dimensions[0]);
      }
      if (h==final_target_rank) 
	{
	  temp_hi=hii;
	}
      else {
	temp_hi=(map_u[h]+1)*(dim_v/dimensions[0])-1;
      }
      MPI_Win_lock(MPI_LOCK_EXCLUSIVE, target_rank, MPI_MODE_NOCHECK, 
		   win);
      //==================================================
      //       ||     ||  ||        ||     ||  ||  ||   ||
      //   values get                     displ   
      MPI_Type_extent(nonc_da_ty2,&extentd);
      if (h == 0) {
	lo_target=0;
      }
      else {
	lo_target=(map_u[h-1]+1)*(dim_v/dimensions[0]);
      }
      //      MPI_Put(&ptr[ih],(temp_hi-temp_lo+1),data_T,target_rank,(temp_lo-lo_target),
      //      (temp_hi-temp_lo+1),nonc_da_ty2,win);
            MPI_Accumulate(&ptr[ih],(temp_hi-temp_lo+1),data_T,target_rank,(temp_lo-lo_target),
		    (temp_hi-temp_lo+1),data_T,MPI_PROD,win);

      MPI_Win_unlock( target_rank, win ); 
      ih=ih+temp_hi-temp_lo+1;
    }

  MPI_Type_free(&nonc_da_ty2);
  return 0;
}







//======================
//******************************************************************************************


template <typename T>
int ba<T>::ba_fill(T fv){
  T *fill_vect;

  vector<int> local_rows(2);
  vector<int> tmp_lo(number_of_dimensions);
  vector<int> tmp_hi(number_of_dimensions);

  local_rows=ba_distribution(rank);
  tmp_lo[0]=local_rows[0];
  tmp_hi[0]=local_rows[1];

  int local_size=(tmp_hi[0]-tmp_lo[0]+1);

  for (int kn=1;kn<number_of_dimensions;kn++)
    {
      tmp_lo[kn]=0;
      tmp_hi[kn]=dimensions[kn]-1;
      local_size=local_size*(tmp_hi[kn]-tmp_lo[kn]+1);
    }

  fill_vect= new T[local_size];
  for (int j=0;j<local_size;j++) fill_vect[j]=fv;
  
  this->ba_putF(tmp_lo,tmp_hi,fill_vect);
  delete[] fill_vect;

  return -1;
}


template <typename T>
int ba<T>::ba_size() const{return dim_v;}

template <typename T>
vector<int> ba<T>::ba_get_dims() const
{
  vector<int> to_return;
  for(int i=0; i<int(dimensions.size());i++)
    to_return.push_back(dimensions[i]-1);
  return to_return;
}


template <typename T>
int ba<T>::ba_get_loc_dimv() const
{
  int to_return=loc_dim_v;;
  return to_return;
}


template <typename T>
vector<int> ba<T>::ba_get_dimensions() const
{
  vector<int> to_return;
  for(int i=0; i<int(dimensions.size());i++)
    to_return.push_back(dimensions[i]);
  return to_return;
}

template <typename T>
MPI_Comm ba<T>::ba_get_comm() const
{ 
  MPI_Comm to_return;
  to_return=comm;
  return to_return;
}


//*******************
template <typename T>
void ba<T>::ba_get_access(T** ptr ) const
{
  *ptr=(this->g);
  
}
template <typename T>
T* ba<T>::ba_get_access() const
{
  return (this->g);
}
template <typename T>
T ba<T>::ba_get_access(int i) const
{
  return (this->g[i]);
}
template <typename T>
void ba<T>::ba_set(int i,T sv)
{
  (this->g[i]=sv);
}
//=========================
template <typename T>
int ba<T>::add(ba<T> a, ba<T> b)
{
  if (a.dimensions!=b.dimensions)
    {
      cerr<<"Error: you are trying to add two ba with different dimensions!"<<endl;
      return -1;
    }
  if (a.dimensions!=this->dimensions)
    {
      cerr<<"Error: the dimensions of the ba where you are trying to put the result of the sum are not correct!"<<endl;
      return -1;
    }
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      T temp=a.g[i]+b.g[i];
      //      this->ba_set(i,temp);
      this->g[i]=temp;
    }
  return 0;
}

template <typename T>
int ba<T>::subtract(ba<T> a, ba<T> b)
{
  if (a.dimensions!=b.dimensions)
    {
      cerr<<"Error: you are trying to subtract two ba with different dimensions!"<<endl;
      return -1;
    }
  if (a.dimensions!=this->dimensions)
    {
      cerr<<"Error: the dimensions of the ba where you are trying to put the result of the sum are not correct!"<<endl;
      return -1;
    }
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      T temp=a.g[i]-b.g[i];
      //      this->ba_set(i,temp);
      this->g[i]=temp;
    }
  return 0;
}
template <typename T>
int ba<T>::multiply(ba<T> a, ba<T> b)
{
  if (a.dimensions!=b.dimensions)
    {
      cerr<<"Error: you are trying to multiply two ba with different dimensions!"<<endl;
      return -1;
    }
  if (a.dimensions!=this->dimensions)
    {
      cerr<<"Error: the dimensions of the ba where you are trying to put the result of the sum are not correct!"<<endl;
      return -1;
    }
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      T temp=a.g[i]*b.g[i];
      //      this->ba_set(i,temp);
      this->g[i]=temp;
    }
  return 0;
}
template <typename T>
int ba<T>::divide(ba<T> a, ba<T> b)
{
  if (a.dimensions!=b.dimensions)
    {
      cerr<<"Error: you are trying to divide two ba with different dimensions!"<<endl;
      return -1;
    }
  if (a.dimensions!=this->dimensions)
    {
      cerr<<"Error: the dimensions of the ba where you are trying to put the result of the sum are not correct!"<<endl;
      return -1;
    }
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      if( b.g[i]!=0)
	{
	  T temp=a.g[i]/b.g[i];
	  this->g[i]=temp;
	}
      else
	{
	  cerr<<"division by zero: error!"<<endl;
	  return -1;
	}
    }
  return 0;
}

template <typename T>
int ba<T>::evaluate(ba<T> a, std::string function)
{
  //  std::string function;
  FunctionParser fparser;
      T * vals;
      vals=new T[1];
  //  std::getline(std::cin, function);
  int res = fparser.Parse(function, "x");
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      vals[0]=a.g[i];
      T temp=fparser.Eval(vals);
      this->g[i]=temp;
    }
  return 0;
}
template <typename T>
int ba<T>::evaluate(ba<T> a, ba<T> b,std::string function)
{
  //  std::string function;
  FunctionParser fparser;
  //  std::getline(std::cin, function);
  int res = fparser.Parse(function, "x,y");
      T * vals;
      vals=new T[2];
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      vals[0]=(a.g[i]);
      vals[1]=(b.g[i]);
      //      T temp=fparser.Eval(&(a.g[i]),&(b.g[i]));
      T temp=fparser.Eval(vals);
      //      this->ba_set(i,temp);
      this->g[i]=temp;
    }
  delete []vals;
  return 0;
}


template <typename T>
int ba<T>::evaluate(ba<T> a, ba<T> b, ba<T> c, std::string function)
{
  //  std::string function;
  FunctionParser fparser;
  //  std::getline(std::cin, function);
  int res = fparser.Parse(function, "x,y,z");
      T * vals;
      vals=new T[3];
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      vals[0]=(a.g[i]);
      vals[1]=(b.g[i]);
      vals[2]=(c.g[i]);
      //      T temp=fparser.Eval(&(a.g[i]),&(b.g[i]));
      T temp=fparser.Eval(vals);
      //      this->ba_set(i,temp);
      this->g[i]=temp;
    }
  delete []vals;
  return 0;
}


template <typename T>
T ba<T>::operator[] (const int index)
{
  //THIS WORKS ONLY ON THE RHS <===
  T * temp;
  temp=new T;
  this->ba_get_opt(index,temp);
  //  this->ba_get(index,index,temp);
  return *temp;
}


template <typename T>
ba<T> ba<T>::operator^ (T y)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<T> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      T temp=pow(this->g[i],y);
      local_ba.g[i]=temp;

    }
  return local_ba;
}
/*
template <typename T>  template <typename U>
ba<typename promote_trait<T,U>::T_promote> ba<T>::operator+ (U y)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<typename promote_trait<T,U>::T_promote> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      T temp=y+this->g[i];
      local_ba.g[i]=temp;

    }
  return local_ba;
}
template <typename T>
ba<T> ba<T>::operator- (T y)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<T> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      T temp=this->g[i]-y;
      local_ba.g[i]=temp;

    }
  return local_ba;
}


template <typename T>
ba<T> ba<T>::operator* (T y)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<T> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      T temp=y*this->g[i];
      local_ba.g[i]=temp;

    }
  return local_ba;
}



template <typename T>
ba<T> ba<T>::operator/ (T y)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<T> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      T temp=this->g[i]/y;
      local_ba.g[i]=temp;

    }
  return local_ba;
}
*/

template <typename T> template <typename U> 
ba<typename promote_trait<T,U>::T_promote > ba<T>::operator+(const ba<U> &b)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<typename promote_trait<T,U>::T_promote > local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"SUM NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=x+y what you got is z=0"<<endl;
  return local_ba;
};

template <typename T>  template <typename U>
ba<typename promote_trait<T,U>::T_promote> ba<T>::operator+ (U y)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<typename promote_trait<T,U>::T_promote> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"SUM NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=x+y what you got is z=0"<<endl;
  return local_ba;
}

template <typename T> template <typename U> 
ba<typename promote_trait<T,U>::T_promote> ba<T>::operator-(const ba<U> &b)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<typename promote_trait<T,U>::T_promote> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"SUBTRACTION NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=x-y what you got is z=0"<<endl;
  return local_ba;
};

template <typename T>  template <typename U>
ba<typename promote_trait<T,U>::T_promote> ba<T>::operator- (U y)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<typename promote_trait<T,U>::T_promote> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"SUBTRACTION NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=x-y what you got is z=0"<<endl;
  return local_ba;
}


template <typename T> template <typename U> 
ba<typename promote_trait<T,U>::T_promote_pr> ba<T>::operator*(const ba<U> &b)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<typename promote_trait<T,U>::T_promote_pr> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"PRODUCT NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=x*y what you got is z=0"<<endl;
  return local_ba;
};

template <typename T>  template <typename U>
ba<typename promote_trait<T,U>::T_promote_pr> ba<T>::operator* (U y)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<typename promote_trait<T,U>::T_promote_pr> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"PRODUCT NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  //  cerr<<"If you typed z=x*y what you got is z=0"<<endl;
  return local_ba;
}


template <typename T> template <typename U> 
ba<typename promote_trait<T,U>::T_promote> ba<T>::operator/(const ba<U> &b)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<typename promote_trait<T,U>::T_promote> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"DIVISION NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=x*y what you got is z=0"<<endl;
  return local_ba;
};

template <typename T>  template <typename U>
ba<typename promote_trait<T,U>::T_promote> ba<T>::operator/ (U y)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<typename promote_trait<T,U>::T_promote> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"DIVISION NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=x*y what you got is z=0"<<endl;
  return local_ba;
}




//==================
#include "operatorS.C"


//=======RELATIONAL OPERATORS===============
template <typename T> template <typename U> 
ba<int> ba<T>::operator<(const ba<U> &b)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING:less than NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=y<x what you got is z=0"<<endl;
  return local_ba;
};

template <typename T>  template <typename U>
ba<int> ba<T>::operator< (U y)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING:less than NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=y<x what you got is z=0"<<endl;
  return local_ba;
}

template <typename T> template <typename U> 
ba<int> ba<T>::operator>(const ba<U> &b)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING:larger than NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=y>x what you got is z=0"<<endl;
  return local_ba;
};

template <typename T>  template <typename U>
ba<int> ba<T>::operator> (U y)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING:larger than NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=y>x what you got is z=0"<<endl;
  return local_ba;
}
template <typename T> template <typename U> 
ba<int> ba<T>::operator>=(const ba<U> &b)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING:larger-eq than NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=y>=x what you got is z=0"<<endl;
  return local_ba;
};

template <typename T>  template <typename U>
ba<int> ba<T>::operator>= (U y)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING:larger-eq than NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=y>=x what you got is z=0"<<endl;
  return local_ba;
}


template <typename T> template <typename U> 
ba<int> ba<T>::operator<=(const ba<U> &b)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING:larger-eq than NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=y<=x what you got is z=0"<<endl;
  return local_ba;
};

template <typename T>  template <typename U>
ba<int> ba<T>::operator<= (U y)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING:larger-eq than NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=y<=x what you got is z=0"<<endl;
  return local_ba;
}


template <typename T> template <typename U> 
ba<int> ba<T>::operator==(const ba<U> &b)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING:larger than NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=y==x what you got is z=0"<<endl;
  return local_ba;
};

template <typename T>  template <typename U>
ba<int> ba<T>::operator== (U y)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING:larger than NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=y==x what you got is z=0"<<endl;
  return local_ba;
}

template <typename T> template <typename U> 
ba<int> ba<T>::operator!=(const ba<U> &b)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING:larger than NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=y!=x what you got is z=0"<<endl;
  return local_ba;
};

template <typename T>  template <typename U>
ba<int> ba<T>::operator!= (U y)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING:larger than NOT DEFINED FOR THESE TWO TYPES; Nothing done"<<endl;
  cerr<<"If you typed z=y!=x what you got is z=0"<<endl;
  return local_ba;
}


template <typename T> template <typename U> 
ba<int> ba<T>::operator&&(const ba<U> &b)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING: logical AND operator used with wrong BA types.Nothing done"<<endl;
  cerr<<"If you typed z=x && y what you got is z=0"<<endl;
  return local_ba;
};
template <typename T> template <typename U> 
ba<int> ba<T>::operator||(const ba<U> &b)
{
  ba<int> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  cerr<<"WARNING: logical AND operator used with wrong BA types.Nothing done"<<endl;
  cerr<<"If you typed z=x || y what you got is z=0"<<endl;
  return local_ba;
};


//==============

template <typename T>
  ba<T>& ba<T>::operator=(const ba<T> & b)
{
  //   cout<<" I am in the BA operator="<<endl;
    if(this==&b) 
      {
	cerr<<"what am i doing?"<<endl;
	return *this; 
      }
    if ((this->dim_v)==0) 
      {
	//	cerr<<"calling ba_create from the operator="<<endl;
	this->ba_create(b.dimensions,b.comm,0);
      }
    if ( (this->dimensions != b.dimensions) )
      {
	//	cerr<<"WARNING: the dimensions of the lhs BA were different from  those rhs BA:lhs BA deleted and new one with the right dimensions created "<<endl; 
	//	return *this;
	//	  cerr<<"OUT OF ="<<endl;
	this->ba_free();
	this->ba_create(b.dimensions,b.comm,0);
      }
    if ((this->comm != b.comm))
      {
	//cerr<<"WARNING: the MPI_COMM_WORLD of the lhs BA was different from that of the rhs BA:lhs BA deleted and new one with the right MPI_COMM_WORLD created "<<endl; 
	//	return *this;
	//	cerr<<"OUT OF ="<<endl;
	this->ba_free();
	this->ba_create(b.dimensions,b.comm,0);

      }	
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      T temp=b.g[i];
      this->g[i]=temp;
    }
  //  cerr<<"OUT OF ="<<endl;
    return *this;
};




/*template <typename T>
ba<T> ba<T>::operator-(const ba<T> &b)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<T> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  if (this->dimensions!=b.dimensions)
    {
      cerr<<"Error: you are trying to subtract two ba with different dimensions!"<<endl;
            return local_ba;
    }
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      T temp=this->g[i]+b.g[i];
      local_ba.g[i]=temp;

    }
  return local_ba;
}


template <typename T>
ba<T> ba<T>::operator*(const ba<T> &b)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<T> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  if (this->dimensions!=b.dimensions)
    {
      cerr<<"Error: you are trying to multiply two ba with different dimensions!"<<endl;
            return local_ba;
    }
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      T temp=this->g[i]*b.g[i];
      local_ba.g[i]=temp;

    }
  return local_ba;
}


template <typename T>
ba<T> ba<T>::operator/(const ba<T> &b)
{
  //  ba<T> * local_ba =new ba<T>(const_cast<ba<T> &> (b)) ;
  ba<T> local_ba;
  local_ba.ba_create(dimensions,comm,0);
  if (this->dimensions!=b.dimensions)
    {
      cerr<<"Error: you are trying to divide two ba with different dimensions!"<<endl;
            return local_ba;
    }
  for (int i=0;i<(loc_dim_v)*(dim_v/dimensions[0]);i++)
    {
      T temp=this->g[i]/b.g[i];
      local_ba.g[i]=temp;

    }
  return local_ba;
}
*/

//==================================
  using namespace std;
  namespace std{
    template<typename T>
    struct greater<aux_vect<T> > : public binary_function<aux_vect<T>,aux_vect<T>,bool> 
    {
      bool operator()(const aux_vect<T>& __x, const aux_vect<T>& __y) const { return __x.val > __y.val; }
    };
    template<typename T>
    struct less<aux_vect<T> > : public binary_function<aux_vect<T> ,aux_vect<T> ,bool> 
    {
      bool operator()(const aux_vect<T>& __x, const aux_vect<T>& __y) const { return __x.val < __y.val; }
    };


  }



//==================
template <typename T>
int ba<T>::ba_Sort(vector<aux_vect<T> >& myVector)
{
  cerr<<"Error:this type is not sortable"<<endl;
  return 1;
  /*  mpi_type_trait2<aux_vect<T> > mpt_av;
  MPI_Datatype data_T_av;
  data_T_av=mpt_av.get();

  if (number_of_dimensions>1) return -1;

  sort(myVector.begin(), myVector.end(),less<aux_vect<T> >());
  int size=(dim_v-1+n_proc)/n_proc;
  int rsize=(size+n_proc-1)/n_proc;

  vector<T> sample((n_proc-1));
  vector<T> local_sample(n_proc*(n_proc-1));
  for(int j = 1; j < n_proc; j++){
    if(j*rsize <= map_u[n_proc-1])
      sample[j-1] = (myVector[j*rsize]).val;
    else
      sample[j-1] = (myVector[map_u[n_proc-1]]).val;
  }
  MPI_Barrier(comm);

  MPI_Gather(&sample[0],(n_proc-1),data_T,&local_sample[0],(n_proc-1),data_T,0,comm);
  vector<T> pivots(n_proc-1);
  if (rank==0)
    {
      
      sort(local_sample.begin(),local_sample.end());
      for (int i=0;i<n_proc-1;i++)
	{
	  pivots[i]=(local_sample[i*n_proc+n_proc/2]);
	}
    }
  MPI_Barrier(comm);
  MPI_Bcast(&pivots[0],n_proc-1,data_T,0,comm);

  MPI_Barrier(comm);

  vector<aux_vect<T> > buckets[n_proc];

  int * count;
  count =new int[n_proc];

  int * rec_count;
  rec_count =new int[n_proc];

  for (int k=0;k<n_proc;k++)
    {
      count[k]=0;
    }

  for(int i = 0; i < loc_dim_v; i++){
    int flag=0;
    for(int j = 0; j < n_proc-1; j++){
      if(myVector[i].val < pivots[j]){
	buckets[j].push_back(myVector[i]);
	count[j]+=1;
	flag = 1;
	break;
      }
    }
    if(flag==0)
      {
      buckets[n_proc-1].push_back(myVector[i]);
      count[n_proc-1]+=1;
      }
  }

  MPI_Alltoall(count, 1, MPI_INT, rec_count,
        1, MPI_INT, MPI_COMM_WORLD);
  int tot=0;
  for (int h=0;h<n_proc;h++)
    tot+=rec_count[h];

  int * displacement=new int[n_proc];
  int * rec_displacement=new int[n_proc];
  displacement[0]=0;
  rec_displacement[0]=0;
  for (int i=1;i<n_proc;i++){
    displacement[i]=count[i-1]+displacement[i-1];
    rec_displacement[i]=rec_count[i-1]+rec_displacement[i-1];
  }

  MPI_Barrier(comm);
 
  aux_vect<T> * send_buckets;
  send_buckets= new aux_vect<T>[loc_dim_v];

  int jj=0;
  for (int i=0;i<n_proc;i++){
    for(int j=0;j<count[i];j++){
      send_buckets[jj]=buckets[i][j];
      jj+=1;
    }
    buckets[i].clear();
  }

  aux_vect<T> * rec_buckets=new aux_vect<T>[tot];
  MPI_Barrier(comm);

  MPI_Alltoallv(send_buckets,count,displacement,data_T_av,rec_buckets,rec_count,rec_displacement,data_T_av,comm);

    MPI_Barrier(comm);

    delete []send_buckets ;
    delete []count;
    delete []rec_count;
    delete []displacement;
    delete []rec_displacement;

    vector<aux_vect<T> > local_buckets;
    for(int i=0;i<tot; i++)
      local_buckets.push_back(rec_buckets[i]);
    
    MPI_Barrier(comm);
    
    delete []rec_buckets ;
    
    
    sort(local_buckets.begin(),local_buckets.end(),less<aux_vect<T> >());
    myVector.clear();
    myVector=local_buckets;
    */
  //Sorted

}



//=================
template <> inline
int ba<double>::ba_Sort(vector<aux_vect<double> >& myVector)
{
  
  mpi_type_trait2<aux_vect<double> > mpt_av;
  MPI_Datatype data_T_av;
  data_T_av=mpt_av.get();

  if (number_of_dimensions>1) return -1;

  sort(myVector.begin(), myVector.end(),less<aux_vect<double> >());
  int size=(dim_v-1+n_proc)/n_proc;
  int rsize=(size+n_proc-1)/n_proc;

  vector<double> sample((n_proc-1));
  vector<double> local_sample(n_proc*(n_proc-1));
  for(int j = 1; j < n_proc; j++){
    if(j*rsize <= map_u[n_proc-1])
      sample[j-1] = (myVector[j*rsize]).val;
    else
      sample[j-1] = (myVector[map_u[n_proc-1]]).val;
  }
  MPI_Barrier(comm);

  MPI_Gather(&sample[0],(n_proc-1),data_T,&local_sample[0],(n_proc-1),data_T,0,comm);
  vector<double> pivots(n_proc-1);
  if (rank==0)
    {
      
      sort(local_sample.begin(),local_sample.end());
      for (int i=0;i<n_proc-1;i++)
	{
	  pivots[i]=(local_sample[i*n_proc+n_proc/2]);
	}
    }
  MPI_Barrier(comm);
  MPI_Bcast(&pivots[0],n_proc-1,data_T,0,comm);

  /*  if(rank==0){
    std::cerr<<"pivots:";
  for (int i=0;i<n_proc-1;i++)
    {
            std::cerr<<" "<<pivots[i];
    }
  std::cerr<<std::endl;
  }*/
  MPI_Barrier(comm);

  vector<aux_vect<double> > buckets[n_proc];

  int * count;
  count =new int[n_proc];

  int * rec_count;
  rec_count =new int[n_proc];

  for (int k=0;k<n_proc;k++)
    {
      count[k]=0;
    }

  for(int i = 0; i < loc_dim_v; i++){
    int flag=0;
    for(int j = 0; j < n_proc-1; j++){
      if(myVector[i].val < pivots[j]){
	buckets[j].push_back(myVector[i]);
	count[j]+=1;
	flag = 1;
	break;
      }
    }
    if(flag==0)
      {
      buckets[n_proc-1].push_back(myVector[i]);
      count[n_proc-1]+=1;
      }
  }

  MPI_Alltoall(count, 1, MPI_INT, rec_count,
        1, MPI_INT, comm);
  int tot=0;
  for (int h=0;h<n_proc;h++)
    tot+=rec_count[h];

  int * displacement=new int[n_proc];
  int * rec_displacement=new int[n_proc];
  displacement[0]=0;
  rec_displacement[0]=0;
  for (int i=1;i<n_proc;i++){
    displacement[i]=count[i-1]+displacement[i-1];
    rec_displacement[i]=rec_count[i-1]+rec_displacement[i-1];
  }

  MPI_Barrier(comm);
 
  aux_vect<double> * send_buckets;
  send_buckets= new aux_vect<double>[loc_dim_v];

  int jj=0;
  for (int i=0;i<n_proc;i++){
    for(int j=0;j<count[i];j++){
      send_buckets[jj]=buckets[i][j];
      jj+=1;
    }
    buckets[i].clear();
  }

  aux_vect<double> * rec_buckets=new aux_vect<double>[tot];
  MPI_Barrier(comm);

  MPI_Alltoallv(send_buckets,count,displacement,data_T_av,rec_buckets,rec_count,rec_displacement,data_T_av,comm);

    MPI_Barrier(comm);

    delete []send_buckets ;
    delete []count;
    delete []rec_count;
    delete []displacement;
    delete []rec_displacement;

    vector<aux_vect<double> > local_buckets;
    for(int i=0;i<tot; i++)
      local_buckets.push_back(rec_buckets[i]);
    
    MPI_Barrier(comm);
    
    delete []rec_buckets ;
    
    
    sort(local_buckets.begin(),local_buckets.end(),less<aux_vect<double> >());
    myVector.clear();
    MPI_Type_free(&data_T_av);

    myVector=local_buckets;
    
  //Sorted

}

template <typename T>
vector<aux_vect<T> > ba<T>::make_sortable(T *y,int lo, int hi)
{
  aux_vect<T> temp;
  vector<aux_vect<T> > toReturn;
  for ( int j=0;j<hi+1-lo;j++)
    {
      temp.val=y[j];
      temp.ind=j+lo;
      toReturn.push_back(temp);
    }
  return toReturn;
}

//new version taking an STL vector
template <typename T>
vector<aux_vect<T> > ba<T>::make_sortable(vector<T> y,int lo, int hi)
{
  aux_vect<T> temp;
  vector<aux_vect<T> > toReturn;
  if (int(y.size())!=(hi+1-lo))
    {
      cerr<<"ERROR in make_sortable: the size of the vector is not consistent with the limits given. Empty vector returned."<<endl;
      return toReturn;
    }
  else
    {
      for ( int j=0;j<hi+1-lo;j++)
	{
	  temp.val=y[j];
	  temp.ind=j+lo;
	  toReturn.push_back(temp);
	}
      return toReturn;
    }
}


//new version assuming y, lo, hi are those of the local values
template <typename T>
vector<aux_vect<T> > ba<T>::make_sortable()
{
  aux_vect<T> temp;
  vector<aux_vect<T> > toReturn;
  if (int(ba_get_dims().size())==1)
    {
      int lo,hi;
      //check that rank is the right one
      ba_distribution_index(rank,&lo,&hi);
      for ( int j=0;j<hi+1-lo;j++)
	{
	  temp.val=ba_get_access(j);
	  temp.ind=j+lo;
	  toReturn.push_back(temp);
	}
      return toReturn;
    }
  else
    {
      cerr<<"Error:only 1-d Arrays are sortable"<<endl;
      return toReturn;
    }
}






//=================================================
// FUNCTIONS
//=================================================

template <typename T>
ba<T> ba_where(ba<int> &a, ba<T> &j, ba<T> &h)
{

  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  vector<int> dims_j;
  vector<int> dims_h;
  dims=a.ba_get_dimensions();
  dims_j=j.ba_get_dimensions();
  dims_h=h.ba_get_dimensions();

  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  if (dims!=dims_j || dims!=dims_h)
    {
      cerr<<"Error: the dimensions of the three Basin Arrays passed as arguments to the where function are not equal"<<endl;
      return local_ba;
    }
  int loc_d_v=a.ba_get_loc_dimv();
  T temp;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      if (a.ba_get_access(i))
	temp=j.ba_get_access(i);
      else
	temp=h.ba_get_access(i);
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}


template <typename T>
ba<T> ba_where(ba<int> &a, ba<T> &j,const  T &h)
{

  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  vector<int> dims_j;
  vector<int> dims_h;
  dims=a.ba_get_dimensions();
  dims_j=j.ba_get_dimensions();

  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  if (dims!=dims_j )
    {
      cerr<<"Error: the dimensions of the Basin Arrays passed as arguments to the where function are not equal"<<endl;
      return local_ba;
    }
  int loc_d_v=a.ba_get_loc_dimv();
  T temp;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      if (a.ba_get_access(i))
	temp=j.ba_get_access(i);
      else
	temp=h;
	local_ba.ba_set(i,temp);
    }
  return local_ba;
}


template <typename T>
ba<T> ba_where(ba<int> &a,const T &j, ba<T> &h)
{

  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  vector<int> dims_j;
  vector<int> dims_h;
  dims=a.ba_get_dimensions();
  dims_h=h.ba_get_dimensions();

  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  if (dims!=dims_h )
    {
      cerr<<"Error: the dimensions of the Basin Arrays passed as arguments to the where function are not equal"<<endl;
      return local_ba;
    }
  int loc_d_v=a.ba_get_loc_dimv();
  T temp;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      if (a.ba_get_access(i))
	temp=j;
      else
	temp=h.ba_get_access(i);
	local_ba.ba_set(i,temp);
    }
  return local_ba;
}



template <typename T>
ba<T> ba_where(ba<int> &a,const  T &j, const T &h)
{

  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  vector<int> dims_j;
  vector<int> dims_h;
  dims=a.ba_get_dimensions();

  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T temp;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      if (a.ba_get_access(i))
	temp=j;
      else
	temp=h;
	local_ba.ba_set(i,temp);
    }
  return local_ba;
}






template <typename T>
ba<T> pow(ba<T> &a,double z)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
 /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=(T)pow((double)loc_g,z);
      //HERE I AM CASTING TO double BUT TO BE CHANGED
      //to take care of other types such as vector3, strings etc.
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}


template <typename T>
ba<fftw_complex> pow(ba<T> &a,fftw_complex z)
{
  ba<fftw_complex> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  fftw_complex loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  return local_ba;
}


template <typename T>
ba<T> real(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=log(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}
template <typename T>
ba<T> imag(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=log(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}




template <typename T>
ba<T> log(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=log(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}

template <typename T>
ba<T> log10(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /* for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=log10(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}

template <typename T>
ba<T> sin(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=sin(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}



template <typename T>
ba<T> cos(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=cos(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}


template <typename T>
ba<T> cosh(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=cosh(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}


template <typename T>
ba<T> sinh(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=sinh(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}


template <typename T>
ba<T> asin(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=asin(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}


template <typename T>
ba<T> acos(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=acos(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}




template <typename T>
ba<T> asinh(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=asinh(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}
template <typename T>
ba<T> acosh(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=acosh(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}


template <typename T>
ba<T> abs(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=abs(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}

template <typename T>
ba<T> sqrt(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=sqrt(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}


template <typename T> 
ba<T> exp(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=0;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}






template <typename T>
ba<T> exp10(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=pow(10.0,loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}



template <typename T>
ba<T> atan(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"OPERATION NOT DEFINED"<<endl;
  /*  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      T temp=atan(loc_g);
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;
}


inline
ba<double> ba_random(vector<int> dims,unsigned int seed=0)
{
  //======
    vector<int>::iterator pos;
    int d_v;
    d_v=1;
    for (pos=dims.begin();pos<dims.end();++pos)
    d_v=d_v*(*pos);

  //=========
  ba<double> local_ba;
  MPI_Comm cm=BASIN_COMM;
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=local_ba.ba_get_loc_dimv();
  int local_rank;
  MPI_Comm_rank(cm,&local_rank);
  if (seed==0)
    srand ( time(NULL)*(local_rank+1));
  else 
    srand (seed*(local_rank+1));

  double temp;
   for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      temp=(double)rand()/(double)RAND_MAX;
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}

inline
ba<double> ba_random(int dim1,unsigned int seed=0)
{
  //======
  int d_v;
    d_v=dim1;
    vector<int> dims(1);
    dims[0]=dim1;
  //=========
  ba<double> local_ba;
  MPI_Comm cm=BASIN_COMM;
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=local_ba.ba_get_loc_dimv();
  int local_rank;
  MPI_Comm_rank(cm,&local_rank);
  if (seed==0)
    srand ( time(NULL)*(local_rank+1));
  else 
    srand (seed*(local_rank+1));
  double temp;
   for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      temp=(double)rand()/(double)RAND_MAX;
      local_ba.ba_set(i,temp);
    }
  return local_ba;
}





template <typename T>
ba<T> ba_Sort_BA(ba<T> &a)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  T loc_g;
  cerr<<"Error:this type is not sortable"<<endl;
  return local_ba;

}


template <typename T>
ba<int> index_sorted(ba<T> &a)
{
  ba<int> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  cerr<<"Error:this type is not sortable"<<endl;
  return local_ba;

}

template <typename T>
T sum_BA(ba<T> &a)
{
  T local_ba;
  cerr<<"Error:you can not sum elements of this type"<<endl;
  return local_ba;

}



template <typename T>
ba<T> userf(ba<T> &a,string func)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  local_ba.evaluate(a,func);
  return local_ba;
}

template <typename T>
ba<T> userf(ba<T> &a,ba<T> &b,string func)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  local_ba.evaluate(a,b,func);
  return local_ba;
}

template <typename T>
ba<T> userf(ba<T> &a,ba<T> &b, ba<T> &c,string func)
{
  ba<T> local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  int loc_d_v=a.ba_get_loc_dimv();
  local_ba.evaluate(a,b,c,func);
  return local_ba;
}
//=====================================
//OPERATORS
//=====================================
template <typename T> 
ba<typename promote_trait<T,double>::T_promote > operator+ (double y, const ba<T> &a)
{
  ba<typename promote_trait<T,double>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"SUM NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}

template <typename T> 
ba<typename promote_trait<T,float>::T_promote > operator+ (float y, const ba<T> &a)
{
  ba<typename promote_trait<T,float>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"SUM NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}

template <typename T> 
ba<typename promote_trait<T,int>::T_promote > operator+ (int y, const ba<T> &a)
{
  ba<typename promote_trait<T,int>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"SUM NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}

#ifdef VECTOR23
template <typename T> 
ba<typename promote_trait<T,Vector2>::T_promote > operator+ (Vector2 y, const ba<T> &a)
{
  ba<typename promote_trait<T,Vector2>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"SUM NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}
template <typename T> 
ba<typename promote_trait<T,Vector3>::T_promote > operator+ (Vector3 y, const ba<T> &a)
{
  ba<typename promote_trait<T,Vector3>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"SUM NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}
#endif

template <typename T> 
ba<typename promote_trait<T,fftw_complex>::T_promote > operator+ (fftw_complex y, const ba<T> &a)
{
  ba<typename promote_trait<T,fftw_complex>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"SUM NOT DEFINED FOR THESE TWO TYPES"<<endl;
  return local_ba;

}



template <typename T> 
ba<typename promote_trait<T,double>::T_promote > operator- (double y, const ba<T> &a)
{
  ba<typename promote_trait<T,double>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"SUBTRACTION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}

template <typename T> 
ba<typename promote_trait<T,float>::T_promote > operator- (float y, const ba<T> &a)
{
  ba<typename promote_trait<T,float>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"SUBTRACTION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}

template <typename T> 
ba<typename promote_trait<T,int>::T_promote > operator- (int y, const ba<T> &a)
{
  ba<typename promote_trait<T,int>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"SUBTRACTION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*    int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}

#ifdef VECTOR23
template <typename T> 
ba<typename promote_trait<T,Vector2>::T_promote > operator- (Vector2 y, const ba<T> &a)
{
  ba<typename promote_trait<T,Vector2>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"SUBTRACTION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*    int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}
template <typename T> 
ba<typename promote_trait<T,Vector3>::T_promote > operator- (Vector3 y, const ba<T> &a)
{
  ba<typename promote_trait<T,Vector3>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"SUBTRACTION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*    int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}
#endif

template <typename T> 
ba<typename promote_trait<T,fftw_complex>::T_promote > operator- (fftw_complex y, const ba<T> &a)
{
  ba<typename promote_trait<T,fftw_complex>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"SUBTRACTION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  return local_ba;
}




template <typename T> 
ba<typename promote_trait<T,double>::T_promote_pr > operator* (double y, const ba<T> &a)
{
  ba<typename promote_trait<T,double>::T_promote_pr > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"MULTIPLICATION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}

template <typename T> 
ba<typename promote_trait<T,float>::T_promote_pr > operator* (float y, const ba<T> &a)
{
  ba<typename promote_trait<T,float>::T_promote_pr > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"MULTIPLICATION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}

template <typename T> 
ba<typename promote_trait<T,int>::T_promote_pr > operator* (int y, const ba<T> &a)
{
  ba<typename promote_trait<T,int>::T_promote_pr > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"MULTIPLICATION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}


#ifdef VECTOR23
template <typename T> 
ba<typename promote_trait<T,Vector2>::T_promote_pr > operator* (Vector2 y, const ba<T> &a)
{
  ba<typename promote_trait<T,Vector2>::T_promote_pr > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"MULTIPLICATION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}
template <typename T> 
ba<typename promote_trait<T,Vector3>::T_promote_pr > operator* (Vector3 y, const ba<T> &a)
{
  ba<typename promote_trait<T,Vector3>::T_promote_pr > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"MULTIPLICATION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}
#endif

template <typename T> 
ba<typename promote_trait<T,fftw_complex>::T_promote_pr > operator* (fftw_complex y, const ba<T> &a)
{
  ba<typename promote_trait<T,fftw_complex>::T_promote_pr > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"MULTIPLICATION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  return local_ba;

}








template <typename T> 
ba<typename promote_trait<T,double>::T_promote > operator/ (double y, const ba<T> &a)
{
  ba<typename promote_trait<T,double>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"DIVISION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}

template <typename T> 
ba<typename promote_trait<T,float>::T_promote > operator/ (float y, const ba<T> &a)
{
  ba<typename promote_trait<T,float>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"DIVISION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}

template <typename T> 
ba<typename promote_trait<T,int>::T_promote > operator/ (int y, const ba<T> &a)
{
  ba<typename promote_trait<T,int>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"DIVISION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}


#ifdef VECTOR23
template <typename T> 
ba<typename promote_trait<T,Vector2>::T_promote > operator/ (Vector2 y, const ba<T> &a)
{
  ba<typename promote_trait<T,Vector2>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"DIVISION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}
template <typename T> 
ba<typename promote_trait<T,Vector3>::T_promote > operator/ (Vector3 y, const ba<T> &a)
{
  ba<typename promote_trait<T,Vector3>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"DIVISION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  /*  int loc_d_v=a.ba_get_loc_dimv();
  typename promote_trait<T,U>::T_promote loc_g;
  for (int i=0;i<(loc_d_v)*(d_v/dims[0]);i++)
    {
      loc_g=a.ba_get_access(i);
      typename promote_trait<T,U>::T_promote temp=loc_g+y;
      local_ba.ba_set(i,temp);
      }*/
  return local_ba;

}
#endif

template <typename T> 
ba<typename promote_trait<T,fftw_complex>::T_promote > operator/ (fftw_complex y, const ba<T> &a)
{
  ba<typename promote_trait<T,fftw_complex>::T_promote > local_ba;
  int d_v=a.ba_size();
  vector<int> dims;
  dims=a.ba_get_dimensions();
  MPI_Comm cm=a.ba_get_comm();
  local_ba.ba_create(dims,cm,0);
  cerr<<"DIVISION NOT DEFINED FOR THESE TWO TYPES"<<endl;
  return local_ba;

}



#include "operatorS2.C"
#include "functionS.C"

#include <boost/variant.hpp>
#endif
