#include <string>
#include <sstream>
#include <iostream>
#include <vector>
#include <stack>
#include <queue>
#include <cmath>
#include "Region.h"
#include "basin_base.h"
#include "hdf5.h"

using namespace std;

class HDF5_IO: public BASIN_IO {
	public:
		HDF5_IO (string, Region*);
		~HDF5_IO ();
	private:
		bool CloseFile();

		/* Input */
		bool OpenFile ();
		bool OpenRoot ();
		bool CloseRoot ();
		bool GetDatasets ();
		bool GetGroups ();
		bool NextDataset ();
		bool NextGroup ();
		bool MakeDataObjects ();
		bool MakeList (vector<int>);
		bool MakeGrid (vector<int>);
		bool IsGrid (vector<int>);
		bool MakeSubGroups (); 
		vector<Attribute*> HDF5GetDataset (hid_t, const char*);
		vector<Attribute*> HDF5Integer (unsigned long int, hid_t, hid_t, hsize_t, hsize_t*);
		vector<Attribute*> HDF5Float (unsigned long int, hid_t, hid_t, hsize_t, hsize_t*);
		vector<Attribute*> HDF5Enum (unsigned long int, hid_t, hid_t, hsize_t, hsize_t*);
		vector<Attribute*> HDF5Array (unsigned long int, hid_t, hid_t, hsize_t, hsize_t*);
		vector<Attribute*> HDF5Compound (unsigned long int, hid_t, hid_t, hsize_t, hsize_t*);

		/* Output */
		bool CreateFile ();
		bool CloseWrittenFile ();
		bool WriteDatasets ();
		bool WriteGrid (Grid*);
		bool WriteList (List*);
		bool WriteSubGroups ();
		bool WriteIntegers (Attribute*);
		bool WriteFloats (Attribute*);
		bool WriteDoubles (Attribute*);
		bool WriteCharacter (Attribute*);
		bool WriteComplex (Attribute*);
#ifdef VECTOR23		
		bool WriteVector2 (Attribute*);
		bool WriteVector3 (Attribute*);
#endif
		hid_t _fileID;
		stack<hid_t> _groupID;
		stack<Region*> _currentRegion;
		vector<string> _datasetsLeft;
		vector<vector<int> > _datasetsSorted;
		stack<stack<string> > _groupsLeft;
};
