#include <string>
#include <sstream>
#include <iostream>
#include <vector>
#include "Region.h"
#include "basin_base.h"

using namespace std;

class TEXT_IO: public BASIN_IO {
 public:
  TEXT_IO (string, Region*);
  ~TEXT_IO ();
 private:
  bool OpenFile ();
  bool CloseFile ();
  bool NextDataset ();
  bool MakeDataObjects ();
  bool MakeList ();
  bool MakeGrid ();
  void SkipLines();
  void SkipInLine(string &data, int &pos);
  void RevSkipInLine(string &data, int &pos);
  bool AssignVariable(string &data, map<const string, string> &header);
  bool ReadHeader (map<const string, string> &header);
  bool ReadDataLine (vector<string> &data, int capacity);
  bool IsGrid();
  int NumDimensions();
  unsigned long int NumElements(int nCols);
  vector<int> NumCell();
  vector<double> LowerLeft ();
  vector<double> UpperRight ();
  void AttInfo (unsigned int nCols, vector<string> &colName, vector<int> &colType, 
		vector<int> &colRank1, vector<int> &colRank2);
  void FillLine (vector<string>&line, vector<int> &colType, vector<int> &colRank1, vector<int> &colRank2);
  int ReadData(int isGrid, vector<string> &colName, vector<int> colType, vector<int> &colRank1, vector<int> &colRank2, string index);
  bool NextGroup ();
  bool CreateFile();
  bool CloseWrittenFile();
  bool WriteDatasets();
  bool WriteData(Data *data, bool isGrid);
  bool PrintAttributes(Data* data, vector<string> attName, vector<string> attType, vector<int> colRank1);
  bool PrintHeader(Data *data, bool isGrid, vector<string> attName, vector<string> attType);

  bool _hasHeader;
  bool _isFirstData;
  bool _nTotKnown;
  bool _isGrid;
  int _nProcs;
  List *_newList;
  Grid *_newGrid;
  map<const string, string> _header;
  unsigned long int _iList;
  unsigned long int _iGrid;
  unsigned int _nInts;
  unsigned int _nDoubles;
  unsigned int _nStrings;
  unsigned int _nVec2ds;
  unsigned int _nVec3ds; 
  unsigned int _nFloats;

  fstream _fileStr;

  string _equalKey;
  string _gridKey;
  string _nDimsKey;
  string _llKey;
  string _urKey;
  string _nTotKey;
  //string _posNameKey;
  string _colNameKey;
  string _colTypeKey;
  //string _colRankKey;
  string _nColKey;
  string _nxKey;
  //string _metricKey;
  //string _boundaryKey;
  //string _spaceKey;
  string _extraVarName;

  string _doubleTypeKey;
  string _intTypeKey;
  string _stringTypeKey;
  string _floatTypeKey;
  string _vec2TypeKey;
  string _vec3TypeKey;
  string _delimiters;

  vector<vector<int>*> _iArray;
  vector<vector<double>*> _dArray;
  vector<vector<char>*> _sArray;
#ifdef VECTOR23
  vector<vector<Vector2>*> _vec2DArray;
  vector<vector<Vector3>*> _vec3DArray; 
#endif
  vector<vector<float>*> _fArray;
};
