/*
 *  C++ core and functions for BASIN.
 *  Copyright (C) 2006  BASIN team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*! A simple conversion from redshift to distance.  Uses a linear approximation.  */

class cosmology{
 public:
  //! Create a lambda CDM cosmology by default
  cosmology();

  //! Convert a single attribute to comoving distance
  Attribute* redshift_to_distance(List*,string zname="",string rname="r",int 
				  distance_type=0,int hflag=0);

  //! Convert a single attribute to comoving distance
  Attribute& redshift_to_distance(Attribute& redshift,int distance_type=0,int hflag=0);

  //! Compute the Hubble Constant at a particular redshift
  Attribute hubble(Attribute& a);
  double hubble(double a);

  //! Convert a single attribute to linear growth factor
  Attribute* redshift_to_growth(List*,string zname="",string dname="d");
  Attribute redshift_to_growth(Attribute& redshift);
  double redshift_to_growth(double z);


  //! Compute the power spectrum (from the fitting form of Sugiyama, 1995) 
  //! of a single value of k (in h/Mpc) or
  //! many, at a particular epoch
  double power_spectrum(double k,double z=0);
  vector<double> power_spectrum(vector<double> k, double z=0);
  Attribute& power_spectrum(Attribute& k,double z=0);

  double get_omega_m() {
    return omega_m;
  };
  int set_omega_m(double omega_m_new) {
    omega_m=omega_m_new;
    return 0;
  };
  
  double get_h() {
    return h;
  };
  int set_h(double h_new) {
    h=h_new;
    return 0;
  };
  
  double get_omega_l() {
    return omega_l;
  };
  int set_omega_l(double omega_l_new) {
    omega_l=omega_l_new;
    return 0;
  };

  double get_sig_8() {
    return sig_8;
  };

  int set_sig_8(double val_new) {
    sig_8=val_new;
  };
  


 private:
  double omega_m, omega_b, omega_l,omega_r,T_0,h,n_s,sig_8,w;
  double norm;
  int normset,k_curv;

  int normalize();

};


