#include "basin_package.h"
#include "linearalgebra.h"

matrix::matrix() {
  n=2;
  m=2;
  data=(double*)malloc(n*m*sizeof(double));
  for (int i=0; i < n*m; i++) {data[i]=0;}
}

matrix::matrix(int n1,int m1) {
  n=n1;
  m=m1;
  data=(double*)malloc(n*m*sizeof(double));
  for (int i=0; i < n*m; i++) {data[i]=0;}
}

matrix::matrix(int n1) {
  // Make an identity matrix;
  n=n1;
  m=n1;
  data=(double*)malloc(n*m*sizeof(double));
  for (int i=0; i < n; i++) {
    for (int j=0; j < n; j++) {
      if (i == j) {
	data[i*n+j]=1.;
      } else {
	data[i*n+j]=0.;
      }
    }
  }
}

matrix::matrix(double *arr,int n1, int m1) {
  data=arr;
  n=n1;
  m=m1;

}

matrix& matrix::operator+(matrix &b) {
  if ((n==b.get_n()) && (m==b.get_m())) {
    matrix* c=new matrix(n,m);
    double* cdat=c->get_data();
    double* bdat=b.get_data();
    for (int i=0; i < n*m; i++) {
      cdat[i]=data[i]+bdat[i];
    }
    return *c;
  } else {
    return b;
  }
}

matrix& matrix::operator-(matrix &b) {
  if ((n==b.get_n()) && (m==b.get_m())) {
    matrix* c=new matrix(n,m);
    double* cdat=c->get_data();
    double* bdat=b.get_data();
    for (int  i=0; i < n*m; i++) {
      cdat[i]=data[i]-bdat[i];
    }
    return *c;
  }
  else {
    return *this;
  }
}

matrix& matrix::operator*(matrix &b) {
  matrix *a=this;
  return(matrix_mult(*a,b));
}

matrix& matrix::operator*(double c){
  return (c*(*this));
}

matrix& matrix::operator*(int c){
  return (double(c)*(*this));
}

matrix& operator*(double c,matrix& a){
  matrix* b=new matrix(a.get_n(),a.get_m());
  double* adat=a.get_data();
  double* bdat=b->get_data();
  for (int i=0; i < a.get_n()*a.get_m(); i++) {
    bdat[i]=c*adat[i];
  }
  return *b;
}

matrix& operator*(int c, matrix& a) {
  return(double(c)*a);
}

double& matrix::get_element(int i, int j) {
  return(data[i*m+j]);
}

double& matrix::operator()(int i, int j) {
  return(get_element(i,j));
}




double* matrix::get_data(){
  return(data);
}

double* matrix::copy_data() {
  double* outdata;
  double* indata=get_data();
  int n1=n;
  int m1=m;
  outdata=(double*)malloc(sizeof(double)*n*m);
  for (int i=0; i< n*m; i++) outdata[i]=indata[i];
  return outdata;
}

int matrix::get_n() {
  return n;
}

int matrix::get_m() {
  return m;
}



matrix& matrix_invert(matrix &M) {
  // Only work for square matrices
  // For now, we use Gaussian elimination.
  if (M.get_n() != M.get_m()) {return(M);}
  int n=M.get_n();
  double *A=M.copy_data();
  matrix* Mi_ptr=new matrix(n);
  
  double *Ai=Mi_ptr->get_data();

  
  for (int i=0; i < n; i++) {
    // Do each row, and look at the leading term.                                    
    double val0=A[i*n+i];
    for (int k=0; k < n ; k++) {
      A[i*n+k]=A[i*n+k]/val0;
      Ai[i*n+k]=Ai[i*n+k]/val0;
    }
    for (int j=0; j < n; j++) {
      if (i != j) {
        double val1=A[j*n+i];
        for (int k=0; k < n; k++) {
          A[j*n+k]=A[j*n+k]-val1*A[i*n+k];
          Ai[j*n+k]=Ai[j*n+k]-val1*Ai[i*n+k];
        }
      }
    }
  }

  return *(Mi_ptr);
}

matrix& matrix_mult(matrix &a, matrix &b) {
  int n1=a.get_n();
  int m1=a.get_m();
  int n2=b.get_n();
  int m2=b.get_m();
  if (m1 != n2) {return a;}
  matrix* c=new matrix(n1,m2);
  double* dat=c->get_data();
  double* d1=a.get_data();
  double* d2=b.get_data();
  for (int i=0; i < n1; i++) {
    for (int j=0; j < m2; j++) {
      dat[i*m2+j]=0.;
      for (int k=0; k < m1; k++) {
	dat[i*m2+j]+=d1[i*m1+k]*d2[k*m2+j];
      }
    }
  }
  return *c;

}
