#include "basin_package.h"
#include "simple_plot.h"

void plot(List* mylist,string var1, string var2, string title) {
  Attribute* attr1=mylist->get_attribute_pointer(var1);
  Attribute* attr2=mylist->get_attribute_pointer(var2);
  plot(*attr1,*attr2,title);
}

void plot(Attribute &x, Attribute &y, string title) {
  if (BASIN_ID == 0) {
    // This one is a scatter plot
    gnuplot plot;
    
    plot.title=title;
    plot << "set size square;";
    int n=x.size();
    
    for (int i=0; i < n; i++) {
      plot << double(x[i]) << double(y[i]);
    }
  }
}

void plot(Grid* gridptr,string key,string title) {
  // Does the key exist?
  Attribute *attr_ptr;
  if (gridptr->attribute_exist(key)) {
    attr_ptr=gridptr->get_attribute_pointer(key);
  } else {
    attr_ptr=gridptr->get_attribute_pointer(0);
  }
  plot(*attr_ptr,title);
}


void plot(Attribute &attr, string title) {
  if (BASIN_ID == 0) {
    // Do a greyscale plot (if 2d), or a line plot (if 1d)
    vector<int> nx=attr.get_dims();
    int ndims=nx.size();
    gnuplot plot;
    plot.title=title;
    
    if (ndims == 1) {
      plot.style="lines";
      for (int i=0; i < nx[0]; i++) {
	plot << i << double(attr[i]);
      }
    }
    else if (ndims == 2) {
      plot.plot3d=true;
      plot.style="pm3d";
      plot << "set pm3d map;";
      plot << "set size square;";
      for (int i=0; i < nx[0]; i++) {
	for (int j=0; j < nx[1]; j++) {
	  plot << i << j << double(attr(i,j));
	}
	plot.insert_nl();
      }    
    }
  }
}

void plot(vector<double> x, vector<double> y,int type,int axes,string title) {
  if (BASIN_ID == 0) {
    gnuplot plot;
    plot.title=title;
    // axes=0 -> lin-lin, axes=1 -> lin-log, axes=2 log-lin, axes=3, log,log
    if (axes == 1) {
      plot.set_log("y");
    } else if (axes == 2) {
      plot.set_log("x");
    } else if (axes == 3) {
      plot.set_log("xy");
    }
    int n=x.size();
    if (type == 1) {
      plot.style="line";
    } else {
      plot << "set size square;";
  }
    for (int i=0; i < n ; i++) {
    plot << x[i] << y[i];
    }
  }
}

void plot(vector<double> x, vector<double>y1, vector<double> y2) {
  if (BASIN_ID == 0) {
  gnuplot plot;
  plot.title="data overlay";
  int n=x.size();
  plot.set_multiplot();
  for (int i=0; i < n ; i++) plot << x[i] << y1[i];
  plot << flush ;
  plot.style="line";
  for (int i=0; i < n; i++)  plot << x[i] << y2[i];
  }

}

