#include "basin_package.h"
#include "statistics.h"


template <typename T> vector<double> bin_data_template(Attribute& x, Attribute& y, 
						       vector<double> xbin, T dum) {
  
  double xmin=xbin[0],xmax=xbin[xbin.size()-1];
  double dx=xbin[1]-xbin[0];
  int nbins=xbin.size();
  T *ydata;
  double *xdata;
  int lo,hi;
  int basin_id,basin_nprocs;
  MPI_Comm_size(MPI_COMM_WORLD,&basin_nprocs); 
  MPI_Comm_rank(MPI_COMM_WORLD,&basin_id);


  double *tmparr,*tmparr_loc;
  int *numarr,*numarr_loc;
  vector<double>outarr(nbins,0.);

  y.get_gen_distribution_index(basin_id, &lo, &hi);
  xdata=(double*)malloc(sizeof(double)*(hi-lo+1));
  x.get(lo,hi,xdata);
  ydata=(T*)malloc(sizeof(T)*(hi-lo+1));
  y.get(lo,hi,ydata);

  tmparr=(double*)malloc(sizeof(double)*nbins);
  tmparr_loc=(double*)malloc(sizeof(double)*nbins);
  numarr=(int*)malloc(sizeof(int)*nbins);
  numarr_loc=(int*)malloc(sizeof(int)*nbins);
  for (int i=0; i < nbins; i++) {
    numarr[i]=0;
    tmparr[i]=0.;
  }
  int id;
  for (int i=0; i < hi-lo+1; i++) {
    id=int(0.5+(xdata[i]-xmin)/dx);
    //berr << i << " " << xdata[i] << " " << id << "\n";
    if ((id >=0 )&&(id < nbins)) {
      numarr[id]++;
      tmparr[id]+=ydata[i];
    }
  }
  free(xdata);
  free(ydata);
  // Now, we need to reduce the data
  MPI_Allreduce(tmparr,tmparr_loc,nbins,MPI_DOUBLE,MPI_SUM,BASIN_COMM);
  MPI_Allreduce(numarr,numarr_loc,nbins,MPI_INT,MPI_SUM,BASIN_COMM);
  for (int i=0; i < nbins; i++) {
    if (numarr_loc[i] > 0) {
      outarr[i]=tmparr_loc[i]/numarr_loc[i];
    }
    else {
      outarr[i]=0;
    }
  }
  free(tmparr);
  free(tmparr_loc);
  free(numarr);
  free(numarr_loc);
  return outarr;
}

vector<double> bin_data(Attribute& x, Attribute& y, vector<double> xbin) {
  // Takes the average of data within a particular bin (much like
  // histogram).  The bin is specified by xbin, which is assumed to be of type double.
  // xbin is assumed to be evenly spaced.
  int type=y.get_data_type(); 
  if (type == BASIN_DATA_INT) {
    return bin_data_template(x,y,xbin,int(0));
  } else if (type == BASIN_DATA_FLOAT) {
    return bin_data_template(x,y,xbin,float(0.));
  } else if (type == BASIN_DATA_DOUBLE) {
    return bin_data_template(x,y,xbin,double(0.));
  } else {
    vector<double> outbin=xbin;
    return outbin;
  }
}

