#include "basin_package.h"
#include "linearalgebra.h"
#include "statistics.h"


function& poly_fit(Attribute& x, Attribute& y, Attribute& w, int order) {
  function* fit= new function(0,order);
  vector<double> coeff(order+1,0.);
  int n=order+1;
  double adat[n*n];
  double bdat[n];
  for (int l=0; l < n; l++) {
    bdat[l]=sum(pow(x,l)*w*y);
    for (int m=0; m < n; m++) {
      adat[l*n+m]=sum(pow(x,l+m)*w);
    }
  }
  matrix A(adat,n,n);


  matrix B(bdat,n,1);

  matrix Ai=matrix_invert(A);


  matrix C=Ai*B;

  double *dat=C.get_data();
  for (int i=0; i < n; i++) {
    coeff[i]=dat[i];
  }



  fit->set_coeffs(coeff);
  return *fit;
}

function& poly_fit(Attribute& x, Attribute& y, int order) {
  Attribute w=x*0.+1.;
  return poly_fit(x,y,w,order);
}
// Now for vector <double>
function& poly_fit(vector<double>& x, vector<double>& y, vector<double>& w, int order) {
  function* fit= new function(0,order);
  vector<double> coeff(order+1,0.);
  int n=order+1;
  double adat[n*n];
  double bdat[n];
  for (int l=0; l < n; l++) {
    bdat[l]=0;
    for (int i=0; i < x.size(); i++) { 
      bdat[l]=bdat[l]+pow(x[i],l)*w[i]*y[i];}
    for (int m=0; m < n; m++) {
      adat[l*n+m]=0;
      for (int i=0; i < x.size(); i++) {
	adat[l*n+m]=adat[l*n+m]+pow(x[i],l+m)*w[i];
      }
    }
  }
  matrix A(adat,n,n);


  matrix B(bdat,n,1);

  matrix Ai=matrix_invert(A);


  matrix C=Ai*B;

  double *dat=C.get_data();
  for (int i=0; i < n; i++) {
    coeff[i]=dat[i];
  }



  fit->set_coeffs(coeff);
  return *fit;
}

function& poly_fit(vector<double>& x, vector<double>& y, int order) {
  vector<double> w(x.size(),1);
  return poly_fit(x,y,w,order);
}
