#include "basin_package.h"
#include "stellardynamics.h"

#define ROTATE(a,i,j,k,l) g=a[i][j];h=a[k][l];a[i][j]=g-s*(h+g*tau);\
	a[k][l]=h+s*(g-h*tau);

using namespace std;



vector<double> center_of_mass(Attribute &x, Attribute &y, Attribute &z, Attribute &m)
{
  double x_sum = double(sum(x * m));
  double y_sum = double(sum(y * m));
  double z_sum = double(sum(z * m));
  double m_sum = sum(m);
  vector<double> com(3);
  com[0] = double(x_sum / m_sum);
  com[1] = double(y_sum / m_sum);
  com[2] = double(z_sum / m_sum);
  return com;
}

vector<double> center_of_mass_2d(Attribute &x, Attribute &y, Attribute &m)
{
  double x_sum = double(sum(x * m));
  double y_sum = double(sum(y * m));
  double m_sum = sum(m);
  vector<double> com(2);
  com[0] = double(x_sum / m_sum);
  com[1] = double(y_sum / m_sum);
  return com;
}

bool compare_vector (const vector<double>* a, const vector<double>* b)
{
  return ( (*a)[0]  < (*b)[0] );
}

//Pass local copies of x, y, z, and m
double compute_local_density(vector<double> &x, vector<double> &y, vector<double> &z, vector<double> &m, const int currPart, const int k)
{
  int npart = x.size();
    vector<vector<double> > temp(npart-1);
    //    vector<double> temp1(2);
    //temp1[0]=0.0;
    //temp1[1]=0.0;
  vector<vector<double>* > r(npart-1);
    int j = 0;
    for(int i = 0; i < npart; i++)
      {
	if(i != currPart)
	  {
	    double xtemp = x[i] - x[currPart];
	    double ytemp = y[i] - y[currPart];
	    double ztemp = z[i] - z[currPart];
	    temp[j].reserve(2);
	    temp[j][0]=(sqrt(xtemp * xtemp + ytemp * ytemp + ztemp * ztemp));
	    temp[j][1]=(m[i]);
	    //	    temp1[0]=(sqrt(xtemp * xtemp + ytemp * ytemp + ztemp * ztemp));
	    //	    temp1[1]=(m[i]);
	    r[j] = &(temp[j]);
	    j++;
	  }
      }
  
  nth_element(r.begin(), r.begin() + k - 1, r.end(), compare_vector);


  
  //compute local density
  double density = 0;
  for(int i = 0; i < k - 1; i++)
    density += (*r[i])[1];
  density *= (3./ (4. * M_PI * pow((*r[k - 1])[0], 3)));
  //  cerr<<density<<" DENSITY STL"<<endl;
  return density;
}



//Pass local copies of x, y, z, and m
double compute_local_density(double x[], double y[], double z[], double m[], const int currPart, const int npart, const int k, double MAX_DIST)
{
  double r2[npart];
  for(int i = 0; i < npart; i++)
  {
    double xtemp = x[i] - x[currPart];
    double ytemp = y[i] - y[currPart];
    double ztemp = z[i] - z[currPart];
    r2[i] = (xtemp * xtemp + ytemp * ytemp + ztemp * ztemp);
  }

  
  double near_r2[k];
  double near_mass[k];  
  for(int i = 0; i < k; i++)
    {
      near_r2[i] = MAX_DIST;
    }
  
  for(int i = 0; i < npart; i++)
  {
    if(i != currPart)
    {
      int j = k - 1;
      //find least distance farther than r2[i]
      while(r2[i] < near_r2[j] && j > -1)
        j--;
      //move up all distances farther than r2[i]
      for(int l = k-1; l > j+1; l--)
      {
        near_r2[l] = near_r2[l-1];
	near_mass[l] = near_mass[l-1];
      }
      //add r2[i] below all farther distances
      if(j < k-1)
      {
	near_r2[j+1] = r2[i];
	near_mass[j+1] = m[i];
      }
    }
  }


  double density = 0;
  for(int i = 0; i < k-1; i++)
    {
      density += near_mass[i];
    }
  density /= (4./3. * M_PI * pow(near_r2[k-1], 1.5));

  return density;
}

Attribute density_list(Attribute &x, Attribute &y, Attribute &z, Attribute &mass, int neighbors)
{
  int lo,hi,npart;
  npart = x.size();
  
  x.get_gen_distribution_index(BASIN_ID,&lo,&hi);
  double x_loc[npart], y_loc[npart], z_loc[npart], m_loc[npart], dens_list[hi-lo+1];
  Attribute dens_att = 0 * x;
  x.get(0, npart-1, x_loc);
  y.get(0, npart-1, y_loc);   
  z.get(0, npart-1, z_loc);
  mass.get(0, npart-1, m_loc);

  for(int i = lo; i <= hi; i++)
    {
      dens_list[i-lo] = compute_local_density(x_loc, y_loc, z_loc, m_loc, i, npart, neighbors);
    }

  dens_att.put(lo,hi,dens_list);
  MPI_Barrier(BASIN_COMM);
  
  return dens_att;  
}

Attribute density_list_vector(Attribute &x, Attribute &y, Attribute &z, Attribute &mass, int neighbors)
{
  int lo,hi,npart;
  npart = x.size();

  x.get_gen_distribution_index(BASIN_ID,&lo,&hi);
  vector<double> x_loc(npart), y_loc(npart), z_loc(npart), m_loc(npart), dens_list(hi-lo+1);
  Attribute dens_att = 0 * x;
  x.get(0, npart-1, &x_loc[0]);
  y.get(0, npart-1, &y_loc[0]);
  z.get(0, npart-1, &z_loc[0]);
  mass.get(0, npart-1, &m_loc[0]);

  for(int i = lo; i <= hi; i++)
    {
      dens_list[i-lo] = compute_local_density(x_loc, y_loc, z_loc, m_loc, i, neighbors);
    }

  dens_att.put(lo,hi,dens_list);
  MPI_Barrier(BASIN_COMM);

  return dens_att;
}

    
vector<double> density_center(Attribute &x, Attribute &y, Attribute &z, Attribute &dens_list)
{
  double rx, ry, rz;
  double tot_dens = sum(dens_list);
  rx = double(sum(dens_list * x)) / tot_dens;
  ry = double(sum(dens_list * y)) / tot_dens;
  rz = double(sum(dens_list * z)) / tot_dens;
  vector<double> center(3);
  center[0] = rx;
  center[1] = ry;
  center[2] = rz;
  return center;
}

vector<double> density_center_maxpos(Attribute &x, Attribute &y, Attribute &z, Attribute &dens_list)
{
  int lo,hi,npart;
  npart = x.size();
  
  x.get_gen_distribution_index(BASIN_ID,&lo,&hi);
  vector<double> x_loc(npart), y_loc(npart), z_loc(npart), dens_list_loc(npart);
  x.get(0, npart-1, &x_loc[0]);
  y.get(0, npart-1, &y_loc[0]);
  z.get(0, npart-1, &z_loc[0]);
  dens_list.get(0, npart-1, &dens_list_loc[0]);
  
  double max_density = -1;
  int max_density_index;
  
  for(int i = 0; i < npart; i++)
    {
      if(dens_list_loc[i] > max_density)
	{
	  if(sqrt(x_loc[i] * x_loc[i]* y_loc[i]* y_loc[i]* z_loc[i]* z_loc[i]) < 1)
	    {
	      max_density = dens_list_loc[i];
	      max_density_index = i;
	    }
	}
    }
  
  vector<double> center(3);
  center[0] = x_loc[max_density_index];
  center[1] = y_loc[max_density_index];
  center[2] = z_loc[max_density_index];
  return center;
}  
  

double core_radius(Attribute &x, Attribute &y, Attribute &z, Attribute &dens_list, vector<double> dens_center)
{ 
  Attribute xdiff = x - dens_center[0];
  Attribute ydiff = y - dens_center[1];
  Attribute zdiff = z - dens_center[2];
  Attribute rdiff = sqrt(xdiff * xdiff + ydiff * ydiff + zdiff * zdiff);

  return (double(sum(dens_list * rdiff)) / double(sum(dens_list)));
}


vector<vector<double> > radial_profile(Attribute &x, Attribute &y, Attribute &z, Attribute &A, int nbins, double rmin, double rmax, string scale)
{
  if(scale == "log")
    {
      if(rmin == 0.)
	{
	  if (BASIN_ID==0) cerr << "ERROR: Cannot compute log(0), Use rmin > 0" << endl;
	}
    }

  int lo,hi,npart, nprocs;
  npart = x.size();
  MPI_Comm_size(BASIN_COMM, &nprocs);
  x.get_gen_distribution_index(BASIN_ID,&lo,&hi);

  int npart_loc = hi-lo+1;

  vector<double> x_loc(npart_loc), y_loc(npart_loc), z_loc(npart_loc), r_loc(npart_loc), A_loc(npart_loc), bintot_loc(nbins, 0.), npart_bin_loc(nbins, 0.);
  x.get(lo, hi, &x_loc[0]);
  y.get(lo, hi, &y_loc[0]);
  z.get(lo, hi, &z_loc[0]);
  A.get(lo, hi, &A_loc[0]);

  
  for(int i = 0; i < (npart_loc) ; i++)
    r_loc[i] = sqrt(x_loc[i] * x_loc[i] + y_loc[i] * y_loc[i] + z_loc[i] * z_loc[i]);

  vector<double> rbins(nbins);
  vector<double> bintot(nbins, 0.);
  vector<double> npart_bin(nbins, 0.);
  vector<double> average(nbins);

  if(scale == "log")
    {
      double log_rmin = log10(rmin);
      double log_rmax = log10(rmax);
      double binsize = (log_rmax - log_rmin) / nbins;
      vector<double> loglimits(nbins+1);
      for(int i = 0; i <= nbins; i++)
	loglimits[i] = pow(10, log_rmin + i * binsize);

      bool binned;
      for(int j = 0; j < (npart_loc); j++)
	{
	  binned = false;
	  for(int i = 0; i < nbins && !binned; i++)
	    {
	      double lower = loglimits[i];
	      double upper = loglimits[i+1];
	      if(r_loc[j] >= lower && r_loc[j] < upper)
		{
		  npart_bin_loc[i]++;
		  bintot_loc[i] += A_loc[j];
		  binned = true;
		}
	    }
	}

      for(int i = 0; i < nbins; i++)
	rbins[i] = (loglimits[i] + loglimits[i+1]) / 2.;

    }
  
  else
    {
      double binsize = (rmax - rmin) / nbins;
      bool binned;
      for(int j = 0; j < (npart_loc); j++)
	{
	  binned = false;
	  for(int i = 0; i < nbins && !binned; i++)
	    {
	      double lower = rmin + i * binsize;
	      double upper = lower + binsize;
	      if(r_loc[j] >= lower && r_loc[j] < upper)
		{
		  npart_bin_loc[i]++;
		  bintot_loc[i] += A_loc[j];
		  binned = true;
		}
	    }
	}
      
      for(int i = 0; i < nbins; i++)
	rbins[i] = rmin + (i + .5) * binsize;
      
    }  

  MPI_Allreduce(&bintot_loc[0], &bintot[0], nbins, MPI_DOUBLE, MPI_SUM, BASIN_COMM);
  MPI_Allreduce(&npart_bin_loc[0], &npart_bin[0], nbins, MPI_DOUBLE, MPI_SUM, BASIN_COMM);
  
  for(int i = 0; i < nbins; i++)
    {
      if(npart_bin[i] == 0)
	average[i] = 0.;
      else
	average[i] = bintot[i] / npart_bin[i];
    }
  vector<vector<double> > data(2);
  data[0] = rbins;
  data[1] = average;

  return data;
}


vector<vector<double> > radial_profile(Attribute &x, Attribute &y, Attribute &z, Attribute &A, Attribute &filter, int nbins, double rmin, double rmax, vector<double> filter_range, string scale)
{
  if(scale == "log")
    {
      if(rmin == 0.)
	{
	  if (BASIN_ID==0) cerr << "ERROR: Cannot compute log(0), Use rmin > 0" << endl;
	}
    }

  int lo,hi,npart, nprocs;
  npart = x.size();
  MPI_Comm_size(BASIN_COMM, &nprocs);
  x.get_gen_distribution_index(BASIN_ID,&lo,&hi);

  int npart_loc = hi-lo+1;

  int nfilter_comp = filter_range.size() - 1;
  int frange_size = filter_range.size();

  vector<double> x_loc(npart_loc), y_loc(npart_loc), z_loc(npart_loc), r_loc(npart_loc), A_loc(npart_loc), f_loc(npart_loc), temp(nbins, 0.);
  vector<vector<double> > bintot_loc(nfilter_comp, temp), npart_bin_loc(nfilter_comp, temp);
  x.get(lo, hi, &x_loc[0]);
  y.get(lo, hi, &y_loc[0]);
  z.get(lo, hi, &z_loc[0]);
  filter.get(lo, hi, &f_loc[0]);
  A.get(lo, hi, &A_loc[0]);

  
  for(int i = 0; i < (npart_loc) ; i++)
    r_loc[i] = sqrt(x_loc[i] * x_loc[i] + y_loc[i] * y_loc[i] + z_loc[i] * z_loc[i]);

  vector<double> rbins(nbins);
  vector<vector<double> > bintot(nfilter_comp, temp);
  vector<vector<double> > npart_bin(nfilter_comp, temp);
  vector<vector<double> > average(nfilter_comp, temp);


  if(scale == "log")
    {
      double log_rmin = log10(rmin);
      double log_rmax = log10(rmax);
      double binsize = (log_rmax - log_rmin) / nbins;

      vector<double> loglimits(nbins+1);
      for(int i = 0; i <= nbins; i++)
	loglimits[i] = pow(10, log_rmin + i * binsize);

      bool binned;
      for(int j = 0; j < (npart_loc); j++)
	{
	  binned = false;
	  for(int i = 0; i < nbins && !binned; i++)
	    {
	      for(int k = 0; k < nfilter_comp; k++)
		{
		  if(f_loc[j] >= filter_range[k] && f_loc[j] < filter_range[k+1])
		    {
		      double lower = loglimits[i];
		      double upper = loglimits[i+1];
		      if(r_loc[j] >= lower && r_loc[j] < upper)
			{
			  npart_bin_loc[k][i]++;
			  bintot_loc[k][i] += A_loc[j];
			  binned = true;
			}
		    }
		}
	    }
	}

      for(int i = 0; i < nbins; i++)
	rbins[i] = (loglimits[i] + loglimits[i+1]) / 2.;

    }
  
  else
    {
      double binsize = (rmax - rmin) / nbins;
      bool binned;
      for(int j = 0; j < (npart_loc); j++)
	{
	  binned = false;
	  for(int i = 0; i < nbins && !binned; i++)
	    {
	      for(int k = 0; k < nfilter_comp; k++)
		{
		  if(f_loc[j] >= filter_range[k] && f_loc[j] < filter_range[k+1])
		    {
		      double lower = rmin + i * binsize;
		      double upper = lower + binsize;
		      if(r_loc[j] >= lower && r_loc[j] < upper)
			{
			  npart_bin_loc[k][i]++;
			  bintot_loc[k][i] += A_loc[j];
			  binned = true;
			}
		    }
		}
	    }
	}

      for(int i = 0; i < nbins; i++)
	rbins[i] = rmin + (i + .5) * binsize;
      
    }

  for(int i = 0; i < nfilter_comp; i++)
    {
      MPI_Allreduce(&bintot_loc[i][0], &bintot[i][0], nbins, MPI_DOUBLE, MPI_SUM, BASIN_COMM);
      MPI_Allreduce(&npart_bin_loc[i][0], &npart_bin[i][0], nbins, MPI_DOUBLE, MPI_SUM, BASIN_COMM);
    }
  
  for(int i = 0; i < nfilter_comp; i++)
    {
      for(int j = 0; j < nbins; j++)
	{
	  if(npart_bin[i][j] == 0)
	    average[i][j] = 0;
	  else
	    average[i][j] = bintot[i][j] / npart_bin[i][j];
	}
    }
  
  vector<vector<double> > data(1+nfilter_comp);
  data[0] = rbins;
  for(int i = 0; i < nfilter_comp; i++)
    data[i+1] = average[i];

  return data;
}


vector<vector<double> > density_profile(Attribute &x, Attribute &y, Attribute &z, Attribute &mass, int nbins, double rmin, double rmax, string compute, string scale)
{
  if(scale == "log")
    {
      if(rmin == 0.)
	{
	  if (BASIN_ID==0) cerr << "ERROR: Cannot compute log(0), Use rmin > 0" << endl;
	}
    }
  bool mass_bool, number, both;
  mass_bool = number = both = false;
  if(compute == "mn")
    {
      mass_bool = number = true;
      both = true;
    }
  else if(compute == "mass")
    mass_bool = true;
  else if(compute == "number")
    number = true;
  int lo,hi,npart, nprocs;
  npart = x.size();
  MPI_Comm_size(BASIN_COMM, &nprocs);
  x.get_gen_distribution_index(BASIN_ID,&lo,&hi);

  int npart_loc = hi-lo+1;

  vector<double> x_loc(npart_loc), y_loc(npart_loc), z_loc(npart_loc), r_loc(npart_loc), m_loc(npart_loc);
  vector<double> bintot_loc(nbins, 0.), binvol(nbins, 0.), npart_bin_loc(nbins, 0.);
  x.get(lo, hi, &x_loc[0]);
  y.get(lo, hi, &y_loc[0]);
  z.get(lo, hi, &z_loc[0]);
  mass.get(lo, hi, &m_loc[0]);

  for(int i = 0; i < (npart_loc); i++)
    r_loc[i] = sqrt(x_loc[i] * x_loc[i] + y_loc[i] * y_loc[i] + z_loc[i] * z_loc[i]);

  vector<double> rbins(nbins);
  vector<double> bintot(nbins, 0.);
  vector<double> average_m(nbins,0.);
  vector<double> npart_bin(nbins, 0.);
  vector<double> average_n(nbins, 0.);

  if(scale == "log")
    {

      double log_rmin = log10(rmin);
      double log_rmax = log10(rmax);
      double binsize = (log_rmax - log_rmin) / nbins;
      vector<double> loglimits(nbins+1);
      for(int i = 0; i <= nbins; i++)
	loglimits[i] = pow(10, log_rmin + i * binsize);
      
      bool binned;
      for(int i = 0; i < (npart_loc); i++)
	{
	  binned = false;
	  for(int j = 0; j < nbins && !binned; j++)
	    {
	      double lower = loglimits[j];
	      double upper = loglimits[j+1];
	      if(r_loc[i] >= lower && r_loc[i] < upper)
		{
		  if(number)
		    npart_bin_loc[j]++;
		  if(mass_bool)
		    bintot_loc[j] += m_loc[i];
		  binned = true;
		}
	    }
	}

      for(int i = 0; i < nbins; i++)
	{
	  double v0 = pow(loglimits[i], 3.);
	  double v1 = pow(loglimits[i+1], 3.);
	  binvol[i] = 4./3. * M_PI * (v1 - v0);
	}
      
      for(int i = 0; i < nbins; i++)
	rbins[i] = (loglimits[i] + loglimits[i+1]) / 2.;
    }

  
  else
    {
      double binsize = (rmax - rmin) / nbins;
      bool binned;
      for(int i = 0; i < (npart_loc); i++)
	{
	  binned = false;
	  for(int j = 0; j < nbins && !binned; j++)
	    {
	      double lower = rmin + j * binsize;
	      double upper = lower + binsize;
	      if(r_loc[i] >= lower && r_loc[i] < upper)
		{
		  if(number)
		    npart_bin_loc[j]++;
		  if(mass_bool)
		    bintot_loc[j] += m_loc[i];
		  binned = true;
		}
	    }
	}
      
      for(int i = 0; i < nbins; i++)
	{
	  rbins[i] = rmin + (i + .5) * binsize;
	  double v0 = pow(rbins[i] - .5 * binsize, 3.);
	  double v1 = pow(rbins[i] + .5 * binsize, 3.);
	  binvol[i] = 4./3. * M_PI * (v1 - v0);
	}

    }


  if(mass_bool)
    {
      MPI_Allreduce(&bintot_loc[0], &bintot[0], nbins, MPI_DOUBLE, MPI_SUM, BASIN_COMM);
      for(int i = 0; i < nbins; i++)
	  average_m[i] = bintot[i] / binvol[i];
    }
  
  if(number)
    {
      MPI_Allreduce(&npart_bin_loc[0], &npart_bin[0], nbins, MPI_DOUBLE, MPI_SUM, BASIN_COMM);
      for(int i = 0; i < nbins; i++)
	average_n[i] = npart_bin[i] / binvol[i];
    }

  vector<vector<double> > data(3);
  
  if(both)
    {
      data[0] = rbins;
      data[1] = average_m;
      data[2] = average_n;
    }
  
  else if(mass_bool)
    {
      data[0] = rbins;
      data[1] = average_m;
    }
  
  else if(number)
    {
      data[0] = rbins;
      data[1] = average_n;
    }

  return data;
}




vector<vector<double> > density_profile(Attribute &x, Attribute &y, Attribute &z, Attribute &mass, Attribute &filter, int nbins, double rmin, double rmax, vector<double> filter_range, string compute, string scale)
{
  if(scale == "log")
    {
      if(rmin == 0.)
	{
	  if (BASIN_ID==0) cerr << "ERROR: Cannot compute log(0), Use rmin > 0" << endl;
	}
    }
  
  bool mass_bool, number, both;
  mass_bool = number = both = false;
  if(compute == "mn")
    mass_bool = number = both = true;
  else if(compute == "mass")
    mass_bool = true;
  else if(compute == "number")
    number = true;
  
  int lo,hi,npart, nprocs;
  npart = x.size();
  MPI_Comm_size(BASIN_COMM, &nprocs);
  x.get_gen_distribution_index(BASIN_ID,&lo,&hi);

  int npart_loc = hi-lo+1;

  int nfilter_comp = filter_range.size() - 1;
  int frange_size = filter_range.size();
  
  vector<double> x_loc(npart_loc), y_loc(npart_loc), z_loc(npart_loc), r_loc(npart_loc), m_loc(npart_loc), f_loc(npart_loc), binvol(nbins, 0.);
  vector<vector<double> > bintot_loc(nfilter_comp, binvol), npart_bin_loc(nfilter_comp, binvol);
  vector<vector<double> > bintot(nfilter_comp, binvol), average_m(nfilter_comp, binvol), npart_bin(nfilter_comp, binvol), average_n(nfilter_comp, binvol);
  x.get(lo, hi, &x_loc[0]);
  y.get(lo, hi, &y_loc[0]);
  z.get(lo, hi, &z_loc[0]);
  mass.get(lo, hi, &m_loc[0]);
  filter.get(lo, hi, &f_loc[0]);


  for(int i = 0; i < (npart_loc); i++)
    r_loc[i] = sqrt(x_loc[i] * x_loc[i] + y_loc[i] * y_loc[i] + z_loc[i] * z_loc[i]);
  

  vector<double> rbins(nbins);
 
  if(scale == "log")
    {
      double log_rmin = log10(rmin);
      double log_rmax = log10(rmax);
      double binsize = (log_rmax - log_rmin) / nbins;
      vector<double> loglimits(nbins+1);
      for(int i = 0; i <= nbins; i++)
	loglimits[i] = pow(10, log_rmin + i * binsize);
      

      bool binned;
      for(int j = 0; j < (npart_loc); j++)
	{
	  binned = false;
	  for(int i = 0; i < nbins && !binned; i++)
	    {
	      double lower = loglimits[i];
	      double upper = loglimits[i+1];	      
	      for(int k = 0; k < nfilter_comp; k++)
		{
		  if(f_loc[j] >= filter_range[k] && f_loc[j] < filter_range[k+1])
		    {
		      if(r_loc[j] >= lower && r_loc[j] < upper)
			{
			  if(number)
			    npart_bin_loc[k][i]++;
			  if(mass_bool)
			    bintot_loc[k][i] += m_loc[j];
			  binned = true;
			}
		    }
		}
	    }
	}

      for(int i = 0; i < nbins; i++)
	{
	  double v0 = pow(loglimits[i], 3.);
	  double v1 = pow(loglimits[i+1], 3.);
	  binvol[i] = 4./3. * M_PI * (v1 - v0);
	  rbins[i] = (loglimits[i] + loglimits[i+1]) / 2.;
	}
      
    }
  
  else
    {
      bool binned;
      double binsize = (rmax - rmin) / nbins;
      for(int i = 0; i < npart_loc; i++)
	{
	  binned = false;
	  for(int j = 0; j < nbins && !binned; j++)
	    {
	      double lower = rmin + j * binsize;
	      double upper = lower + binsize;
	      for(int k = 0; k < nfilter_comp; k++)
		{
		  if(f_loc[i] >= filter_range[k] && f_loc[i] < filter_range[k+1])
		    {
		      if(r_loc[i] >= lower && r_loc[i] < upper)
			{
			  if(number)
			    npart_bin_loc[k][j]++;
			  if(mass_bool)
			    bintot_loc[k][j] += m_loc[i];
			  binned = true;
			}
		    }
		}
	    }
	}
     
      for(int i = 0; i < nbins; i++)
	{
	  rbins[i] = rmin + (i + .5) * binsize;
	  double v0 = pow(rbins[i] - .5 * binsize, 3.);
	  double v1 = pow(rbins[i] + .5 * binsize, 3.);
	  binvol[i] = 4./3. * M_PI * (v1 - v0);
	}

    }
  

  
  if(mass_bool)
    {
      for(int i = 0; i < nfilter_comp; i++)
      {
	MPI_Allreduce(&bintot_loc[i][0], &bintot[i][0], nbins, MPI_DOUBLE, MPI_SUM, BASIN_COMM);
	for(int j = 0; j < nbins; j++)
	  average_m[i][j] = bintot[i][j] / binvol[j];
      }
    }
  
  if(number)
    {
      for(int i = 0; i < nfilter_comp; i++)
	{
	  MPI_Allreduce(&npart_bin_loc[i][0], &npart_bin[i][0], nbins, MPI_DOUBLE, MPI_SUM, BASIN_COMM);
	  for(int j = 0; j < nbins; j++)
	    average_n[i][j] = npart_bin[i][j] / binvol[j];
	}
    }

  vector<vector<double> > data(1 + 2 * nfilter_comp);
  for(int i = 0; i < (1 + 2 * nfilter_comp); i++)
    data[i].reserve(nbins);
  data[0] = rbins;
  
  if(both)
    {
      for(int i = 0; i < nfilter_comp; i++)
	  data[i+1] = average_m[i];
      for(int i = 0; i < nfilter_comp; i++)
	  data[i+nfilter_comp+1] = average_n[i];      
    }  
  else if(mass_bool)
    {
      for(int i = 0; i < nfilter_comp; i++)
	data[i+1] = average_m[i];
    }
  
  else if(number)
    {
    for(int i = 0; i < nfilter_comp; i++)
       data[i+1] = average_n[i];
    }
  
  return data;
}


vector<vector<double> > disp_profile(Attribute &x, Attribute &y, Attribute &z, Attribute &A, int nbins, double rmin, double rmax, string scale)
{
  if(scale == "log")
    {
      if(rmin == 0.)
	{
	  if (BASIN_ID==0) cerr << "ERROR: Cannot compute log(0), Use rmin > 0" << endl;
	}
    }
  
  Attribute A_sq = A * A;
  vector<vector<double> > A_prof = radial_profile(x, y, z, A, nbins, rmin, rmax, scale);
  vector<vector<double> > A_sq_prof = radial_profile(x, y, z, A_sq, nbins, rmin, rmax, scale);
  vector<double> A_prof_sq(nbins), sigma_sq(nbins), sigma(nbins);
  for(int i = 0; i < nbins; i++)
    {
      A_prof_sq[i] = A_prof[1][i] * A_prof[1][i];
      sigma_sq[i] = A_sq_prof[1][i] - A_prof_sq[i];
      sigma[i] = sqrt(sigma_sq[i]);
    }

 
  vector<vector<double> > data(2);
  data[0] = A_prof[0];
  data[1] = sigma;
  return data;
}

vector<vector<double> > disp_profile(Attribute &x, Attribute &y, Attribute &z, Attribute &A, Attribute &filter, int nbins, double rmin, double rmax, vector<double> filter_range, string scale)
{
  if(scale == "log")
    {
      if(rmin == 0.)
	{
	  if (BASIN_ID==0) cerr << "ERROR: Cannot compute log(0), Use rmin > 0" << endl;
	}
    }
  
  int nfilter_comp = filter_range.size() - 1;
  Attribute A_sq = A * A;
  vector<vector<double> > A_prof = radial_profile(x, y, z, A, filter, nbins, rmin, rmax, filter_range, scale);
  vector<vector<double> > A_sq_prof = radial_profile(x, y, z, A_sq, filter, nbins, rmin, rmax, filter_range, scale);
  vector<double> temp(nbins, 0.);
  vector<vector<double> > sigma_sq(nfilter_comp, temp), sigma(nfilter_comp, temp), A_prof_sq(nfilter_comp, temp);

  for(int i = 0; i < nfilter_comp; i++)
    {
      for(int j = 0; j < nbins; j++)
	{
	  A_prof_sq[i][j] = A_prof[i+1][j] * A_prof[i+1][j];
	}
    }
  for(int i = 0; i < nfilter_comp; i++)
    {
      for(int j = 0; j < nbins; j++)
	{
	  sigma_sq[i][j] = A_sq_prof[i+1][j] - A_prof_sq[i][j];
	  sigma[i][j] = sqrt(sigma_sq[i][j]);
	}
    }
  vector<vector<double> > data(1+nfilter_comp);
  data[0] = A_prof[0];
  for(int i = 0; i < nfilter_comp; i++)
    data[i+1] = sigma[i];

  return data;

}

//if radii[i] < 0, results[i] = -1. and if radii[i] == 0., results[i] = 0.
vector<double> lag_rad(Attribute &x, Attribute &y, Attribute &z, Attribute &mass, vector<double> radii)
{
  int npart = x.size();
  int nrad = radii.size();
  Attribute r = sqrt(x * x + y * y + z * z);
  Attribute r_index = index_sort(r);
  vector<double> results(nrad, -1.);
  double mtot = sum(mass);
  if(BASIN_ID == 0)
    {
      vector<double> r_loc(npart), m_loc(npart);
      vector<int> r_index_loc(npart);
      r.get(0, npart-1, &r_loc[0]);
      mass.get(0, npart-1, &m_loc[0]);
      r_index.get(0, npart-1, &r_index_loc[0]);
      double r_curr, r_prev, m_curr, m_prev;
      int currPart;
      for(int i = 0; i < nrad; i++)
	{
	  m_prev = m_curr = 0;
	  double mtest = radii[i] * mtot;
	  for(int j = 0; !(m_prev <= mtest && m_curr >= mtest) && radii[i] > 0; j++)
	    {
	      currPart = r_index_loc[j];
	      m_prev = m_curr;
	      r_prev = r_curr;
	      m_curr += m_loc[currPart];
	      r_curr = r_loc[currPart];
	    }
	  if(radii[i] > 0)
	    results[i] =  r_prev + (mtest - m_prev) * (r_curr - r_prev) / (m_curr - m_prev);
	  else if(radii[i] == 0)
	    results[i] = 0.;
	}
    }
  
  MPI_Bcast(&results[0], nrad, MPI_DOUBLE, 0, BASIN_COMM);
  MPI_Barrier(BASIN_COMM);
  
  return results;
				
}

//if radii[i] < 0, results[i] = -1. and if radii[i] == 0., results[i] = 0.
vector<vector<double> > lag_rad(Attribute &x, Attribute &y, Attribute &z, Attribute &mass, Attribute &filter, vector<double> radii, vector<double> filter_range)
{
  int npart = x.size();
  int nrad = radii.size();
  int frange_size = filter_range.size();
  int nfilter_comp = frange_size - 1;

  Attribute r = sqrt(x * x + y * y + z * z);
  Attribute r_index = index_sort(r);
  Attribute temp, tempMass;
  vector<double> tempvec(nrad, -1.);
  vector<vector<double> > results(nfilter_comp, tempvec);
  double mtot;
  //if(BASIN_ID == 0)
  //{
  //cerr << "TEST" << endl;
  vector<double> r_loc(npart), m_loc(npart), f_loc(npart);
  vector<int> r_index_loc(npart);
  r.get(0, npart-1, &r_loc[0]);
  mass.get(0, npart-1, &m_loc[0]);
  r_index.get(0, npart-1, &r_index_loc[0]);
  filter.get(0, npart-1, &f_loc[0]);
  double r_curr, r_prev, m_curr, m_prev;
  int currPart;
  for(int k = 0; k < nfilter_comp; k++) {
    temp = where(filter >= filter_range[k],mass,0.0);
    tempMass = where(temp < filter_range[k+1],temp,0.0);
    mtot = sum(tempMass);
    tempMass.get(0, npart-1, &m_loc[0]);
    
    if(BASIN_ID == 0) {
      for(int i = 0; i < nrad; i++)
	{
	  m_prev = m_curr = 0;
	  double mtest = radii[i] * mtot;
	  for(int j = 0; !(m_prev <= mtest && m_curr >= mtest) && radii[i] > 0; j++)
	    {
	      //cerr << "TEST" << endl;
	      currPart = r_index_loc[j];
	      m_prev = m_curr;
	      r_prev = r_curr;
	      m_curr += m_loc[currPart];
	      r_curr = r_loc[currPart];
	    }
	  if(radii[i] > 0)
	    results[k][i] =  r_prev + (mtest - m_prev) * (r_curr - r_prev) / (m_curr - m_prev);
	  else if(radii[i] == 0)
	    results[k][i] = 0.;
	}	
    }  
  }
    //}
  
  for(int i = 0; i < nfilter_comp; i++)
    MPI_Bcast(&results[i][0], nrad, MPI_DOUBLE, 0, BASIN_COMM);
  MPI_Barrier(BASIN_COMM);
  
  return results;
				
}

//Make (n+1) x (n+1) copy of nxn matrix a and put all values in indices 1..n, 1..n (ignore first row and column)
void jacobi(vector<vector<double> > &a, int n,vector<double> &d,vector<vector<double> > &v,int *nrot)
  //double a[4][4],d[4],v[4][4];
  //int n, *nrot;
{
  int j,iq,ip,i;
  double tresh,theta,tau,t,sm,s,h,g,c ;
  double b[4],z[4] ;
  
  for (ip=1;ip<=n;ip++) {
    for (iq=1;iq<=n;iq++) v[ip][iq]=0.0;
    v[ip][ip]=1.0;
  }
  for (ip=1;ip<=n;ip++) {
    b[ip]=d[ip]=a[ip][ip];
    z[ip]=0.0;
  }
  *nrot=0;
  for (i=1;i<=100;i++) {
    sm=0.0;
    for (ip=1;ip<=n-1;ip++) {
      for (iq=ip+1;iq<=n;iq++)
	sm += fabs(a[ip][iq]);
    }
    if (sm <= 1.0e-12) {
      return;
    }
    if (i < 4)
      tresh=0.2*sm/(n*n);
    else
      tresh=0.0;
    for (ip=1;ip<=n-1;ip++) {
      for (iq=ip+1;iq<=n;iq++) {
	g=100.0*fabs(a[ip][iq]);
	if (i > 4 && fabs(d[ip])+g == fabs(d[ip]) && fabs(d[iq])+g == fabs(d[iq]))
	  a[ip][iq]=0.0;
	else if (fabs(a[ip][iq]) > tresh) {
	  h=d[iq]-d[ip];
	  if (fabs(h)+g == fabs(h))
	    t=(a[ip][iq])/h;
	  else {
	    theta=0.5*h/(a[ip][iq]);
	    t=1.0/(fabs(theta)+
		   sqrt(1.0+theta*theta));
	    if (theta < 0.0) t = -t;
	  }
	  c=1.0/sqrt(1+t*t);
	  s=t*c;
	  tau=s/(1.0+c);
	  h=t*a[ip][iq];
	  z[ip] -= h;
	  z[iq] += h;
	  d[ip] -= h;
	  d[iq] += h;
	  a[ip][iq]=0.0;
	  for (j=1;j<=ip-1;j++) {
	    ROTATE(a,j,ip,j,iq)
	      }
	  for (j=ip+1;j<=iq-1;j++) {
	    ROTATE(a,ip,j,j,iq)
	      }
	  for (j=iq+1;j<=n;j++) {
	    ROTATE(a,ip,j,iq,j)
	      }
	  for (j=1;j<=n;j++) {
	    ROTATE(v,j,ip,j,iq)
	      }
	  ++(*nrot);
	}
      }
    }
    for (ip=1;ip<=n;ip++) {
      b[ip] += z[ip];
      d[ip]=b[ip];
      z[ip]=0.0;
    }
  }
  if (BASIN_ID==0) cerr << "JACOBI FUNCTION ERROR " << endl;
}


void shape(vector<double> &x, vector<double> &y, vector<double> &z, vector<double> &m_loc, double &ba, double &ca, double d)
{
  vector<double> temp1(4, 0.);
  vector<vector<double> > inertia_input(4, temp1);
  vector<vector<double> > evec_input = inertia_input;
  vector<double> evalues(3, 0.);
  vector<vector<double> > evectors(3, evalues);
  vector<vector<double> > inertia = evectors;
  int nrot, ia, ib, ic;

  vector<double> x_copy, y_copy, z_copy;
  int npart_loc = x.size();
  x_copy.reserve(npart_loc);
  y_copy.reserve(npart_loc);
  z_copy.reserve(npart_loc);
  double sum00_loc = 0., sum11_loc  = 0., sum22_loc  = 0., sum01_loc  = 0., sum02_loc  = 0., sum12_loc  = 0., delta;
  double sum00, sum11, sum22, sum01, sum02, sum12;
  double ba_old, ca_old;

  vector<bool> include(npart_loc, true);

  int niter = 0;
  do{
    sum00_loc = sum11_loc = sum22_loc = sum01_loc = sum02_loc = sum12_loc = 0;
    for(int i = 0; i < npart_loc; i++)
      {
	if(include[i])
	  {
	    if(niter == 0) delta = sqrt(x[i] * x[i] + y[i] * y[i] + z[i] * z[i]);
	    else delta = x[i] * x[i] + y[i] * y[i] / ba / ba + z[i] * z[i] / ca / ca;
	    if(delta <= d)
	      {
		sum00_loc += m_loc[i] * (y[i] * y[i] + z[i] * z[i]);
		sum11_loc += m_loc[i] * (x[i] * x[i] + z[i] * z[i]);
		sum22_loc += m_loc[i] * (x[i] * x[i] + y[i] * y[i]);
		sum01_loc -= m_loc[i] * x[i] * y[i];
		sum02_loc -= m_loc[i] * x[i] * z[i];
		sum12_loc -= m_loc[i] * y[i] * z[i];
	      }
	    else
	      include[i] = false;
	  }
      }
    
    MPI_Allreduce(&sum00_loc, &sum00, 1, MPI_DOUBLE, MPI_SUM, BASIN_COMM);
    MPI_Allreduce(&sum11_loc, &sum11, 1, MPI_DOUBLE, MPI_SUM, BASIN_COMM);
    MPI_Allreduce(&sum22_loc, &sum22, 1, MPI_DOUBLE, MPI_SUM, BASIN_COMM);
    MPI_Allreduce(&sum01_loc, &sum01, 1, MPI_DOUBLE, MPI_SUM, BASIN_COMM);
    MPI_Allreduce(&sum02_loc, &sum02, 1, MPI_DOUBLE, MPI_SUM, BASIN_COMM);
    MPI_Allreduce(&sum12_loc, &sum12, 1, MPI_DOUBLE, MPI_SUM, BASIN_COMM);

    inertia[0][0] = sum00;
    inertia[1][1] = sum11;
    inertia[2][2] = sum22;
    inertia[0][1] = inertia[1][0] = sum01;
    inertia[0][2] = inertia[2][0] = sum02;
    inertia[1][2] = inertia[2][1] = sum12;
  
    //Use Jacobi function
    for(int k = 1; k < 4; k++)
      for(int j = 1; j < 4; j++)
	inertia_input[k][j] = inertia[k-1][j-1];
    jacobi(inertia_input, 3, temp1, evec_input, &nrot);
    
    //Move values to n=3 vector and matrix
    for(int k = 0; k < 3; k++)
      {
	evalues[k] = temp1[k+1];
	for(int j = 0; j < 3; j++)
	  evectors[k][j] = evec_input[k+1][j+1];
      }

    if(evalues[0] >= evalues[1] && evalues[0] >= evalues[2])
      {
	ia = 0;
	if(evalues[1] >= evalues[2])
	  {
	    ib = 1;
	    ic = 2;
	  }
	else
	  {
	    ib = 2;
	    ic = 1;
	  }
      }
    
    else if(evalues[1] >= evalues[0] && evalues[1] >= evalues[2])
      {
	ia = 1;
	if(evalues[0] >= evalues[2])
	  {
	    ib = 0;
	    ic = 2;
	  }
	else
	  {
	    ib = 2;
	    ic = 0;
	  }
      }
    
    else if(evalues[2] >= evalues[0] && evalues[2] >= evalues[1])
      {
	ia = 2;
	if(evalues[0] >= evalues[1])
	  {
	    ib = 0;
	    ic = 1;
	  }
	else
	  {
	    ib = 1;
	    ic = 0;
	  }
      }

    ba_old = ba;
    ca_old = ca;

    ba = sqrt(evalues[ib] / evalues[ia]);
    ca = sqrt(evalues[ic] / evalues[ia]);
    niter++;
  }while((fabs(ba - ba_old) / ba_old > 1.e-4 || fabs(ca - ca_old) / ca_old > 1.e-4) && niter < 20);
  //bout << niter << endl;
}

  


vector<vector<double> > shape_profile(Attribute &x, Attribute &y, Attribute &z, Attribute &mass, int nbins, double rmin, double rmax, string scale)
{
  if(scale == "log")
    {
      if(rmin == 0.)
	{
	  if (BASIN_ID==0) cerr << "ERROR: Cannot compute log(0), Use rmin > 0" << endl;
	}
    }

  vector<double> rlimits(nbins+1);
  vector<double> rbins(nbins);
  if(scale == "log")
    {
      double binsize = (log10(rmax) - log10(rmin)) / nbins;
      for(int i = 0; i <= nbins; i++)
	rlimits[i] = pow(10, log10(rmin) + i * binsize);
    }
  else
    {
      double binsize = (rmax - rmin) / nbins;
      for(int i = 0; i <= nbins; i++)
	rlimits[i] = rmin + i * binsize;      
    }
  for(int i = 0; i < nbins; i++)
    rbins[i] = (rlimits[i] + rlimits[i+1]) / 2.;
	 
  int lo,hi,npart, nprocs, my_id;
  npart = x.size();
  MPI_Comm_size(BASIN_COMM, &nprocs);
  MPI_Comm_rank(BASIN_COMM, &my_id);
  x.get_gen_distribution_index(my_id,&lo,&hi);
  int npart_loc = hi-lo+1;

  double ba, ca;
  vector<double> data1(nbins), data2(nbins);
  vector<double> x_loc(npart_loc), y_loc(npart_loc), z_loc(npart_loc), m_loc(npart_loc);
  x.get(lo, hi, &x_loc[0]);
  y.get(lo, hi, &y_loc[0]);
  z.get(lo, hi, &z_loc[0]);
  mass.get(lo, hi, &m_loc[0]);
  
  for(int i = 0; i < nbins; i++)
    {
      shape(x_loc, y_loc, z_loc, m_loc, ba, ca, rlimits[i+1]);
      data1[i] = ba;
      data2[i] = ca;
    }
   
  
  vector<vector<double> > data(3);
  data[0] = rbins;
  data[1] = data1;
  data[2] = data2;
  return data;  
}

vector<vector<double> > shape_profile(Attribute &x, Attribute &y, Attribute &z, Attribute &mass, Attribute &filter, int nbins, double rmin, double rmax, vector<double> filter_range, string scale)
{
  if(scale == "log")
    {
      if(rmin == 0.)
	{
	  if (BASIN_ID==0) cerr << "ERROR: Cannot compute log(0), Use rmin > 0" << endl;
	}
    }
  vector<double> rlimits(nbins+1);
  vector<double> rbins(nbins);
  if(scale == "log")
    {
      double binsize = (log10(rmax) - log10(rmin)) / nbins;
      for(int i = 0; i <= nbins; i++)
	rlimits[i] = pow(10, log10(rmin) + i * binsize);
    }
  else
    {
      double binsize = (rmax - rmin) / nbins;
      for(int i = 0; i <= nbins; i++)
	rlimits[i] = rmin + i * binsize;      
    }
  for(int i = 0; i < nbins; i++)
    rbins[i] = (rlimits[i] + rlimits[i+1]) / 2.;
	 
  int lo,hi,npart, nprocs, my_id;
  npart = x.size();
  MPI_Comm_size(BASIN_COMM, &nprocs);
  MPI_Comm_rank(BASIN_COMM, &my_id);
  x.get_gen_distribution_index(my_id,&lo,&hi);
  int npart_loc = hi-lo+1;

  int frange_size = filter_range.size();
  int nfilter_comp = frange_size - 1;

  double ba, ca;
  vector<vector<double> > data1(nfilter_comp), data2(nfilter_comp);
  vector<double> x_loc(npart_loc), y_loc(npart_loc), z_loc(npart_loc), m_loc(npart_loc), f_loc(npart_loc);
  x.get(lo, hi, &x_loc[0]);
  y.get(lo, hi, &y_loc[0]);
  z.get(lo, hi, &z_loc[0]);
  mass.get(lo, hi, &m_loc[0]);
  filter.get(lo, hi, &f_loc[0]);

  vector<vector<double> > xgroup(nfilter_comp), ygroup(nfilter_comp), zgroup(nfilter_comp), mgroup(nfilter_comp);
  for(int i = 0; i < nfilter_comp; i++)
    {
      //data1[i].reserve(nbins);
      //data2[i].reserve(nbins);
      data1[i] = vector<double> (nbins);
      data2[i] = vector<double> (nbins);
      xgroup[i].reserve(npart_loc);
      ygroup[i].reserve(npart_loc);
      zgroup[i].reserve(npart_loc);
      mgroup[i].reserve(npart_loc);
    }
  
  bool binned;
  for(int j = 0; j < npart_loc; j++)
    {
      binned = false;
      for(int i = 0; i < nfilter_comp && !binned; i++)
	{
	  if(f_loc[j] >= filter_range[i] && f_loc[j] < filter_range[i+1])
	    {
	      xgroup[i].push_back(x_loc[j]);
	      ygroup[i].push_back(y_loc[j]);
	      zgroup[i].push_back(z_loc[j]);
	      mgroup[i].push_back(m_loc[j]);
	      binned = true;
	    }
	}
    }

  int count = 0;
  for(int i = 0; i < nfilter_comp; i++)
    {
      for(int j = 0; j < nbins; j++)
	{
	  shape(xgroup[i], ygroup[i], zgroup[i], mgroup[i], ba, ca, rlimits[j+1]);
	  data1[i][j] = ba;
	  data2[i][j] = ca;
	}
    }  

  vector<vector<double> > data(1 + 2 * nfilter_comp);
  data[0] = rbins;
  
  for(int i = 0; i < nfilter_comp; i += 2)
    {
      data[i+1] = data1[i];
      data[i+2] = data2[i];
    }
  return data;  
}

