#include "basin_package.h"
#include "transforms.h"

ConvKernel::ConvKernel() {
  // By default, we're creating a 2-d Gaussian which is 5x5.  Why?  Why not?
  ndims=2;
  nx=5;
  double sig=1.;
  double total=0;
  double xc=(nx-1.)/2.;
  elements.resize(nx*nx);
  for (int i=0; i< nx; i++) {
    double dx=i-xc;
    for (int j=0; j< nx; j++) {
      double dy=j-xc;
      elements[i*nx+j]=exp(-(dx*dx+dy*dy)/(2*sig*sig));
      total=total+elements[i*nx+j];
    }
  }
  for (int n=0; n< nx*nx; n++) {
    elements[n]=elements[n]/total;
  }
}

ConvKernel::ConvKernel(Attribute& attr,int nxmax) {
  vector<int> nx1=attr.get_dims();
  ndims=nx1.size();
  nx=2*nxmax+1;
  elements.resize(int(pow(double(nx),ndims)));
  // Assume (for now) that attr is a double.
  Context* c1=attr.get_context_pointer();
  int type=c1->get_order_type();
  // Assume it's symmetric about all 8 octants
  if (ndims == 3) {
    for (int i=-nxmax; i <=nxmax; i++) {
      for (int j=-nxmax; j <=nxmax; j++) {
	for (int k=-nxmax; k <=nxmax; k++) {
	  if (type == NORMAL) {
	    this->set_element(attr(abs(i),abs(j),abs(k)),i,j,k);
	  }
	  else if (type == FFT_HALF_HERMITIAN) {
	    int jout=j;
	    if (j < 0) jout=j+nx1[1];
	    int iout=i;
	    if (i < 0) iout=i+nx1[0];
	    this->set_element(attr(iout,jout,abs(k)),i,j,k);
	  }
	}
	
      }
    }
  }
}

double ConvKernel::get_element(int i, int j, int k) {
  int ic=(nx-1)/2;
  if (ndims == 3) {
    if ((abs(i) <= ic) && (abs(j) <= ic) && (abs(k) <= ic)) {
      return elements[(i+ic)*nx*nx+(j+ic)*nx+(k+ic)];
    } else{
      return 0;
    }
  } else if (ndims == 2) {
    if ((abs(i) <= ic) && (abs(j) <= ic)) {
      return elements[(i+ic)*nx+(j+ic)];
    } else{
      return 0;
    }
  } else {
    if (abs(i) <= ic) {
      return elements[i+ic];
    } else{
      return 0;
    }
  }
}

int ConvKernel::get_nx() {
  return nx;
}

int ConvKernel::get_ndims() {
  return ndims;
}

int ConvKernel::set_nx(int nxnew) {
  nx=nxnew;
  elements.resize(int(pow(double(nx),ndims)));
  return 0;
}

int ConvKernel::set_ndims(int ndimsnew) {
  ndims=ndimsnew;
  elements.resize(int(pow(double(nx),ndims)));
  return 0;
}

int ConvKernel::set_element(double val, int i, int j, int k) {
  int ic=(nx-1)/2;
  if (ndims == 3) {
    if ((abs(i) <= ic) && (abs(j) <= ic) && (abs(k) <= ic)) {
      elements[(i+ic)*nx*nx+(j+ic)*nx+(k+ic)]=val;
    } else{
      return 0;
    }
  } else if (ndims == 2) {
    if ((abs(i) <= ic) && (abs(j) <= ic)) {
      elements[(i+ic)*nx+(j+ic)]=val;
    } else{
      return 0;
    }
  } else {
    if (abs(i) <= ic) {
      elements[i+ic]=val;
    } else{
      return 0;
    }
  }
}
