#include "basin_package.h"
#include "transforms.h"

// 1d with weighting or 2d without
Grid* compute_density(List* listptr,string xname,string name2, string densname,
		      vector<int> nx, vector<double> ll, vector<double> ur,int ndims) {
  Attribute* a1=listptr->get_attribute_pointer(xname);
  Attribute* a2=listptr->get_attribute_pointer(name2);
  Attribute* adum=new Attribute();
  Attribute* adum2=new Attribute();
  Attribute dens;
  //cerr << "I am here\n";
  //cerr << ll[0] << " " << nx[0] << "\n";
  if (ndims == 2) {
    //cerr << "pt1\n";
    dens=compute_density(*a1,*a2,*adum,*adum2,nx,ll,ur,0);
  } else {
    dens=compute_density(*a1,*adum,*adum2,*a2,nx,ll,ur,0);
  }
  //cerr << "pt 2\n";
  //cerr << ll[0] << " " << nx[0] << "\n";
  Region* myregion=listptr->get_parent();
  Grid* outgrid=myregion->add_grid("DensGrid",nx);
  outgrid->set_context(ll,ur,nx);
  outgrid->add_attribute(densname,dens);
  delete(adum);
  delete(adum2);
  return outgrid;
}
Grid* compute_density(List* listptr,string xname, string name2, string densname, 
		      vector<int> nx,int ndims) {
  return compute_density(listptr,xname,name2,densname,nx,BASIN_posDefault,BASIN_posDefault);
}
Grid* compute_density(List* listptr,string xname, string name2, string densname,
		      int ndims) {
  return compute_density(listptr,xname,name2,densname,BASIN_nxDefault,ndims);
}

// 2d with weighting or 3d without
Grid* compute_density(List* listptr,string xname,string yname, string name3, string densname,
		      vector<int> nx, vector<double> ll, vector<double> ur,int ndims) {
  Attribute* a1=listptr->get_attribute_pointer(xname);
  Attribute* a2=listptr->get_attribute_pointer(yname);
  Attribute* a3=listptr->get_attribute_pointer(name3);
  Attribute* adum=new Attribute();
  Attribute dens;
  if (ndims == 3) {
    dens=compute_density(*a1,*a2,*a3,*adum,nx,ll,ur,0);
  } else {
    dens=compute_density(*a1,*a2,*adum,*a3,nx,ll,ur,0);
  }
  Region* myregion=listptr->get_parent();
  Grid* outgrid=myregion->add_grid("DensGrid",nx);
  outgrid->set_context(ll,ur,nx);
  outgrid->add_attribute(densname,dens);
  delete(adum);
  return outgrid;
}
Grid* compute_density(List* listptr,string xname, string yname, string name3, string densname,
		      vector<int> nx,int ndims) {
  return compute_density(listptr,xname,yname,name3,densname,nx,BASIN_posDefault,BASIN_posDefault);
}
Grid* compute_density(List* listptr, string xname, string yname, string name3, string densname,
		      int ndims) {
  return compute_density(listptr,xname,yname,name3,densname,BASIN_nxDefault);
}

// This is the proper calling procedure for 3-d with weighting
Grid* compute_density(List* listptr,string xname,string yname, string zname, string massname, 
		      string densname,vector<int> nx,  vector<double> ll, vector<double> ur){
  // In this call, we know that we're creating a 3-d grid with weighting given by massname;
  Attribute* a1=listptr->get_attribute_pointer(xname);
  Attribute* a2=listptr->get_attribute_pointer(yname);
  Attribute* a3=listptr->get_attribute_pointer(zname);
  Attribute* a4=listptr->get_attribute_pointer(massname);
  int ndims=3;
  
  Attribute dens=compute_density(*a1,*a2,*a3,*a4,nx,ll,ur,0);
  Region* myregion=listptr->get_parent();
  Grid* outgrid=myregion->add_grid("DensGrid",nx);
  outgrid->set_context(ll,ur,nx);
  outgrid->add_attribute(densname,dens);
  return outgrid;
}
Grid* compute_density(List* listptr,string xname, string yname, string zname, string massname,
		      string densname, vector<int> nx) {
  return compute_density(listptr,xname,yname,zname,massname,densname,nx,BASIN_posDefault,
			 BASIN_posDefault);
}
Grid* compute_density(List* listptr,string xname, string yname, string zname, string massname,
		      string densname) {
  return compute_density(listptr,xname,yname,zname,massname,densname,BASIN_nxDefault);
} 

// If a density grid has already been created, and you want to add to it.
Attribute* compute_density(List* listptr,Grid* gridptr, string xname, string name2, string densname) {
  Attribute* a1=listptr->get_attribute_pointer(xname);
  Attribute* a2=listptr->get_attribute_pointer(name2);
  Attribute* adum=new Attribute();
  Attribute* adum2=new Attribute();
  Attribute dens;
  Context* context_ptr=gridptr->get_context_pointer();
  vector<int> nx=context_ptr->get_n_grid_points();
  vector<double> ll=context_ptr->get_boundary()[0];
  vector<double> ur=context_ptr->get_boundary()[1];
  int ndims=nx.size();
  if (ndims == 2) {
    dens=compute_density(*a1,*a2,*adum,*adum2,nx,ll,ur,0);
  } else {
    dens=compute_density(*a1,*adum,*adum2,*a2,nx,ll,ur,0);
  }
  gridptr->add_attribute(densname,dens);
  Attribute* attr_ptr=gridptr->get_attribute_pointer(densname);
  delete(adum);
  delete(adum2);
  return attr_ptr;
}


// If a density grid has already been created, and you want to add to it. 2d+weight or 3d
Attribute* compute_density(List* listptr,Grid* gridptr, string xname, string yname, 
			   string name3, string densname) {
  Attribute* a1=listptr->get_attribute_pointer(xname);
  Attribute* a2=listptr->get_attribute_pointer(yname);
  Attribute* a3=listptr->get_attribute_pointer(name3);
  Attribute* adum=new Attribute();
  Attribute dens;
  Context* context_ptr=gridptr->get_context_pointer();
  vector<int> nx=context_ptr->get_n_grid_points();
  vector<double> ll=context_ptr->get_boundary()[0];
  vector<double> ur=context_ptr->get_boundary()[1];
  int ndims=nx.size();
  if (ndims == 3) {
    dens=compute_density(*a1,*a2,*a3,*adum,nx,ll,ur,0);
  } else {
    dens=compute_density(*a1,*a2,*adum,*a3,nx,ll,ur,0);
  }
  gridptr->add_attribute(densname,dens);
  Attribute* attr_ptr=gridptr->get_attribute_pointer(densname);
  delete(adum);
  return attr_ptr;
}

// If a density grid has already been created, and you want to add to it. 3d+weight
Attribute* compute_density(List* listptr,Grid* gridptr, string xname, string yname, string zname,
			   string massname, string densname) {
  Attribute* a1=listptr->get_attribute_pointer(xname);
  Attribute* a2=listptr->get_attribute_pointer(yname);
  Attribute* a3=listptr->get_attribute_pointer(zname);
  Attribute* a4=listptr->get_attribute_pointer(massname);
  Attribute dens;
  Context* context_ptr=gridptr->get_context_pointer();
  vector<int> nx=context_ptr->get_n_grid_points();
  vector<double> ll=context_ptr->get_boundary()[0];
  vector<double> ur=context_ptr->get_boundary()[1];
  int ndims=nx.size();
  dens=compute_density(*a1,*a2,*a3,*a4,nx,ll,ur,0);
  gridptr->add_attribute(densname,dens);
  Attribute* attr_ptr=gridptr->get_attribute_pointer(densname);
  return attr_ptr;
}




// Set nx with a single vlaue
Attribute compute_density(Attribute& x, Attribute& y, Attribute& z, Attribute& m, int nx,int ndims) {
  vector<int> nx_vec(ndims,nx);
  return compute_density(x,y,z,m,nx_vec);
}

Attribute compute_density(Attribute& x, Attribute& y, Attribute& val3, int nx,int ndims) {
  vector<int> nx_vec(ndims,nx);
  return compute_density(x,y,val3,nx_vec);
}

Attribute compute_density(Attribute& x, Attribute& val2, int nx,int ndims) {
  vector<int> nx_vec(ndims,nx);
  return compute_density(x,val2,nx_vec);
}

Attribute compute_density(Attribute &x,int nx) {
  vector<int> nx_vec(1,nx);
  return compute_density(x,nx_vec);
}


// Don't specify the corners
Attribute compute_density(Attribute& x, Attribute& y, Attribute& z, Attribute& m, vector<int> &nx) {
  vector<double>ll=BASIN_posDefault;
  vector<double>ur=BASIN_posDefault;
  return compute_density(x,y,z,m,nx,ll,ur);
}
Attribute compute_density(Attribute& x, Attribute& y, Attribute& val3, vector<int> &nx) {
  vector<double>ll=BASIN_posDefault;
  vector<double>ur=BASIN_posDefault;
  return compute_density(x,y,val3,nx,ll,ur);
}
Attribute compute_density(Attribute& x, Attribute& val2, vector<int> &nx) {
  vector<double>ll=BASIN_posDefault;
  vector<double>ur=BASIN_posDefault;
  return compute_density(x,val2,nx,ll,ur);
}

Attribute compute_density(Attribute& x, vector<int> &nx) {
  vector<double>ll=BASIN_posDefault;
  vector<double>ur=BASIN_posDefault;
  Attribute dum1=Attribute();
  Attribute dum2=Attribute();
  Attribute dum3=Attribute();
  return compute_density(x,dum1,dum2,dum3,nx,ll,ur);
}


// Not available to the script side (for now), because of the vector argument

// 3 Attributes
Attribute compute_density(Attribute& x, Attribute& y, Attribute& val3, vector<int> &nx,vector<double>& ll, vector<double>& ur) {
  // Is this a 1-d or 2-d call?
  int ndims=nx.size();
  Attribute dum1=Attribute();
  if (ndims == 2) {
    return compute_density(x,y,dum1,val3,nx,ll,ur,0);
  } else {
    return compute_density(x,y,val3,dum1,nx,ll,ur,0);    
  }
}

// 2 Attributes
Attribute compute_density(Attribute& x, Attribute& val2, vector<int> &nx, vector<double>& ll, vector<double>& ur) {
  // Is this a 1-d or 2-d call?
  int ndims=nx.size();
  Attribute dum1=Attribute();
  Attribute dum2=Attribute();
  if (ndims == 1) {
    return compute_density(x,dum1,dum2,val2,nx,ll,ur,0);
  } else {
    return compute_density(x,val2,dum1,dum2,nx,ll,ur,0);    
  }
}


// 3-D density computations
Attribute compute_density(Attribute& x, Attribute& y, Attribute& z, Attribute& mass, vector<int> &nx, vector<double> &ll, vector<double> &ur,int Smoothing) {


  int ndims=1;// maximum of 3
  int hi,lo,isWeight=0;
  int numprocs;
  int myid;
  MPI_Comm_rank(BASIN_COMM, &myid );
  // For diagnostics
  MPI_Comm_size(BASIN_COMM,&numprocs);


  if ((mass.get_dims())[0] == 0) {
    isWeight=0;
  }
  
  
  if ((y.get_dims())[0] > 0) {
    ndims=2;
    if ((z.get_dims())[0] > 0) {
      ndims=3;
    }
  }

  nx.resize(ndims);  
  
  // What are the ids of local particles?  (We'll loop over those)
  x.get_gen_distribution_index(myid,&lo,&hi);
  int npart_local=hi-lo+1;
  
  // If the ranges aren't defined, loop in order to find them
  if ((ll == BASIN_posDefault) || (ur == BASIN_posDefault)) {
    double *min_arr=(double*)malloc(sizeof(double)*ndims);
    double *max_arr=(double*)malloc(sizeof(double)*ndims);

    double *min_arr_red=(double*)malloc(sizeof(double)*ndims);
    double *max_arr_red=(double*)malloc(sizeof(double)*ndims);
    for (int i=0; i < ndims; i++) {
      min_arr[i]=1e8;
      max_arr[i]=-1e8;
    }
    for (int i=0; i < npart_local; i++) {
      if (double(x[i+lo]) < min_arr[0]) min_arr[0]=x[i+lo];
      if (double(x[i+lo]) > max_arr[0]) max_arr[0]=x[i+lo];
      if (ndims > 1) {
	if (double(y[i+lo]) < min_arr[1]) min_arr[1]=y[i+lo];
	if (double(y[i+lo]) > max_arr[1]) max_arr[1]=y[i+lo];
	if (ndims > 2) {
	  if (double(z[i+lo]) < min_arr[2]) min_arr[2]=z[i+lo];
	  if (double(z[i+lo]) > max_arr[2]) max_arr[2]=z[i+lo];
	}
      }
    }
    
    // Combine the min and max of the ll and ur to get a true ll and ur.
    MPI_Allreduce(min_arr,min_arr_red,ndims,MPI_DOUBLE,MPI_MIN,BASIN_COMM);
    MPI_Allreduce(max_arr,max_arr_red,ndims,MPI_DOUBLE,MPI_MAX,BASIN_COMM);
    ll.resize(ndims);
    ur.resize(ndims);
    double dlmax=0.;
    for (int i=0; i < ndims; i++) {
      double range=max_arr[i]-min_arr[i];
      ll[i]=min_arr_red[i]-0.01*range;
      ur[i]=max_arr_red[i]+0.01*range;
      if ((ur[i]-ll[i])/nx[i] > dlmax) {dlmax=(ur[i]-ll[i])/nx[i];}
    }
    // 
    for (int i=0; i < ndims; i++) {
      double lnew=nx[i]*dlmax;
      double l=ur[i]-ll[i];
      ur[i]=ur[i]+(lnew-l)/2.;
      ll[i]=ll[i]-(lnew-l)/2.;
    }


     free(min_arr);
     free(max_arr);
     free(min_arr_red);
     free(max_arr_red);
  }

  // Create an empty attribute with the appropriate size.
  Attribute gridout(nx,double(0.));
  
  // What are the local indices of the grid?
  int idx_lo,idx_hi;
  gridout.get_gen_distribution_index(myid,&idx_lo,&idx_hi);
  // Encrico will fix this in getdims;
  double del[3];
  for (int i=0; i < ndims; i++) {
    del[i]=(ur[i]-ll[i])/nx[i];
  }

  // Now, loop over all (local) the particles...
  vector<int> idx;
  vector<double> w;
  int idx_tmp;
  double w_tmp,wx,wy,wz;
  int i0,j0,k0;

  // Given the plethora of "if" statements, it would help to split this up...

  // For now, no smoothing...
  vector<double> del_local(idx_hi-idx_lo+1,0);
  if (ndims == 1) {
    for (int ipart=lo; ipart < npart_local+lo; ipart++) {
      if ((double(x[ipart]) >= ll[0]) && (double(x[ipart]) <= ur[0])) {
	i0=int((double(x[ipart])-ll[0])/del[0]);
	idx_tmp=i0;
	w_tmp=1.;
	if (isWeight == 1) w_tmp=mass[ipart];
	if ((idx_tmp >= idx_lo) && (idx_tmp <=idx_hi)) {
	  del_local[idx_tmp-idx_lo]+=w_tmp;	  
	} else {
	  idx.push_back(idx_tmp);
	  w.push_back(w_tmp);
	}
      }
    }
  } else if (ndims == 2) {
    for (int ipart=lo; ipart < npart_local+lo; ipart++) {
      if ((double(x[ipart]) >= ll[0]) && (double(x[ipart]) <= ur[0]) && 
	  (double(y[ipart]) >= ll[1]) && (double(y[ipart]) <= ur[1])) {
	i0=int((double(x[ipart])-ll[0])/del[0]);
	j0=int((double(y[ipart])-ll[1])/del[1]);
	idx_tmp=i0*nx[1]+j0;
	w_tmp=1.;
	if (isWeight == 1) w_tmp=mass[ipart];
	if ((idx_tmp >= idx_lo) && (idx_tmp <=idx_hi)) {
	  //cerr << idx_tmp << " " << w_tmp << "\n";
	  del_local[idx_tmp-idx_lo]=del_local[idx_tmp-idx_lo]+w_tmp;	  
	} else {
	  idx.push_back(idx_tmp);
	  w.push_back(w_tmp);
	}
      }
    }

  } else if (ndims == 3) {
    for (int ipart=lo; ipart < npart_local+lo; ipart++) {
      if ((double(x[ipart]) >= ll[0]) && (double(x[ipart]) <= ur[0]) && 
	  (double(y[ipart]) >= ll[1]) && (double(y[ipart]) <= ur[1]) && 
	  (double(z[ipart]) >= ll[2]) && (double(z[ipart]) <= ur[2])) {
	i0=int((double(x[ipart])-ll[0])/del[0]);
	j0=int((double(y[ipart])-ll[1])/del[1]);
	k0=int((double(z[ipart])-ll[2])/del[2]);
	idx_tmp=i0*nx[1]*nx[2]+j0*nx[2]+k0;
	w_tmp=1.;
	if (isWeight == 1) w_tmp=mass[ipart];

	if ((idx_tmp >= idx_lo) && (idx_tmp <=idx_hi)) {
	  del_local[idx_tmp-idx_lo]+=w_tmp;	  
	} else {
	  idx.push_back(idx_tmp);
	  w.push_back(w_tmp);
	}
      }
    }
  } 



  gridout.put(idx_lo,idx_hi,del_local);

  MPI_Barrier(BASIN_COMM);

  // Now, do all of the remaining processors.
  if (numprocs > 1) {
    for (int d_id=1; d_id < numprocs; d_id++) {
      int tid=((myid+d_id) % numprocs);
      gridout.get_gen_distribution_index(tid,&idx_lo,&idx_hi);

      del_local.resize(idx_hi-idx_lo+1);
      del_local.assign(idx_hi-idx_lo+1,0);
      for (int i=0; i < idx.size(); i++) {
      	if ((idx[i] >= idx_lo) and (idx[i] <= idx_hi)) {
      	  del_local[idx[i]-idx_lo]=del_local[idx[i]-idx_lo]+w[i];
      	}
      }
      gridout.add(idx_lo,idx_hi,del_local);
      MPI_Barrier(BASIN_COMM);
    }

  }


  return gridout;
}


