/*
 *  C++ core and functions for BASIN.
 *  Copyright (C) 2006  BASIN team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "basin_package.h"
#include "transforms.h"

Attribute convolve(Attribute& attr,ConvKernel kernel) {
  int rowmin,rowmax;
  Attribute attr_out=attr*0;
  attr.get_gen_distribution(BASIN_ID,&rowmin,&rowmax);
  vector <int> nx=attr.get_dims();
  int ndims=nx.size();
  int nx0=kernel.get_nx();
  int ic=(nx0-1)/2;
  double val;
  vector<int> idx;
  idx.resize(ndims);

  if (ndims == 3) {
    for (int k=rowmin; k<= rowmax; k++) {
      for (int j=0; j < nx[1]; j++) {
	for (int i=0; i < nx[0]; i++) {
	  for (int dk=-ic; dk<= ic; dk++) {
	    idx[2]=k+dk;
	    for (int dj=-ic; dj<= ic; dj++) {
	      idx[1]=j+dj;
	      for (int di=-ic; di<= ic; di++) {
		idx[0]=i+di;
		if ((i+di >=0)&&(i+di <nx[0])&&(j+dj >=0)&&(j+dj <nx[1])&&(k+dk >=0)&&(k+dk < nx[2])) {
		  double val=double(attr_out(i+di,j+dj,k+dk))+
		    kernel.get_element(di,dj,dk)*(double(attr(i,j,k)));
		  attr_out.set_element(idx,val);
		}
	      }
	    }
	  }
	}
      }
    }
  }
  else if (ndims == 2)  {
    for (int j=rowmin; j <=rowmax; j++) {
      for (int i=0; i < nx[0]; i++) {
	for (int dj=-ic; dj<= ic; dj++) {
	  idx[1]=j+dj;
	  for (int di=-ic; di<= ic; di++) {
	    idx[0]=i+di;
	    if ((i+di >=0)&&(i+di <nx[0])&&(j+dj >=0)&&(j+dj <nx[1])) {
	      double val=double(attr_out(i+di,j+dj))+
		kernel.get_element(di,dj)*(double(attr(i,j)));
	      attr_out.set_element(idx,val);
	    }
	  }
	}
      }
    }
  }
  else if (ndims == 1)  {
    for (int i=rowmin; i <=rowmax; i++) {
      for (int di=-ic; di<= ic; di++) {
	idx[0]=i+di;
	if ((i+di >=0)&&(i+di <nx[0])) {
	  double val=double(attr_out(i+di))+ kernel.get_element(di)*(double(attr(i)));
	  attr_out.set_element(idx,val);
	}
      }
    }
  }
  return attr_out;
}

Attribute convolve(vector<double> f,ConvKernel c,vector<int> nx, int cot) {
  // Takes a function, f, which is a scalar function of "r" (appropriately defined), and
  // convolves A(vec{x})=int( C(vec(x'))* B(vec(x')+vec(x)) d^n x')
  // C is the convolution kernel, and all integration units are in terms of gridcells.
  
  // First, create a new attribute and context.
  Attribute A(nx,double(0.));
  int ndims=nx.size();
  int nmax=f.size();
  vector<double> ll(ndims,0),ur(ndims,0);
  for (int i=0; i < ndims;i++) {
    ur[i]=nx[i];
  }
  Context* c1=new Context(ll,ur,nx,NULL);
  c1->set_order_type(cot);
  A.set_context(c1);

  // Now, get the local data
  int basin_id,basin_nprocs;
  MPI_Comm_size(MPI_COMM_WORLD,&basin_nprocs); 
  MPI_Comm_rank(MPI_COMM_WORLD,&basin_id);

  double* mydata;
  int lo,hi;

  A.get_gen_distribution_index(basin_id, &lo, &hi);
  mydata=(double*)malloc(sizeof(double)*(hi-lo+1));
  A.get(lo,hi,mydata);
  vector<int> id;
  double r;
  int nconv=(c.get_nx()-1)/2;


  for (int i=0; i < hi-lo+1; i++) {
    id=c1->get_index(i+lo);
    
    if (ndims == 1) {
      for (int j1=-nconv; j1<=nconv; j1++) {
	r=abs(id[0]+j1);
	int k1=int(r);
	double w=r-k1;
	if (r <= nmax) {
	  mydata[i]+=c.get_element(j1)*((1.-w)*f[k1]+w*f[k1+1]);
	}
      }
    }
    else if (ndims == 2) {
      for (int j1=-nconv; j1<=nconv; j1++) {
	for (int j2=-nconv; j2<=nconv; j2++) {
	  r=sqrt((id[0]+j1)*(id[0]+j1)+(id[1]+j2)*(id[1]+j2));
	  int k1=int(r);
	  double w=r-k1;
	  if (r <= nmax) {
	    mydata[i]+=c.get_element(j1,j2)*((1-w)*f[k1]+w*f[k1+1]);
	  }
	}
      }
    }
    else if (ndims == 3) {
      for (int j1=-nconv; j1<=nconv; j1++) {
	for (int j2=-nconv; j2<=nconv; j2++) {
	  for (int j3=-nconv; j3<=nconv; j3++) {
	    r=sqrt((double)(id[0]+j1)*(id[0]+j1)+(id[1]+j2)*(id[1]+j2)+(id[2]+j3)*(id[2]+j3));
	    int k1=int(r);
	    double w=r-k1;
	    if (r <= nmax) {
	      mydata[i]+=c.get_element(j1,j2,j3)*((1-w)*f[k1]+w*f[k1+1]);
	    }
	  }
	}
      }
    }
  }
  A.put(lo,hi,mydata);

  free(mydata);
  return A;
}

