#include "basin_package.h"
#include "transforms.h"

/* 	cylindric_to_xyz
 * INPUT: Pointer to a List, 6 Strings for Attribute names
 * Given existing attributes 'r', 'theta' and 'z',
 * this function calculates the corresponding
 * 'x', 'y' and 'z' attributes and adds them to
 * the existing list. */
string cylindrical_to_xyz (List* listptr, string r, string theta, string zIn, string x, string y, string z) {
	if (listptr == NULL) {
		return "Bad list pointer.";
	}
	string return_string = "";
	Attribute *r_ptr;
	Attribute *theta_ptr;
	Attribute *z_ptr;

	/* Check that the attributes exist
	 * and grab a pointer to them,
	 * otherwise a default value */
	if (listptr-> attribute_exist (r)) {
		r_ptr = listptr-> get_attribute_pointer (r);
	} else {
		r_ptr = listptr-> get_attribute_pointer (0);
		return_string +=  r + " not found, default 0's\n";
	}
	if (listptr-> attribute_exist (theta)) {
		theta_ptr = listptr-> get_attribute_pointer (theta);
	} else {
		theta_ptr = listptr-> get_attribute_pointer (0);
		return_string += theta + " not found, default 0's\n";
	}
	if (listptr-> attribute_exist (zIn)) {
		z_ptr = listptr-> get_attribute_pointer (zIn);
	} else {
		z_ptr = listptr-> get_attribute_pointer (0);
		return_string += zIn + " not found, default 0's\n";
	}

	/* Calculates the xyz coordinates */
	Attribute xTemp = (*r_ptr) * cos((*theta_ptr));
	Attribute yTemp = (*r_ptr) * sin((*theta_ptr));
	Attribute zTemp = (*z_ptr);
	return_string += "Created " + x + ", " + y + ", and " + z + " attributes in list!";

	/* Adds the new calculated attributes
	 * into the existing List*/
	listptr-> add_attribute (x, xTemp);
	listptr-> add_attribute (y, yTemp);
	listptr-> add_attribute (z, zTemp);

	return return_string;
}


/* 	cylindrical_to_xyz
 * INPUT: 6 pointers o Attributes
 * Given the first three cylindrical attributes, it
 * will compute the xyz coordinates which
 * are modified via pointer*/
string cylindrical_to_xyz (Attribute* r, Attribute* theta, Attribute* z, Attribute* xOut, Attribute* yOut, Attribute* zOut) {
	*xOut = (*r) * cos((*theta));
	*yOut = (*r) * sin((*theta));
	*zOut = (*z);

	return "Success";
}

