//FFT algorithm on a grid.
#include "basin_package.h"
#include "transforms.h"



#ifdef BASIN_HAVE_FFTW

Attribute fft(Attribute& attr_in) {
  // Check to see its type, and then call either fft_c2r or fft_r2c.
  if (attr_in.get_data_type() == BASIN_DATA_DOUBLE) {
    cerr << "About to do real to complex transform\n";
    return fft_r2c(attr_in); }
  else if (attr_in.get_data_type() == BASIN_DATA_COMPLEX) {
    cerr << "About to do complex to real transform\n";
    return fft_c2r(attr_in); }

}


Attribute fft_c2r(Attribute& attr_in) {
  vector<int> nx_k=attr_in.get_dims();
  int ndims=nx_k.size();
  vector<int> nx(ndims);
  double dum=0.0;
  int local_nx,local_x_start,local_ny_after_transpose,local_y_start_after_transpose,total_local_size;
  int nyloc,nzloc;
  // fftw variables
  rfftwnd_mpi_plan plan;
  fftw_complex* cdata;
  fftw_real* rdata;
  fftw_real* rwork;
  //fftw_real* data;
  //fftw_real* work;
  void* data;
  void* work;
  
  fftw_complex* basin_cdata;

  for (int i=0; i < ndims; i++) {
    nx[i]=nx_k[i];
  }
  nx[ndims-1]=2*(nx_k[ndims-1]-1);

  Attribute attr_out(nx,dum);
  
  if (ndims == 1) {

  } 
  else if (ndims > 1) {
    if (ndims == 2) {
      plan= rfftw2d_mpi_create_plan(BASIN_COMM,nx[0],nx[1], FFTW_COMPLEX_TO_REAL, FFTW_ESTIMATE | FFTW_IN_PLACE); 
    } else if (ndims == 3) {
      plan= rfftw3d_mpi_create_plan(BASIN_COMM,nx[0],nx[1],nx[2],
				      FFTW_COMPLEX_TO_REAL, FFTW_ESTIMATE | FFTW_IN_PLACE);
    }
    rfftwnd_mpi_local_sizes(plan, &local_nx, &local_x_start,
                               &local_ny_after_transpose,
                               &local_y_start_after_transpose,
                               &total_local_size);
    //data = (fftw_real*) malloc(sizeof(fftw_real) * total_local_size);
    //work = (fftw_real*) malloc(sizeof(fftw_real) * total_local_size);
    data = new fftw_complex[total_local_size];
    work = new fftw_complex[total_local_size];

    vector<int> lo(ndims,0);
    vector<int> hi(ndims,0);

    lo[0]=0;
    lo[1]=local_y_start_after_transpose;
    hi[0]=nx[0]-1;
    hi[1]=local_y_start_after_transpose+local_ny_after_transpose-1;
    if (ndims == 3) {
      nzloc=(nx[2]/2+1);
      lo[2]=0;
      hi[2]=nzloc-1;
    }
    
    if (ndims == 2) {
      basin_cdata=(fftw_complex*)malloc(sizeof(fftw_complex)*nx[0]*local_ny_after_transpose);
    } else if (ndims == 3) {
      basin_cdata=(fftw_complex*)malloc(sizeof(fftw_complex)*nx[0]*local_ny_after_transpose*nzloc);
    }
    attr_in.get(lo,hi,basin_cdata);
    cdata = (fftw_complex*)data;
    if (ndims == 2) {
      nyloc=2*(nx_k[1]/2+1);
      for (int y = 0; y < local_ny_after_transpose; y++) {
	for (int x = 0; x < nx[0]; x++) {
	  cdata[y*nx[0]+x]=basin_cdata[x*local_ny_after_transpose+y];
	}
      }
    }
    else if (ndims ==3) {
      for (int y = 0; y < local_ny_after_transpose; y++) {
	for (int x = 0; x < nx[0]; x++) {
	  for (int z=0; z < nzloc; z++) {
	    int id=z+nzloc*y+nzloc*local_ny_after_transpose*x;
	    cdata[(y*nx[0] + x) * nzloc + z]=basin_cdata[id];
	  }
	}
      }
    }

    rdata = (fftw_real*) data;
    rwork = (fftw_real*) work;
    MPI_Barrier(BASIN_COMM);
    rfftwnd_mpi(plan, 1, rdata, rwork, FFTW_TRANSPOSED_ORDER);
    
    lo[0]=local_x_start;
    hi[0]=local_x_start+local_nx-1;
    lo[1]=0;
    hi[1]=nx[1]-1;
    if (ndims == 3) {
      lo[2]=0;
      hi[2]=nx[2]-1;
    }


    // Copy the data back
    if (ndims == 2) {
      double* basin_rdata = new double[local_nx * nx[1]];
      nyloc=2*(nx[1]/2+1);
      for (int x=0; x < local_nx; x++) {
	for (int y=0; y < nx[1]; y++) {
	  basin_rdata[y+x*nx[1]]=((fftw_real*)data)[y+x*nyloc];
	}
      }
      attr_out.put(lo,hi,basin_rdata);
      delete [] basin_rdata;
    }
    else if (ndims == 3) {
      double* basin_rdata = new double[local_nx * nx[1] * nx[2]];
      nzloc=2*(nx[2]/2+1);
      for (int x=0; x< local_nx; x++) {
	for (int y=0; y < nx[1]; y++) {
	  for (int z=0; z< nx[2]; z++) {
	    basin_rdata[z+y*nx[2]+x*nx[1]*nx[2]]=((fftw_real*) data)[(x*nx[1] + y) * nzloc + z];
	  }
	}
      }
      attr_out.put(lo,hi,basin_rdata);
      delete [] basin_rdata;
    }
    free(basin_cdata);
    //free((fftw_real*) data);
    //free((fftw_real*) work);
    delete [] (fftw_complex*) data;
    delete [] (fftw_complex*) work;
    rfftwnd_mpi_destroy_plan(plan);
  }

  return attr_out;
}

//**************************************************************************************

Attribute fft_r2c(Attribute& attr_in) {
  vector<int> nx=attr_in.get_dims();
  int ndims=nx.size();
  vector<int> nx_k(ndims);
  fftw_complex cdum;
  int local_nx,local_x_start,local_ny_after_transpose,local_y_start_after_transpose,total_local_size;
  // fftw variables
  rfftwnd_mpi_plan plan;
  fftw_real* data;
  fftw_real* work;
  fftw_complex* cdata;
  fftw_complex* basin_cdata;
  double* basin_rdata;

  for (int i=0; i < ndims; i++) {
    nx_k[i]=nx[i];
  }
  nx_k[ndims-1]=(nx[ndims-1])/2+1;

  Attribute attr_out(nx_k,cdum);
  vector<double> ll(ndims,0.),ur(ndims,0.);
  for (int i=0; i < nx_k.size(); i++) {
    ur[i]=nx_k[i];
  }
  Context *c=new Context(ll,ur,nx_k,NULL);
  c->set_order_type(FFT_HALF_HERMITIAN);  
  attr_out.set_context(c);
  
  //cerr <<"Header stuff done\n";

  if (ndims == 1) {

  } 
  else if (ndims > 1) {
    if (ndims == 2) {
      plan= rfftw2d_mpi_create_plan(BASIN_COMM,nx[0],nx[1], FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE | FFTW_IN_PLACE); 
    } else if (ndims == 3) {
      plan= rfftw3d_mpi_create_plan(BASIN_COMM,nx[0],nx[1],nx[2],
				    FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE | FFTW_IN_PLACE);
      //cerr << "plan created\n";
    }
    rfftwnd_mpi_local_sizes(plan, &local_nx, &local_x_start,
			    &local_ny_after_transpose,
			    &local_y_start_after_transpose,
			    &total_local_size);
    data = (fftw_real*) malloc(sizeof(fftw_real)*total_local_size);
    work = (fftw_real*) malloc(sizeof(fftw_real)*total_local_size);
    
    
    //cerr << "memory allocated\n";

    vector<int> lo(ndims,0);
    vector<int> hi(ndims,0);
    lo[0]=local_x_start;
    hi[0]=local_x_start+local_nx-1;
    lo[1]=0;
    hi[1]=nx[1]-1;
    if (ndims == 2) {
      basin_rdata=(double*)malloc(sizeof(double)*local_nx*nx[1]);
    }
    else if (ndims == 3) {
      lo[2]=0;
      hi[2]=nx[2]-1;
      basin_rdata=(double*)malloc(sizeof(double)*local_nx*nx[1]*nx[2]);
    }
    attr_in.get(lo,hi,basin_rdata);
    int nyloc,nzloc;
    if (ndims == 2) {
      nyloc=2*(nx[1]/2+1);
      for (int x=0; x < local_nx; x++) {
	for (int y=0; y < nx[1]; y++) {
	  data[y+x*nyloc]=basin_rdata[y+x*nx[1]];
	}
      }
    }
    else if (ndims == 3) {
      nzloc=2*(nx[2]/2+1);
      for (int x=0; x< local_nx; x++) {
	for (int y=0; y < nx[1]; y++) {
	  for (int z=0; z< nx[2]; z++) {
	    data[(x*nx[1] + y) * nzloc + z]=basin_rdata[z+y*nx[2]+x*nx[1]*nx[2]];
	  }
	}
      }
    }
    MPI_Barrier(BASIN_COMM);
    rfftwnd_mpi(plan, 1, data, work, FFTW_TRANSPOSED_ORDER);
    //cerr << "The actual fft is done\n";
    cdata=(fftw_complex*)data;
    lo[0]=0;
    lo[1]=local_y_start_after_transpose;
    hi[0]=nx[0]-1;
    hi[1]=local_y_start_after_transpose+local_ny_after_transpose-1;
    if (ndims == 3) {
      nzloc=nzloc/2;
      lo[2]=0;
      hi[2]=nzloc-1;
    }

    // Copy the data back
    if (ndims == 2) {
      basin_cdata=(fftw_complex*)malloc(sizeof(fftw_complex)*nx[0]*local_ny_after_transpose);
      double fact=1./(nx[0]*nx[1]);
      for (int y = 0; y < local_ny_after_transpose; y++) {
	
	for (int x = 0; x < nx[0]; x++) {
	  basin_cdata[x*local_ny_after_transpose+y] = cdata[y*nx[0]+x]*fact;
	}
      }
      attr_out.put(lo,hi,basin_cdata);
    } else if (ndims == 3) {
      //cerr << "About to copy the data back\n";
      basin_cdata=(fftw_complex*)malloc(sizeof(fftw_complex)*nx[0]*local_ny_after_transpose*nzloc);
      double fact=1./(nx[0]*nx[1]*nx[2]);
      for (int y = 0; y < local_ny_after_transpose; y++) {
	for (int x = 0; x < nx[0]; x++) {
	  for (int z=0; z < nzloc; z++) {
	    basin_cdata[z+nzloc*y+nzloc*local_ny_after_transpose*x] = fftw_complex(cdata[(y*nx[0] + x) * nzloc + z]*fact);
	  }
	}
      }
      //cerr << lo[0] << " " << hi[0] << " " << lo[1] << " " << hi[1] << " " << lo[2] << " " << hi[2] << "\n";

      
      for (int x=0; x < nx[0]; x++) {
	//cerr << x << "\n";
	lo[0]=x;
	hi[0]=x;	
	attr_out.put(lo,hi,basin_cdata+nzloc*local_ny_after_transpose*x);
      }
    }
    //cerr << "Data copied back\n";
    MPI_Barrier(BASIN_COMM);
    free(basin_rdata);
    free(basin_cdata);
    free(data);
    free(work);
    MPI_Barrier(BASIN_COMM);
    rfftwnd_mpi_destroy_plan(plan);
  }

  return attr_out;
}


#endif //BASIN_HAVE_FFTW
