#include "basin_package.h"
#include "transforms.h"

/* 	polar_to_xy
 * INPUT: Pointer to a List, 4 Strings for Attribute names
 * Given existing attributes 'r' and 'theta',
 * this function calculates the corresponding
 * 'x' and 'y' attributes and adds them to
 * the existing list. */
string polar_to_xy (List* listptr, string r, string theta, string x, string y) {
	if (listptr == NULL) {
		return "Bad list pointer.";
	}
	string return_string = "";
	Attribute *r_ptr;
	Attribute *theta_ptr;

	/* Check that the attributes exist
	 * and grab a pointer to them,
	 * otherwise a default value */
	if (listptr-> attribute_exist (r)) {
		r_ptr = listptr-> get_attribute_pointer (r);
	} else {
		r_ptr = listptr-> get_attribute_pointer (0);
		return_string += r + " not found, default 0's\n";
	}
	if (listptr-> attribute_exist (theta)) {
		theta_ptr = listptr-> get_attribute_pointer (theta);
	} else {
		theta_ptr = listptr-> get_attribute_pointer (0);
		return_string += theta + " not found, default 0's\n";
	}

	/* Calculates the xy coordinates */
	Attribute xTemp = (*r_ptr) * cos((*theta_ptr));
	Attribute yTemp = (*r_ptr) * sin((*theta_ptr));
	
	return_string += "Created " + x + " and " + y + " attributes in list!";

	/* Adds the new calculated attributes
	 * into the existing List*/
	listptr-> add_attribute (x, xTemp);
	listptr-> add_attribute (y, yTemp);

	return return_string;
}


/* 	polar_to_xy
 * INPUT: 4 pointers o Attributes
 * Given the first two polar attributes, it
 * will compute the second xy coordinates which
 * are modified via pointer*/
string polar_to_xy (Attribute* r, Attribute* theta, Attribute* xOut, Attribute *yOut) {
	*xOut = (*r) * cos((*theta));
	*yOut = (*r) * sin((*theta));

	return "Success";
}

