#include "basin_package.h"
#include "transforms.h"

/* 	spherical_to_xyz
 * INPUT: Pointer to a List, 6 Strings for Attribute names
 * Given existing attributes 'rho', 'phi' and 'theta',
 * this function calculates the corresponding
 * 'x', 'y' and 'z' attributes and adds them to
 * the existing list. */
string spherical_to_xyz (List* listptr, string rho, string theta, string phi, string x, string y, string z) {
	if (listptr == NULL) {
		return "Bad list pointer.";
	}
	string return_string = "";
	Attribute *rho_ptr;
	Attribute *phi_ptr;
	Attribute *theta_ptr;

	/* Check that the attributes exist
	 * and grab a pointer to them,
	 * otherwise a default value */
	if (listptr-> attribute_exist (rho)) {
		rho_ptr = listptr-> get_attribute_pointer (rho);
	} else {
		rho_ptr = listptr-> get_attribute_pointer (0);
		return_string +=  rho + " not found, default 0's\n";
	}
	if (listptr-> attribute_exist (phi)) {
		phi_ptr = listptr-> get_attribute_pointer (phi);
	} else {
		phi_ptr = listptr-> get_attribute_pointer (0);
		return_string += phi + " not found, default 0's\n";
	}
	if (listptr-> attribute_exist (theta)) {
		theta_ptr = listptr-> get_attribute_pointer (theta);
	} else {
		theta_ptr = listptr-> get_attribute_pointer (0);
		return_string += theta + " not found, default 0's\n";
	}

	/* Calculates the xyz coordinates */
	Attribute xTemp = (*rho_ptr) * sin((*phi_ptr)) * cos ((*theta_ptr));
	Attribute yTemp = (*rho_ptr) * sin((*phi_ptr)) * sin((*theta_ptr));
	Attribute zTemp = (*rho_ptr) * cos ((*phi_ptr));
	return_string += "Created " + x + ", " + y + ", and " + z + " attributes in list!";

	/* Adds the new calculated attributes
	 * into the existing List*/
	listptr-> add_attribute (x, xTemp);
	listptr-> add_attribute (y, yTemp);
	listptr-> add_attribute (z, zTemp);

	return return_string;
}


/* 	spherical_to_xyz
 * INPUT: 6 pointers o Attributes
 * Given the first three spherical attributes, it
 * will compute the xyz coordinates which
 * are modified via pointer*/
string spherical_to_xyz (Attribute* rho, Attribute* theta, Attribute* phi, Attribute* xOut, Attribute* yOut, Attribute* zOut) {
	*xOut = (*rho) * sin((*phi)) * cos((*theta));
	*yOut = (*rho) * sin((*phi)) * sin((*theta));
	*zOut = (*rho) * cos((*phi));

	return "Success";
}

