#include "basin_package.h"
#include "transforms.h"

/* 	xy_to_polar
 * INPUT: Pointer to a List, 4 Strings for Attribute names
 * Given existing attributes 'x' and 'y',
 * this function calculates the corresponding
 * 'r' and 'theta' attributes and adds them to
 * the existing list. */
string xy_to_polar (List* listptr, string x, string y, string r, string theta) {
	if (listptr == NULL) {
		return "Bad list pointer.";
	}
	string return_string = "";
	Attribute *x_ptr;
	Attribute *y_ptr;

	/* Check that the attributes exist
	 * and grab a pointer to them,
	 * otherwise a default value */
	if (listptr-> attribute_exist (x)) {
		x_ptr = listptr-> get_attribute_pointer (x);
	} else {
		x_ptr = listptr-> get_attribute_pointer (0);
		return_string += x + " not found, default 0's\n";
	}
	if (listptr-> attribute_exist (y)) {
		y_ptr = listptr-> get_attribute_pointer (y);
	} else {
		y_ptr = listptr-> get_attribute_pointer (0);
		return_string += y + " not found, default 0's\n";
	}

	/* Calculates the r,theta coordinates */
	Attribute rTemp = sqrt (pow ((*x_ptr),2) + pow ((*y_ptr), 2));
	/* Don't divide by 0 */
	Attribute thetaTemp = where((*x_ptr) != 0, atan ((*y_ptr) / (*x_ptr)), (*x_ptr) * 0);
	return_string += "Created " + r + " and " + theta + " attributes in list!";

	/* Adds the new calculated attributes
	 * into the existing List*/
	listptr-> add_attribute (r, rTemp);
	listptr-> add_attribute (theta, thetaTemp);

	return return_string;
}


/* 	xy_to_polar
 * INPUT: 4 pointers o Attributes
 * Given the first two polar attributes, it
 * will compute the xy coordinates which
 * are modified via pointer*/
string xy_to_polar (Attribute* x, Attribute* y, Attribute* rOut, Attribute *thetaOut) {
	*rOut = sqrt (pow ((*x), 2) + pow ((*y), 2));
	*thetaOut = where((*x) != 0, atan ((*y) / (*x)), (*x) * 0);

	return "Success";
}

