#include "basin_package.h"
#include "transforms.h"

/* 	xyz_to_cylindrical
 * INPUT: Pointer to a List, 6 Strings for Attribute names
 * Given existing attributes 'x', 'y' and 'z',
 * this function calculates the corresponding
 * 'r', 'theta' and 'z' attributes and adds them to
 * the existing list. */
string xyz_to_cylindrical (List* listptr, string x, string y, string zIn, string r, string theta, string z) {
	if (listptr == NULL) {
		return "Bad list pointer.";
	}
	string return_string = "";
	Attribute *x_ptr;
	Attribute *y_ptr;
	Attribute *z_ptr;

	/* Check that the attributes exist
	 * and grab a pointer to them,
	 * otherwise a default value */
	if (listptr-> attribute_exist (x)) {
		x_ptr = listptr-> get_attribute_pointer (x);
	} else {
		x_ptr = listptr-> get_attribute_pointer (0);
		return_string +=  x + " not found, default 0's\n";
	}
	if (listptr-> attribute_exist (y)) {
		y_ptr = listptr-> get_attribute_pointer (y);
	} else {
		y_ptr = listptr-> get_attribute_pointer (0);
		return_string += y + " not found, default 0's\n";
	}
	if (listptr-> attribute_exist (zIn)) {
		z_ptr = listptr-> get_attribute_pointer (zIn);
	} else {
		z_ptr = listptr-> get_attribute_pointer (0);
		return_string += zIn + " not found, default 0's\n";
	}

	/* Calculates the rho, phi, theta coordinates */
	Attribute rTemp = sqrt (pow ((*x_ptr), 2) + pow ((*y_ptr), 2));
	// Don't divide by 0
	Attribute thetaTemp = where ((*x_ptr) <= 0, atan((*y_ptr) / (*x_ptr)), (*x_ptr) * 0);
	Attribute zTemp = (*z_ptr);
	return_string += "Created " + r + ", " + theta + ", and " + z + " attributes in list!";

	/* Adds the new calculated attributes
	 * into the existing List*/
	listptr-> add_attribute (r, rTemp);
	listptr-> add_attribute (theta, thetaTemp);
	listptr-> add_attribute (z, zTemp);

	return return_string;
}


/* 	xyz_to_cylindrical
 * INPUT: 6 pointers o Attributes
 * Given the first three cartesian attributes, it
 * will compute the cylindrical coordinates which
 * are modified via pointer*/
string xyz_to_cylindrical (Attribute* x, Attribute* y, Attribute* z, Attribute* rOut, Attribute* thetaOut, Attribute* zOut) {
	*rOut = sqrt (pow ((*x), 2) + pow ((*y), 2));
	*thetaOut = where ((*x) <= 0, atan((*y) / (*x)), (*x) * 0);
	*zOut = (*z);

	return "Success";
}

