#include "basin_package.h"
#include "transforms.h"

/* 	xyz_to_spherical
 * INPUT: Pointer to a List, 6 Strings for Attribute names
 * Given existing attributes 'x', 'y' and 'y',
 * this function calculates the corresponding
 * 'rho', 'phi' and 'theta' attributes and adds them to
 * the existing list. */
string xyz_to_spherical (List* listptr, string x, string y, string z, string rho, string theta, string phi) {
	if (listptr == NULL) {
		return "Bad list pointer.";
	}
	string return_string = "";
	Attribute *x_ptr;
	Attribute *y_ptr;
	Attribute *z_ptr;

	/* Check that the attributes exist
	 * and grab a pointer to them,
	 * otherwise a default value */
	if (listptr-> attribute_exist (x)) {
		x_ptr = listptr-> get_attribute_pointer (x);
	} else {
		x_ptr = listptr-> get_attribute_pointer (0);
		return_string +=  x + " not found, default 0's\n";
	}
	if (listptr-> attribute_exist (y)) {
		y_ptr = listptr-> get_attribute_pointer (y);
	} else {
		y_ptr = listptr-> get_attribute_pointer (0);
		return_string += y + " not found, default 0's\n";
	}
	if (listptr-> attribute_exist (z)) {
		z_ptr = listptr-> get_attribute_pointer (z);
	} else {
		z_ptr = listptr-> get_attribute_pointer (0);
		return_string += z + " not found, default 0's\n";
	}

	/* Calculates the rho, phi, theta coordinates */
	Attribute rhoTemp = sqrt(pow((*x_ptr), 2) + pow((*y_ptr), 2) + pow ((*z_ptr), 2));
	//Don't divide by 0
	Attribute phiTemp = where ((*x_ptr) != 0, atan ((*y_ptr) / (*x_ptr)), (*x_ptr) * 0);
	Attribute thetaTemp = where ((rhoTemp) != 0, acos((*z_ptr) / (rhoTemp)), (*x_ptr) * 0);
	return_string += "Created " + rho + ", " + phi + ", and " + theta + " attributes in list!";

	/* Adds the new calculated attributes
	 * into the existing List*/
	listptr-> add_attribute (rho, rhoTemp);
	listptr-> add_attribute (phi, phiTemp);
	listptr-> add_attribute (theta, thetaTemp);

	return return_string;
}


/* 	xyz_to_spherical
 * INPUT: 6 pointers o Attributes
 * Given the first three cartesian attributes, it
 * will compute the spherical coordinates which
 * are modified via pointer*/
string xyz_to_spherical (Attribute* x, Attribute* y, Attribute* z, Attribute* rhoOut, Attribute* thetaOut, Attribute* phiOut) {
	*rhoOut = sqrt(pow((*x), 2) + pow((*y), 2) + pow ((*z), 2));
	*phiOut = where ((*x) != 0, atan ((*y) / (*x)), (*x) * 0);
	*thetaOut = where ((*rhoOut) != 0, acos((*z) / (*rhoOut)), (*z) * 0);

	return "Success";
}

