#ifndef __VISIT_MODULE_DATATYPES_H__
#define __VISIT_MODULE_DATATYPES_H__


#include <map>
#include <string>
#include <boost/thread.hpp>

#include "VisItDataInterface_V1.h"
#include "VisItControlInterface_V1.h"


// This is not as clean as I'd like it to be with all of the globals and the static
// VisitCache. However, the design of VisIt's simulation interface
// makes this necessary, as it only allows one plotting engine and state
// per process. 

namespace visit_module {

using std::string;
using std::map;


extern boost::thread *visit_thread;

struct Mesh {
    string name;
    int type;
    int nDims;
    VisIt_MeshMetaData *md;
    VisIt_MeshData *data;
};

struct Variable {
    string name;
    string meshName;
    int size;
    VisIt_ScalarMetaData *md;
    VisIt_ScalarData *data;
};




/**
 * This class manages all of the data and metadata 
 * associated with plotting in Visit
 */
class VisitCache {
    
public:
    VisitCache();
    ~VisitCache();

    
    /**
     * the make_* methods will always replace any existing object of the
     * same name.
     */
    void make_point_mesh(string name, int nDims, Attribute *x, Attribute *y, Attribute *z);
    void make_grid_mesh(string name, vector<double>& ll, vector<double>& ur, vector<int>& nx, int li, int hi);
    void make_variable(string name, string meshName, Attribute *d);


    VisIt_SimulationMetaData* get_md();
    VisIt_MeshData* get_mesh(string name);
    VisIt_ScalarData* get_var(string name);

private:
    boost::mutex _mutex;
    
    VisIt_SimulationMetaData *sim_md;
    map<string, Mesh> mesh_map;
    map<string, Variable> var_map;
    

    // Ensures that all the data allocated for these
    // structs is properly released
    void free_Mesh(Mesh &m);
    void free_Variable(Variable &v);

    
    // Some methods to deal with VisIt's Data interface structures
    void free_SimMD();
    void free_MeshMD(VisIt_MeshMetaData *mmd);
    void free_MeshData(int mesh_type, void *mesh);
    void free_ScalarMD(VisIt_ScalarMetaData *md);
    void free_ScalarData(VisIt_ScalarData *sd);
    void free_DataArray(VisIt_DataArray da);

    VisIt_MeshMetaData* copy_MeshMD(const VisIt_MeshMetaData *mmd);
    VisIt_ScalarMetaData* copy_ScalarMD(const VisIt_ScalarMetaData *smd);


};



extern VisitCache cache;
} //end namespace visit_plot

MPI_Comm visit_comm;


#endif //__VISIT_MODULE_DATATYPES_H__

